/*-------------------------------------
 * 
 * Copyright (c) 2000 by The University of North Carolina
 *
 * See the file "packages/mpeg2/nc_license.txt" for information 
 * on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *-------------------------------------
 */

#ifndef _DVM_MPEG_2_H
#define _DVM_MPEG_2_H

#include "dvmbasic.h"
#include "bitparser.h"

#ifdef __cplusplus
extern "C" {
#endif

#include "mpeg2compat.h"

  
  /* Return codes */

#define DVM_MPEG2_OK 0
#define DVM_MPEG2_NOT_FOUND -1
#define DVM_MPEG2_INVALID_START_CODE -2
#define DVM_MPEG2_EXTENSION_ID_MISMATCH -3
#define DVM_MPEG2_ERROR -1

  /* Start codes */

#define MPEG2_PIC_START_CODE 0x00000100
#define MPEG2_SLICE_MIN_START_CODE 0x00000101
#define MPEG2_SLICE_MAX_START_CODE 0x000001af
#define MPEG2_USER_START_CODE 0x000001b2
#define MPEG2_SEQ_START_CODE 0x000001b3
#define MPEG2_SEQ_ERROR_CODE 0x000001b4
#define MPEG2_EXT_START_CODE 0x000001b5
#define MPEG2_SEQ_END_CODE   0x000001b7
#define MPEG2_GOP_START_CODE 0x000001b8

  /* Extension ids */
#define MPEG2_SEQ_HDR_EXT_ID 1
#define MPEG2_SEQ_DISPLAY_EXT_ID 2
#define MPEG2_QUANT_MATRIX_EXT_ID 3
#define MPEG2_COPYRIGHT_EXT_ID 4
#define MPEG2_SEQ_SCALABLE_EXT_ID 5
#define MPEG2_PIC_DISPLAY_EXT_ID 7
#define MPEG2_PIC_CODING_EXT_ID 8
#define MPEG2_PIC_SPATIAL_EXT_ID 9
#define MPEG2_PIC_TEMPORAL_EXT_ID 10

  /* Aspect ratio codes */
#define ASPECT_FORBIDDEN 0
#define ASPECT_SQUARE 1
#define ASPECT_DAR_3_by_4 2
#define ASPECT_DAR_9_by_16 3
#define ASPECT_DAR_1_by_2_21 4
#define ASPECT_DAR_RESERVED 5
 
  /* Frame rate codes */
#define FRAME_RATE_FORBIDDEN 0
#define FRAME_RATE_24000_div_1001 1
#define FRAME_RATE_24 2
#define FRAME_RATE_25 3
#define FRAME_RATE_30000_div_1001 4
#define FRAME_RATE_30 5
#define FRAME_RATE_50 6
#define FRAME_RATE_60000_div_1001 7
#define FRAME_RATE_60 8
#define FRAME_RATE_RESERVED 9

  /* Profile and level codes */
#define PROFILE_ESCAPE_BIT 0x80
#define PROFILE_PROFILE_MASK 0x70
#define PROFILE_LEVEL_MASK 0xf
#define PROFILE_PROFILE_SHIFT 4
#define PROFILE_LEVEL_SHIFT 0
#define PROFILE_GET_PROFILE(c) ((c & PROFILE_PROFILE_MASK) >> PROFILE_PROFILE_SHIFT)
#define PROFILE_GET_LEVEL(c) ((c & PROFILE_LEVEL_MASK) >> PROFILE_LEVEL_SHIFT)

#define PROFILE_CODE_SIMPLE 5
#define PROFILE_CODE_MAIN 4
#define PROFILE_CODE_SNR 3
#define PROFILE_CODE_SPATIAL 2
#define PROFILE_CODE_HIGH 1

#define LEVEL_CODE_LOW 10
#define LEVEL_CODE_MAIN 8
#define LEVEL_CODE_HIGH_1440 6
#define LEVEL_CODE_HIGH 4

  /* Chroma format codes */
#define CHROMA_CODE_RESERVED 0
#define CHROMA_CODE_4_2_0 1
#define CHROMA_CODE_4_2_2 2
#define CHROMA_CODE_4_4_4 3

  /* Video format codes */
#define VIDEO_FORMAT_COMPONENT 0
#define VIDEO_FORMAT_PAL 1
#define VIDEO_FORMAT_NTSC 2
#define VIDEO_FORMAT_SECAM 3
#define VIDEO_FORMAT_MAC 4
#define VIDEO_FORMAT_UNSPECIFIED 5
#define VIDEO_FORMAT_RESERVED1 6
#define VIDEO_FORMAT_RESERVED2 7

  /* Frame type codes */
#define I_FRAME 1
#define P_FRAME 2
#define B_FRAME 3

  /* Structure Codes */
#define STRUCTURE_CODE_FRAME 3
#define STRUCTURE_CODE_BOTTOM 2
#define STRUCTURE_CODE_TOP 1
#define STRUCTURE_CODE_RESERVED 0

  /* Motion Types */
#define MPEG2_FIELD_MOTION_TYPE_FIELD 1
#define MPEG2_FIELD_MOTION_TYPE_16_X_8_MC 2
#define MPEG2_FIELD_MOTION_TYPE_DUAL_PRIME 3

#define MPEG2_FRAME_MOTION_TYPE_FIELD 5
#define MPEG2_FRAME_MOTION_TYPE_DUAL_PRIME 6
#define MPEG2_FRAME_MOTION_TYPE_FRAME 7

  /* Motion Vector Formats */
#define MV_FORMAT_FIELD 0
#define MV_FORMAT_FRAME 1

  /* Q Scale Types */
#define MPEG2_LINEAR_Q_SCALE 0
#define MPEG2_NON_LINEAR_Q_SCALE 1

  /* Structures */

  /*
   *--------------------------
   * type Mpeg2SeqHdr
   *
   *   The Mpeg2SeqHdr structure contains all the parameters specified
   *   in the MPEG2 sequence header, sequence header extension,
   *   sequence display extension, and sequence scalable extension.
   *  
   *   Except for the sequence header extension which is always present
   *   in an MPEG 2 stream, the other extension headers may or may not
   *   be present. Flags for whether they were parsed into this structure
   *   are maintained. If they did not exists (or were not parsed into
   *   this structure) some fields of this structure may be undefined.
   *
   *---------------------------------------------
   */
  typedef struct Mpeg2SeqHdr {
    unsigned short horizontal_size;  /* Width of encoded frames. */
    unsigned short vertical_size;    /* Height of encoded frames. */
    unsigned short mb_width;         /* Width in macroblocks. */
    unsigned short mb_height;        /* Height in macroblocks. */
    short profile_code;              /* Profile and level indicator in 
					sequence extension header. */
    short progressive_seq_flag;      /* Boolean indicating if sequence is
					strictly progressive. */
    short chroma_format_code;        /* Code indicating chroma subsampling
					format (i.e. 4:2:0, 4:2:2, or 4:4:4) */
    short aspect_ratio_code;         /* Aspect ratio code. */
    short frame_rate_ext_n;          /* Frame rate extension factor "n",
					ref: 6.3.3 */
    short frame_rate_ext_d;          /* Frame rate extension factor "d",
					ref: 6.3.3 */
    short frame_rate_code;           /* Frame rate code. */

    short low_delay_flag;            /* Boolean. If set, guarantees no
					B frames */
    int bitrate;                     /* Bitrate in units of 400b/s */
    int vbv_buffer_size;             /* VBV buffer size spec, defined
				        in units of 16KB */
    
    /* 
     * Note: all quant tables in this structure are stored in
     * row major order. Also, the chrominance quant tables are only
     * used for non 4:2:0 data. 
     */

    unsigned char lum_iqt[64];       /* Luminance intra quant table */
    unsigned char lum_niqt[64];      /* Luminance non-intra quant table */
    unsigned char chr_iqt[64];       /* Chrominance intra quant table */
    unsigned char chr_niqt[64];      /* Chrominance  non-intra quant table */

    short display_ext_flag;

    /*
     * Fields below here are given in sequence display extension which
     * may or may not be present. Check display_ext_flag to determine
     * if any of this has been defined (i.e., parsed) first.
     */

    unsigned char video_format_code;          /* Video format */

    unsigned char color_primary_code;         /* Color info   */
    unsigned char color_xfer_characteristic_code;
    unsigned char color_matrix_coeff_code;

    short display_horizontal_size;
    short display_vertical_size;

    short scalable_ext_flag;                  /* Indicate presence of
					       * seq. scalable extension. */
  } Mpeg2SeqHdr;

  typedef struct ShortVector {
    char exists;            /* 0 iff no vector */
    short down;
    short right;
  } ShortVector;
  
  typedef struct Mpeg2PicHdr {
    unsigned short temporal_reference;        /* Timestamp  */
    unsigned short picture_type;              /* I, P, or B */
    unsigned short vbv_delay;                 /* Buffer parameter */
    unsigned char fcode[2][2];                /* Fcodes */
    unsigned char intra_dc_precision;
    unsigned char picture_structure_code;
    unsigned char top_field_first_flag;
    unsigned char frame_pred_frame_dct_flag;
    unsigned char concealment_mv_flag;
    unsigned char q_scale_type;
    unsigned char intra_vlc_format_code;
    unsigned char alternate_scan_flag;
    unsigned char repeat_first_field_flag;
    unsigned char chroma_420_type;
    unsigned char progressive_frame_flag;

    unsigned char composite_display_flag;
    unsigned char v_axis;
    unsigned char field_sequence;
    unsigned char sub_carrier;
    unsigned char burst_amplitude;
    unsigned char sub_carrier_phase;

    ShortVector PMV[2][2];                      /* Motion vector predictors */
  } Mpeg2PicHdr;

  typedef struct Mpeg2MBInfo {
    unsigned char quant;
    unsigned char motion_forward;
    unsigned char motion_backward;
    unsigned char intra;
    unsigned short pattern;
    unsigned char motion_type;
    unsigned char dct_type;
    unsigned char mv_field_select[2][2];
    unsigned char mv_format;
    unsigned char dmv_flag;
    unsigned char mv_count;
    ShortVector mvectors[2][2];
  } Mpeg2MBInfo;

  typedef struct Mpeg2MBInfoArray {
    int width;
    int height;
    Mpeg2MBInfo *mbinfo;
  } Mpeg2MBInfoArray;

  /*
   *----------------------------------
   * 
   * Function prototypes and macro defined accessors.
   *
   *----------------------------------
   */

  /*
   * Mpeg2SeqHdr
   */

  Mpeg2SeqHdr *Mpeg2SeqHdrNew();
  void Mpeg2SeqHdrFree(Mpeg2SeqHdr *hdr);

  int Mpeg2SeqHdrFind(BitParser *);
  int Mpeg2SeqHdrParse(BitParser *, Mpeg2SeqHdr *hdr);

  int Mpeg2SeqHdrExtFind(BitParser *);
  int Mpeg2SeqHdrExtParse(BitParser *, Mpeg2SeqHdr *hdr);
  
  double Mpeg2SeqHdrGetAspectRatio(Mpeg2SeqHdr *hdr);
  int Mpeg2SeqHdrGetChromaFormat(Mpeg2SeqHdr *hdr);
  double Mpeg2SeqHdrGetFrameRate(Mpeg2SeqHdr *hdr);

#define Mpeg2SeqHdrGetWidth(h) (h)->horizontal_size
#define Mpeg2SeqHdrGetHeight(h) (h)->vertical_size
#define Mpeg2SeqHdrGetMBWidth(h) (h)->mb_width
#define Mpeg2SeqHdrGetMBHeight(h) (h)->mb_height
#define Mpeg2SeqHdrGetProfile(h) PROFILE_GET_PROFILE(((h)->profile_code))
#define Mpeg2SeqHdrGetLevel(h) PROFILE_GET_LEVEL(((h)->profile_code))
#define Mpeg2SeqHdrIsProgressive(h) (h)->progressive_seq_flag
#define Mpeg2SeqHdrIsLowDelay(h) (h)->low_delay_flag
#define Mpeg2SeqHdrGetBitRate(h) (h)->bitrate
#define Mpeg2SeqHdrGetVBVBufferSize(h) (h)->vbv_buffer_size
#define Mpeg2SeqHdrGetLumIQuantMatrix(h) (h)->lum_iqt
#define Mpeg2SeqHdrGetLumNonIQuantMatrix(h) (h)->lum_niqt
#define Mpeg2SeqHdrGetChrIQuantMatrix(h) (h)->chr_iqt
#define Mpeg2SeqHdrGetChrNonIQuantMatrix(h) (h)->chr_niqt

  /* Routines/macros dealing with display extension and its fields */
  int Mpeg2SeqHdrDisplayExtParse(BitParser *bp, Mpeg2SeqHdr *hdr);
  int Mpeg2SeqHdrGetDisplayWidth(Mpeg2SeqHdr *hdr);
  int Mpeg2SeqHdrGetDisplayHeight(Mpeg2SeqHdr *hdr);
  
#define Mpeg2SeqHdrDisplayExtPresent(h)  (h)->display_ext_flag
#define Mpeg2SeqHdrGetVideoFormat(h) (h)->video_format_code
#define Mpeg2SeqHdrGetColorPrimary(h) (h)->color_primary_code
#define Mpeg2SeqHdrGetColorXferCharacteristic(h) (h)->color_xfer_characteristic_code
#define Mpeg2SeqHdrGetColorMatrixCoeff(h) (h)->color_matrix_coeff_code
#define Mpeg2SeqHdrScalableExtPresent(h) (h)->scalable_ext_flag
#define Mpeg2SeqHdrGetChromaFormatCode(h) (h)->chroma_format_code

  void Mpeg2SeqEndCodeEncode(BitParser *);

  /*
   * Picture header functions
   */

  Mpeg2PicHdr *Mpeg2PicHdrNew();
  void Mpeg2PicHdrFree(Mpeg2PicHdr *hdr);
  int Mpeg2PicHdrFind(BitParser *bp);
  int Mpeg2PicHdrParse(BitParser *bp, Mpeg2PicHdr *hdr);
  int Mpeg2PicHdrExtFind(BitParser *bp);
  int Mpeg2PicHdrExtParse(BitParser *bp, Mpeg2PicHdr *hdr);
#define Mpeg2PicHdrGetPictureType(h) (h)->picture_type
#define Mpeg2PicHdrGetTopFieldFirstFlag(h) (h)->top_field_first_flag
#define Mpeg2PicHdrGetStructureCode(h) (h)->picture_structure_code
#define Mpeg2PicHdrGetTemporalReference(h) (h)->temporal_reference
#define Mpeg2PicHdrGetConcealmentMVFlag(h) (h)->concealment_mv_flag
#define Mpeg2PicHdrGetFCode(h,i,j) ((h)->fcode[i][j])
#define Mpeg2PicHdrResetPMV(h) \
          ((h)->PMV[0][0]).down = 0; \
          ((h)->PMV[0][0]).right = 0; \
          ((h)->PMV[0][1]).down = 0; \
          ((h)->PMV[0][1]).right = 0; \
          ((h)->PMV[1][0]).down = 0; \
          ((h)->PMV[1][0]).right = 0; \
          ((h)->PMV[1][1]).down = 0; \
          ((h)->PMV[1][1]).right = 0
#define Mpeg2PicHdrUpdatePMV1Fwd(h) \
          ((h)->PMV[1][0]).down = ((h)->PMV[0][0]).down; \
          ((h)->PMV[1][0]).right = ((h)->PMV[0][0]).right
#define Mpeg2PicHdrUpdatePMV1Bwd(h) \
          ((h)->PMV[1][1]).down= ((h)->PMV[0][1]).down; \
          ((h)->PMV[1][1]).right = ((h)->PMV[0][1]).right

#define Mpeg2PicHdrGetPMVHorizComp(h, r, s) ((h)->PMV[r][s]).right
#define Mpeg2PicHdrGetPMVVertComp(h, r, s) ((h)->PMV[r][s]).down
#define Mpeg2PicHdrSetPMVHorizComp(h, r, s, v) ((h)->PMV[r][s]).right = v
#define Mpeg2PicHdrSetPMVVertComp(h, r, s, v) ((h)->PMV[r][s]).down = v
#define Mpeg2PicHdrGetIntraVLCFormat(h) (h)->intra_vlc_format_code
#define Mpeg2PicHdrGetAltScan(h) (h)->alternate_scan_flag
#define Mpeg2PicHdrGetQScaleType(h) (h)->q_scale_type
#define Mpeg2PicHdrGetIntraDCPrecision(h) (h)->intra_dc_precision
#define Mpeg2PicHdrGetFramePredFrameDCT(h) (h)->frame_pred_frame_dct_flag

  /*
   * MB Info functions
   */

  Mpeg2MBInfoArray *Mpeg2MBInfoArrayNew(int w, int h);
  void Mpeg2MBInfoArrayFree(Mpeg2MBInfoArray *mbi_array);
#define Mpeg2MBInfoArrayGetMBInfoEntry(a,i) (&((a)->mbinfo[i]))
#define Mpeg2MBInfoArrayGetWidth(a) (a)->width
#define Mpeg2MBInfoArrayGetHeight(a) (a)->height
#define Mpeg2MBInfoSetQuant(i,v) (i)->quant = v
#define Mpeg2MBInfoGetQuant(i) (i)->quant
#define Mpeg2MBInfoSetMFwd(i,v) (i)->motion_forward = v
#define Mpeg2MBInfoGetMFwd(i) (i)->motion_forward
#define Mpeg2MBInfoSetMBwd(i,v) (i)->motion_backward = v
#define Mpeg2MBInfoGetMBwd(i) (i)->motion_backward
#define Mpeg2MBInfoSetPattern(i,v) (i)->pattern = v
#define Mpeg2MBInfoGetPattern(i) (i)->pattern
#define Mpeg2MBInfoSetIntra(i,v) (i)->intra = v
#define Mpeg2MBInfoSetMotionType(i,v) (i)->motion_type = v
#define Mpeg2MBInfoGetMotionType(i) (i)->motion_type
#define Mpeg2MBInfoSetDCTType(i,v) (i)->dct_type = v
#define Mpeg2MBInfoGetDCTType(i) (i)->dct_type
#define Mpeg2MBInfoSetMVFieldSelect(i,j,k,v) (i)->mv_field_select[j][k] = v
#define Mpeg2MBInfoGetMVFieldSelect(i,j,k) (i)->mv_field_select[j][k]
#define Mpeg2MBInfoSetMVectorHorizComp(i,r,s,v) ((i)->mvectors[r][s]).right = v
#define Mpeg2MBInfoSetMVectorVertComp(i,r,s,v) ((i)->mvectors[r][s]).down = v
#define Mpeg2MBInfoGetMVectorHorizComp(i,r,s) ((i)->mvectors[r][s]).right
#define Mpeg2MBInfoGetMVectorVertComp(i,r,s) ((i)->mvectors[r][s]).down
#define Mpeg2MBInfoSetMVCount(i,v) (i)->mv_count = v
#define Mpeg2MBInfoSetMVFormat(i,v) (i)->mv_format = v
#define Mpeg2MBInfoSetDMVFlag(i,v) (i)->dmv_flag = v
#define Mpeg2MBInfoGetMVCount(i) (i)->mv_count
#define Mpeg2MBInfoGetMVFormat(i) (i)->mv_format
#define Mpeg2MBInfoGetDMVFlag(i) (i)->dmv_flag
#define Mpeg2MBInfoGetIntra(i) (i)->intra
#define Mpeg2MBInfoGetPatternFlag(i) (i)->pattern

  /* Util prototypes */

  int Mpeg2AnyStartCodeFind(BitParser *bp);
  unsigned int Mpeg2GetCurrStartCode(BitParser *bp);
  unsigned int Mpeg2GetCurrExtensionCode(BitParser *bp);
  int Mpeg2SkipToNextStartCode(BitParser *bp);

  /* Parse prototypes */

  int Mpeg2ParseI(BitParser *bp,  Mpeg2SeqHdr *sh, Mpeg2PicHdr *fh, ScImage *y, ScImage *u, ScImage *v, Mpeg2MBInfoArray *mbi_array, int field_flag);
  int Mpeg2ScIFieldToByte(ScImage *sc_image, ByteImage *dest_image);
  int Mpeg2ScIFrameToByte(ScImage *sc_image, Mpeg2MBInfoArray *mbi_array, ByteImage *dest_image, int lum_flag, int non_420_flag);
  int Mpeg2ParseP(BitParser *bp, Mpeg2SeqHdr *sh, Mpeg2PicHdr *fh, ScImage *y, ScImage *u, ScImage *v, Mpeg2MBInfoArray *mbi_array, int field_flag);
  int Mpeg2ScPFieldToByte(ScImage *sc_image, Mpeg2MBInfoArray *mbi_array, ByteImage *prev_top, ByteImage *prev_bot, ByteImage *dest_image);
  int Mpeg2ParseB(BitParser *bp, Mpeg2SeqHdr *sh, Mpeg2PicHdr *fh, ScImage *y, ScImage *u, ScImage *v, Mpeg2MBInfoArray *mbi_array, int field_flag);
  int Mpeg2ScBFrameToByte(ScImage *sc_image, Mpeg2MBInfoArray *mbi_array, ByteImage *prev, ByteImage *future, ByteImage *dest_image, int lum_flag, int non_420_flag);
  int Mpeg2ScBFieldToByte(ScImage *sc_image, Mpeg2MBInfoArray *mbi_array, ByteImage *prev_top, ByteImage *prev_bot, ByteImage *future_top, ByteImage *future_bot, ByteImage *dest_image);
  int Mpeg2ScPFrameToByte(ScImage *sc_image, Mpeg2MBInfoArray *mbi_array, ByteImage *prev, ByteImage *dest_image, int lum_flag, int non_420_flag);


  /* Reference index structures and interface declaration below. */

  typedef struct {
    unsigned int offset;
    unsigned int size;
    unsigned int ref1;
    unsigned int ref2;
    int type_flag;
  } Mpeg2RefIdxRecord;
  
  typedef struct {
    int num_recs;
    unsigned int max_size;
    Mpeg2RefIdxRecord *recs;
  } Mpeg2RefIdx;
  
#define Mpeg2RefIdxRecordGetOffset(rec) (rec)->offset
#define Mpeg2RefIdxRecordGetSize(rec) (rec)->size
#define Mpeg2RefIdxRecordGetRef1(rec) (rec)->ref1
#define Mpeg2RefIdxRecordGetRef2(rec) (rec)->ref2
#define Mpeg2RefIdxRecordGetType(rec) (rec)->type_flag

  Mpeg2RefIdx *Mpeg2RefIdxNew();
  void Mpeg2RefIdxFree(Mpeg2RefIdx *idx);
  int Mpeg2ReadRefIndex(Mpeg2RefIdx *idx, char *idx_filename);
  Mpeg2RefIdxRecord *Mpeg2RefIdxGetRecord(Mpeg2RefIdx *idx, unsigned int frame_num);
#define Mpeg2RefIdxGetMaxSize(idx) (idx)->max_size

  typedef struct Mpeg2CacheEntry {
    unsigned int frame_num;
    ByteImage *y;
    ByteImage *u;
    ByteImage *v;
    struct Mpeg2CacheEntry *next;
  } Mpeg2CacheEntry;

  typedef struct {
    int num_entries;
    Mpeg2CacheEntry *free_entries;
    Mpeg2CacheEntry *cur_entries;
  } Mpeg2Cache;

  Mpeg2Cache *Mpeg2CacheNew(int num_entries);
  void Mpeg2CacheFree(Mpeg2Cache *cache);
  void Mpeg2CachePutEntry(Mpeg2Cache *cache, unsigned int framenum, ByteImage *y, ByteImage *u, ByteImage *v);
  int Mpeg2CacheGetEntry(Mpeg2Cache *cache, unsigned int framenum, ByteImage **y, ByteImage **u, ByteImage **v);


  typedef struct {
    Mpeg2Cache *frame_cache;
    Mpeg2RefIdx *frame_index;
    Mpeg2SeqHdr *sh;
    Mpeg2PicHdr *fh;
    BitParser *bp;
    BitStream *bs;
    FILE *data_fp;
    ScImage *scy;
    ScImage *scv;
    ScImage *scu;
    Mpeg2MBInfoArray *mbi_array;
  } Mpeg2CellMgr;

  Mpeg2CellMgr *Mpeg2CellMgrNew(char *data_file, char *index_file);
  void Mpeg2CellMgrFree(Mpeg2CellMgr *cell_mgr);
  int Mpeg2CellMgrGetWidth(Mpeg2CellMgr *cell_mgr);
  int Mpeg2CellMgrGetHeight(Mpeg2CellMgr *cell_mgr);
  int Mpeg2CellMgrGetFrameAsPlanes(Mpeg2CellMgr *cell_mgr, unsigned int cell_num, unsigned int face_num, ByteImage *r, ByteImage *g, ByteImage *b);

  /* The following is what Andy will eventually need. 
  int Mpeg2CellMgrGetFrame(Mpeg2CellMgr *cell_mgr, unsigned int cell_num, unsigned int frame_num, unsigned char *framebuf);
  */

#ifdef __cplusplus
}           /* extern "C" */
#endif      /* __cplusplus */

#endif      /* __DVM_MPEG_2_H */








