/*-------------------------------------
 * 
 * Copyright (c) 2000 by The University of North Carolina
 *
 * See the file "packages/mpeg2/nc_license.txt" for information 
 * on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *-------------------------------------
 */

#include "mpeg2Int.h"

/*-----------------------------------------
 * Mpeg2PicHdrNew
 *
 * Allocates a new Mpeg2PicHdr structure.
 *
 *-----------------------------------------*/

Mpeg2PicHdr *
Mpeg2PicHdrNew()
{
  Mpeg2PicHdr *hdr;
  hdr = NEW(Mpeg2PicHdr);
  
  return hdr;
}

/*-----------------------------------------
 * Mpeg2PicHdrFree
 *
 * Frees a previously allocated Mpeg2PicHdr
 * structure.
 *
 *-----------------------------------------*/

void
Mpeg2PicHdrFree(Mpeg2PicHdr *hdr)
{
  FREE((char *) hdr);
}

/*-----------------------------------------
 * Mpeg2PicHdrFind
 *
 * Given BitParser bp, finds the next picture
 * header start code. If start code is found,
 * bp is positioned at the start code and the
 * total number of bytes skipped is returned.
 * If start code is not found, returns
 * DVM_MPEG2_NOT_FOUND.
 *
 *-----------------------------------------*/

int
Mpeg2PicHdrFind(BitParser *bp)
{
  unsigned int code, off, total;

  total = 0;
  Bp_InByteAlign(bp);
  while (((code = NextStartCode(bp, &off)) != MPEG2_SEQ_END_CODE) &&
	 (code != 0)) {
    total += off;
    if (code == MPEG2_PIC_START_CODE) {
      Bp_RestoreInt(bp);
      return total - 4;
    }
  }
  return DVM_MPEG2_NOT_FOUND;
}

/*-----------------------------------------
 * Mpeg2PicHdrParse
 *
 * Given BitParser bp and Mpeg2PicHdr hdr,
 * parses the picture header and fills the
 * appropriate fields of hdr. If bp is not
 * positioned at the start of a picture header,
 * returns DVM_MPEG2_INVALID_START_CODE. Otherwise,
 * returns total number of bytes parsed.
 *
 *-----------------------------------------*/

int
Mpeg2PicHdrParse(BitParser *bp, Mpeg2PicHdr *hdr)
{
  unsigned int data;
  int total_bits = 0;

  /* Flush off picture start code. */
  Bp_InByteAlign(bp);
  Bp_GetInt(bp, data);
  if (data != MPEG2_PIC_START_CODE) {
    Bp_RestoreInt(bp);
    return DVM_MPEG2_INVALID_START_CODE;
  }

  total_bits += 32;

  /* Get temporal reference */

  Bp_GetBits(bp, 10, data);
  hdr->temporal_reference = data;

  /* Frame type (I,P, or B) */
  Bp_GetBits(bp, 3, data);
  hdr->picture_type = data;

  /* VBV delay parameter */
  Bp_GetBits(bp, 16, data);
  hdr->vbv_delay = data;

  total_bits += 29;

  /* Flush out MPEG-1 vectors and fcodes not used by MPEG-2 */

  if (hdr->picture_type != I_FRAME) {
    Bp_FlushBits(bp, 4);

    total_bits += 4;

    if (hdr->picture_type == B_FRAME) {
      Bp_FlushBits(bp, 4);

      total_bits += 4;
    }
  }

  /* Flush out extra picture info. */
  Bp_GetBits(bp, 1, data);
  while(data) {
    Bp_FlushBits(bp, 8);
    Bp_GetBits(bp, 1, data);
    total_bits += 9;
  }

  total_bits += 1;

  /* Byte align */
  Bp_FlushBits(bp, (8 - (total_bits % 8)));
  total_bits += (8 - (total_bits % 8));

  Bp_InByteAlign(bp);

  /* Return number of bytes parsed. */
  return (total_bits / 8);
}  

/*-----------------------------------------
 * Mpeg2PicHdrExtFind
 *
 * Given BitParser bp, finds next Mpeg-2 picture
 * coding extension header. If found, bp is
 * positioned at the header start code and total
 * number of bytes parsed is returned. If not
 * found, returns DVM_MPEG2_NOT_FOUND.
 *
 *-----------------------------------------*/

int
Mpeg2PicHdrExtFind(BitParser *bp)
{
  unsigned int code, off, total;

  Bp_InByteAlign(bp);
  total = 0;
  while (((code = NextStartCode(bp, &off)) != MPEG2_SEQ_END_CODE) &&
	 (code != 0)) {
    total += off;
    if (code == MPEG2_EXT_START_CODE) {
      Bp_GetBits(bp, 4, code);
      if (code == MPEG2_PIC_CODING_EXT_ID) {
	Bp_RestoreBits(bp, 4);
	Bp_InByteAlign(bp);
	Bp_RestoreInt(bp);
	return total - 4;
      } else {
	Bp_RestoreBits(bp, 4);
	Bp_InByteAlign(bp);
      }
    }
  }
  return DVM_MPEG2_NOT_FOUND;
}

/*-----------------------------------------
 * Mpeg2PicHdrExtParse
 *
 * Given BitParser bp and Mpeg2PicHdr hdr,
 * parses picture coding extension header and
 * fill appropriate fields of hdr. If successful,
 * total number of bytes parsed is returned.
 * If bp is not positioned at an extenstion header, 
 * DVM_MPEG2_INVALID_START_CODE is returned. If
 * extension header is not a picture coding
 * extension, DVM_MPEG2_EXTENSION_ID_MISMATCH is
 * returned.
 *-----------------------------------------*/

int
Mpeg2PicHdrExtParse(BitParser *bp, Mpeg2PicHdr *hdr)
{
  unsigned int data;
  int i,j;
  int total_bits = 0;

  Bp_InByteAlign(bp);

  /* Flush off seq. ext. code */
  Bp_GetInt(bp, data);
  if (data != MPEG2_EXT_START_CODE) {
    Bp_RestoreInt(bp);
    return DVM_MPEG2_INVALID_START_CODE;
  }

  /* Get extension id. */
  Bp_GetBits(bp, 4, data);
  if (data != MPEG2_PIC_CODING_EXT_ID) {
    Bp_RestoreBits(bp, 4);
    Bp_InByteAlign(bp);
    Bp_RestoreInt(bp);
    return DVM_MPEG2_EXTENSION_ID_MISMATCH;
  }

  /* Get fcodes (these determine range of motion vectors */
  for(i=0; i<2; i++) {
    for(j=0; j<2; j++) {
      Bp_GetBits(bp, 4, data);
      hdr->fcode[i][j] = data;
    }
  }

  Bp_GetBits(bp, 2, data);
  hdr->intra_dc_precision = data;

  Bp_GetBits(bp, 2, data);
  hdr->picture_structure_code = data;

  Bp_GetBits(bp, 1, data);
  hdr->top_field_first_flag = data;

  Bp_GetBits(bp, 1, data);
  hdr->frame_pred_frame_dct_flag = data;

  Bp_GetBits(bp, 1, data);
  hdr->concealment_mv_flag = data;

  Bp_GetBits(bp, 1, data);
  hdr->q_scale_type = data;

  Bp_GetBits(bp, 1, data);
  hdr->intra_vlc_format_code = data;

  Bp_GetBits(bp, 1, data);
  hdr->alternate_scan_flag = data;
  
  Bp_GetBits(bp, 1, data);
  hdr->repeat_first_field_flag = data;

  Bp_GetBits(bp, 1, data);
  hdr->chroma_420_type = data;

  Bp_GetBits(bp, 1, data);
  hdr->progressive_frame_flag = data;

  Bp_GetBits(bp, 1, data);
  hdr->composite_display_flag = data;

  total_bits = 82;

  if (hdr->composite_display_flag) {
    Bp_GetBits(bp, 1, data);
    hdr->v_axis = data;

    Bp_GetBits(bp, 3, data);
    hdr->field_sequence = data;

    Bp_GetBits(bp, 1, data);
    hdr->sub_carrier = data;

    Bp_GetBits(bp, 7, data);
    hdr->burst_amplitude = data;

    Bp_GetBits(bp, 8, data);
    hdr->sub_carrier_phase = data;

    total_bits += 20;
  }
  
  /* Byte align */
  Bp_FlushBits(bp, (8 - (total_bits % 8)));
  total_bits += (8 - (total_bits % 8));
  
  Bp_InByteAlign(bp);
  
  /* Return number of byts parsed. */
  return (total_bits / 8);
}  


  

