/*-------------------------------------
 * 
 * Copyright (c) 2000 by The University of North Carolina
 *
 * See the file "packages/mpeg2/nc_license.txt" for information 
 * on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *-------------------------------------
 */

#include "mpeg2Int.h"

/*-----------------------------------------
 * Mpeg2SeqHdrNew
 *
 * Allocates a new Mpeg2SeqHdr structure.
 * Initial default values for some fields are
 * filled in. Pointer to new structure is 
 * returned.
 *-----------------------------------------*/

Mpeg2SeqHdr *
Mpeg2SeqHdrNew()
{
  Mpeg2SeqHdr *hdr;
  int i;
  static unsigned char defaultIQT[64] =
  {
    8, 16, 19, 22, 26, 27, 29, 34,
    16, 16, 22, 24, 27, 29, 34, 37,
    19, 22, 26, 27, 29, 34, 34, 38,
    22, 22, 26, 27, 29, 34, 37, 40,
    22, 26, 27, 29, 32, 35, 40, 48,
    26, 27, 29, 32, 35, 40, 48, 58,
    26, 27, 29, 34, 38, 46, 56, 69,
    27, 29, 35, 38, 46, 56, 69, 83};

  hdr = NEW(Mpeg2SeqHdr);
  hdr->display_ext_flag = 0;
  hdr->scalable_ext_flag = 0;

  /* Default values for fields within display extension which
   * may never be parsed (i.e., doesn't exist)
   */
  hdr->video_format_code = VIDEO_FORMAT_UNSPECIFIED;
  hdr->color_primary_code = 1;
  hdr->color_xfer_characteristic_code = 1;
  hdr->color_matrix_coeff_code = 1;

  /* Init quant matrices with default values. */
  for (i = 0; i < 64; i++) {
    hdr->lum_iqt[i] = hdr->chr_iqt[i] = defaultIQT[i];
  }

  for (i = 0; i < 64; i++) {
    hdr->lum_niqt[i] = hdr->chr_niqt[i] = 16;
  }

  return hdr;
}

/*-----------------------------------------
 * Mpeg2SeqHdrFree
 *
 * Frees a previously allocated Mpeg2SeqHdr
 * structure.
 *
 *-----------------------------------------*/

void
Mpeg2SeqHdrFree(Mpeg2SeqHdr *hdr)
{
  FREE((char *) hdr);
}

int
Mpeg2SeqHdrFind(BitParser *bp)
{
  unsigned int code, off, total;

  total = 0;
  Bp_InByteAlign(bp);
  while (((code = NextStartCode(bp, &off)) != MPEG2_SEQ_END_CODE) &&
	 (code != 0)) {
    total += off;
    if (code == MPEG2_SEQ_START_CODE) {
      Bp_RestoreInt(bp);
      return total - 4;
    }
  }
  return DVM_MPEG2_NOT_FOUND;
}

/*-----------------------------------------
 * Mpeg2SeqHdrParse
 *
 * Given BitParser bp and Mpeg2SeqHdr hdr,
 * parses sequence header and fills in appropriate
 * fields of hdr. If bp is not positioned at
 * a sequence header start code, returns
 * DVM_MPEG2_INVALID_START_CODE. Otherwise,
 * returns number of bytes parsed.
 *
 *-----------------------------------------*/

int
Mpeg2SeqHdrParse(BitParser *bp, Mpeg2SeqHdr *hdr)
{
  unsigned int data, r1, r2;
  int total, i;


  /* Flush off sequence start code. */
  
  Bp_InByteAlign(bp);
  Bp_GetInt(bp, data);
  if (data != MPEG2_SEQ_START_CODE) {
    Bp_RestoreInt(bp);
    return DVM_MPEG2_INVALID_START_CODE;
  }
  
  /* Get lower 12 bits of horizontal size of image space. */

  Bp_GetBits(bp, 12, data);
  hdr->horizontal_size = data;

  /* Get lower 12 bits vertical size of image space. */

  Bp_GetBits(bp, 12, data);
  hdr->vertical_size = data;

  /* Calculate macroblock width and height of image space.
     Not, mb_height will be recalculated when 
     sequence header extension is read.
  */

  hdr->mb_width = (hdr->horizontal_size + 15) / 16;
  hdr->mb_height = (hdr->vertical_size + 15) / 16;

  /* Parse of pel_aspect_ratio ratio code. */

  Bp_GetBits(bp, 4, data);
  hdr->aspect_ratio_code = data;

  /* Parse off frame rate code. */

  Bp_GetBits(bp, 4, data);
  hdr->frame_rate_code = data;

  /* Parse off lower 18 bits of bit rate. */

  Bp_GetBits(bp, 8, r1);
  Bp_GetBits(bp, 8, r2);
  Bp_GetBits(bp, 2, data);
  data = (r1 << 10) | (r2 << 2) | data;
  hdr->bitrate = data;

  /* Flush marker bit. */

  Bp_FlushBits(bp, 1);

  /*
   * Parse off lower 10 bits of vbv buffer size. 
   */

  Bp_GetBits(bp, 10, data);
  hdr->vbv_buffer_size = data;

  /* Parse off contrained parameter flag. 
     Completely unused in MPEG-2
  */

  Bp_FlushBits(bp, 1);

  /*
   * Total number of bytes so far (with the 2 bits about to be used
   * for quant matrix flags).
   */
  total = 12;

  /*
   * If intra_quant_matrix_flag set, parse off intra quant matrix values.
   */
  Bp_GetBits(bp, 1, data);
  if (data) {
    for (i = 0; i < 64; i++) {
      Bp_GetBits(bp, 8, data);
      hdr->lum_iqt[i] = hdr->chr_iqt[i] = (unsigned char) data;
    }
    total += 64;
  } 

  /*
   * If non intra quant matrix flag set, parse off non intra quant matrix
   * values.
   */

  Bp_GetBits(bp, 1, data);
  if (data) {
    for (i = 0; i < 64; i++) {
      Bp_GetBits(bp, 8, data);
      hdr->lum_niqt[i] = hdr->chr_niqt[i] = (unsigned char) data;
    }
    total += 64;
  } 

  /* Restore byte mode. Should be byte aligned */
  Bp_InByteAlign(bp);
  return total;
}

/*-----------------------------------------
 * Mpeg2SeqHdrExtFind
 *
 * Given BitParser bp, finds next sequence
 * extension header. If successful, returns
 * number of bytes parsed and bp is positioned
 * at start of sequence extension header.
 * If not found, DVM_MPEG2_NOT_FOUND returned.
 *
 *-----------------------------------------*/

int
Mpeg2SeqHdrExtFind(BitParser *bp)
{
  unsigned int code, off, total;

  Bp_InByteAlign(bp);
  total = 0;
  while (((code = NextStartCode(bp, &off)) != MPEG2_SEQ_END_CODE) &&
	 (code != 0)) {
    total += off;
    if (code == MPEG2_EXT_START_CODE) {
      Bp_GetBits(bp, 4, code);
      if (code == MPEG2_SEQ_HDR_EXT_ID) {
	Bp_RestoreBits(bp, 4);
	Bp_InByteAlign(bp);
	Bp_RestoreInt(bp);
	return total - 4;
      } else {
	Bp_RestoreBits(bp, 4);
	Bp_InByteAlign(bp);
      }
    }
  }
  return DVM_MPEG2_NOT_FOUND;
}

/*-----------------------------------------
 * Mpeg2SeqHdrExtParse
 *
 * Given BitParser bp and Mpeg2SeqHdr hdr,
 * parses sequence header extension and fills
 * appropriate fields of hdr. If successful,
 * total number of bytes parsed returned.
 * If bp is not positioned at an extension
 * header, DVM_MPEG2_INVALID_START_CODE returned.
 * If extension header is not a sequence extension
 * header, DVM_MPEG2_EXTENSION_ID_MISMATCH
 * returned.
 *
 *-----------------------------------------*/

int
Mpeg2SeqHdrExtParse(BitParser *bp, Mpeg2SeqHdr *hdr)
{
  unsigned int data;

  Bp_InByteAlign(bp);
  /* Flush off sequence ext code. */
  Bp_GetInt(bp, data);
  if (data != MPEG2_EXT_START_CODE) {
    Bp_RestoreInt(bp);
    return DVM_MPEG2_INVALID_START_CODE;
  }

  /* Get extension id. */
  Bp_GetBits(bp, 4, data);
  if (data != MPEG2_SEQ_HDR_EXT_ID) {
    Bp_RestoreBits(bp, 4);
    Bp_InByteAlign(bp);
    Bp_RestoreInt(bp);
    return DVM_MPEG2_EXTENSION_ID_MISMATCH;
  }
  
  /* Get profile/level code. */
  Bp_GetBits(bp, 8, data);
  
  hdr->profile_code = data;

  /* Get progressive sequence flag. */
  Bp_GetBits(bp, 1, data);
  
  hdr->progressive_seq_flag = data;

  /* Get chroma format code. */
  Bp_GetBits(bp, 2, data);
  
  hdr->chroma_format_code = data;

  /* Get upper 2 bits of horizontal size */
  Bp_GetBits(bp, 2, data);
  
  hdr->horizontal_size = hdr->horizontal_size | (data << 12);
  
  /* Get upper 2 bits of vertical size */
  Bp_GetBits(bp, 2, data);
  
  hdr->vertical_size = hdr->vertical_size | (data << 12);

  /* Recaluclate mb_width and mb_height */
  hdr->mb_width = (hdr->horizontal_size + 15) / 16;
  if (hdr->progressive_seq_flag == 1) {
    hdr->mb_height = (hdr->vertical_size + 15) / 16;
  } else {
    hdr->mb_height = ((hdr->vertical_size + 31) / 32) * 2;
  }

  /* Get upper 12 bits of bit rate */
  Bp_GetBits(bp, 12, data);

  hdr->bitrate = hdr->bitrate | (data << 18);
  
  /* Parse off marker bit. */
  Bp_GetBits(bp, 1, data);

  /* Get upper 8 bits of vbv buffer size. */
  Bp_GetBits(bp, 8, data);

  hdr->vbv_buffer_size = hdr->vbv_buffer_size | (data << 10);

  /* Get low delay flag */
  Bp_GetBits(bp, 1, data);

  hdr->low_delay_flag = data;

  /* Get frame rate extension factor "n" */
  Bp_GetBits(bp, 2, data);

  hdr->frame_rate_ext_n = data;
  
  /* Get frame rate extension factor "d" */
  Bp_GetBits(bp, 5, data);

  hdr->frame_rate_ext_d = data;

  Bp_InByteAlign(bp);
  /* Return number of bytes parsed. */
  return 10;
}

/*-----------------------------------------
 * Mpeg2SeqHdrGetAspectRatio
 *
 * Given Mpeg2SeqHdr *hdr, returns aspect
 * ratio as a double float number (i.e., 3x4
 * aspect ratio returned as 0.75). Aspect ratio
 * information is calculated from encoded width,
 * height, and display information as per 
 * section 6.3.3 in the standard.
 *
 *-----------------------------------------*/

double
Mpeg2SeqHdrGetAspectRatio(Mpeg2SeqHdr *hdr)
{
  double scale_factor;
  double w, h;

  if (hdr->aspect_ratio_code == ASPECT_SQUARE) {
    return 1.0;
  }

  switch (hdr->aspect_ratio_code) {
  case ASPECT_DAR_3_by_4:
    scale_factor = 3.0/4.0;
    break;
  case ASPECT_DAR_9_by_16:
    scale_factor = 9.0/16.0;
    break;
  case ASPECT_DAR_1_by_2_21:
    scale_factor = 1.0/2.21;
    break;
  default:
    return DVM_MPEG2_ERROR;
  }
    
  if (hdr->display_ext_flag) {
    /* Calculate aspect ratio using parameters in display
       extension. XXX -- need to add this.
    */
    
    return 1.0;
  } else {
    w = (double) hdr->horizontal_size;
    h = (double) hdr->vertical_size;
  }
  return (scale_factor * (w / h));
}

/*-----------------------------------------
 * Mpeg2SeqHdrGetChromaFormat
 *
 * Given Mpeg2SeqHdr *hdr, returns format of
 * chrominance planes as an integer.
 * The return value 420 indicates 4:2:0 format,
 * the return value 422 indicates 4:2:2 and 
 * the return value 444 indicates 4:4:4.
 *
 *-----------------------------------------*/

int 
Mpeg2SeqHdrGetChromaFormat(Mpeg2SeqHdr *hdr)
{
  switch (hdr->chroma_format_code) {
  case CHROMA_CODE_4_2_0:
    return 420;
  case CHROMA_CODE_4_2_2:
    return 422;
  case CHROMA_CODE_4_4_4:
    return 444;
  }
  return DVM_MPEG2_ERROR;
}

/*-----------------------------------------
 * Mpeg2SeqHdrGetFrameRate
 *
 * Given Mpeg2SeqHdr *hdr, returns frame
 * rate information as a double float integer.
 *
 *-----------------------------------------*/

double 
Mpeg2SeqHdrGetFrameRate(Mpeg2SeqHdr *hdr)
{
  double base_rate;
  double n, d;

  switch(hdr->frame_rate_code) {
  case FRAME_RATE_24000_div_1001:
    base_rate = 24000.0 / 1001.0;
    break;
  case FRAME_RATE_24:
    base_rate = 24.0;
    break;
  case FRAME_RATE_25:
    base_rate = 25.0;
    break;
  case FRAME_RATE_30000_div_1001:
    base_rate = 30000.0 / 1001.0;
    break;
  case FRAME_RATE_30:
    base_rate = 30.0;
    break;
  case FRAME_RATE_50:
    base_rate = 50.0;
    break;
  case FRAME_RATE_60000_div_1001:
    base_rate = 60000.0 / 1001.0;
    break;
  case FRAME_RATE_60:
    base_rate = 60.0;
    break;
  default:
    return DVM_MPEG2_ERROR;
  }

  n = ((double) hdr->frame_rate_ext_n) + 1.0;
  d = ((double) hdr->frame_rate_ext_d) + 1.0;

  return base_rate * (n / d);
}

/*-----------------------------------------
 * Mpeg2SeqHdrDisplayExtParse
 *
 * Given BitParser *bp and Mpeg2SeqHdr *hdr,
 * parses sequence display extension header,
 * filling in appropriate fields of hdr.
 * If successful, total number of bytes parsed 
 * is returned. If bp is not positioned at the
 * beginning of an extenstion header, 
 * DVM_MPEG2_INVALID_START_CODE returned. If
 * extension header is not a sequence display
 * extenstion header, DVM_MPEG2_EXTENSION_ID_MISMATCH
 * returned.
 *
 *-----------------------------------------*/

int
Mpeg2SeqHdrDisplayExtParse(BitParser *bp, Mpeg2SeqHdr *hdr)
{
  unsigned int data;
  int total;

  total = 0;

  Bp_InByteAlign(bp);
  
  /* Flush off sequence ext code. */
  Bp_GetInt(bp, data);
  if (data != MPEG2_EXT_START_CODE) {
    Bp_RestoreInt(bp);
    return DVM_MPEG2_INVALID_START_CODE;
  }

  /* Get extension id. */
  Bp_GetBits(bp, 4, data);
  if (data != MPEG2_SEQ_DISPLAY_EXT_ID) {
    Bp_RestoreBits(bp, 4);
    Bp_InByteAlign(bp);
    Bp_RestoreInt(bp);
    return DVM_MPEG2_EXTENSION_ID_MISMATCH;
  }

  /* Get video format code */
  Bp_GetBits(bp, 3, data);
  hdr->video_format_code = data;
  
  /* Check for color info */
  Bp_GetBits(bp, 1, data);
  if (data) {
    /* Parse off color info codes */
    Bp_GetBits(bp, 8, data);
    hdr->color_primary_code = data;

    Bp_GetBits(bp, 8, data);
    hdr->color_xfer_characteristic_code = data;

    Bp_GetBits(bp, 8, data);
    hdr->color_matrix_coeff_code = data;
    total += 3;
  }

  /* Display width */
  Bp_GetBits(bp, 14, data);
  hdr->display_horizontal_size = data;

  /* Parse off marker bit */
  Bp_FlushBits(bp, 1);

  /* Display height */
  Bp_GetBits(bp, 14, data);
  hdr->display_vertical_size = data;

  /* Set display ext_flag*/
  hdr->display_ext_flag = 1;

  /* Parse of 3 remaining extra bits 
     to byte align
  */
  Bp_FlushBits(bp, 3);
  Bp_InByteAlign(bp);

  return (total + 9);
}

/*-----------------------------------------
 * Mpeg2SeqHdrGetDisplayWidth
 *
 * Given Mpeg2SeqHdr *hdr, returns value of
 * display_horizontal_size field. This
 * field is only defined if an optional sequence
 * display extension header is present and
 * has been parsed. If the field is not defined,
 * DVM_MPEG2_ERROR is returned.
 *-----------------------------------------*/

int 
Mpeg2SeqHdrGetDisplayWidth(Mpeg2SeqHdr *hdr)
{
  if (hdr->display_ext_flag) {
    return hdr->display_horizontal_size;
  }
  return DVM_MPEG2_ERROR;
}

/*-----------------------------------------
 * Mpeg2SeqHdrGetDisplayHeight
 *
 * Given Mpeg2SeqHdr *hdr, returns value of
 * display_vertical_size field. This
 * field is only defined if an optional sequence
 * display extension header is present and
 * has been parsed. If the field is not defined,
 * DVM_MPEG2_ERROR is returned.
 *-----------------------------------------*/

int Mpeg2SeqHdrGetDisplayHeight(Mpeg2SeqHdr *hdr)
{
  if (hdr->display_ext_flag) {
    return hdr->display_vertical_size;
  }
  return DVM_MPEG2_ERROR;
}


