
/*-------------------------------------
 * 
 * Copyright (c) 2000 by The University of North Carolina
 *
 * See the file "packages/mpeg2/nc_license.txt" for information 
 * on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 *-------------------------------------
 */

#include "mpeg2Int.h"

/*-----------------------------------------
 * Mpeg2AnyStartCodeFind
 *
 * Given BitParser *bp, parses until next
 * start code found. Bp is positioned at the
 * next start code and then number of bytes 
 * parsed is returned. If no start code is found,
 * DVM_MPEG2_NOT_FOUND is returned.
 *-----------------------------------------*/

int
Mpeg2AnyStartCodeFind(BitParser *bp)
{
  unsigned int code, off;

  Bp_InByteAlign(bp);
  code = NextStartCode(bp, &off);
  if (code != 0) {
    Bp_RestoreInt(bp);
    return off - 4;
  }
  return DVM_MPEG2_NOT_FOUND;
}

/*-----------------------------------------
 * Mpeg2GetCurrStartCode
 * 
 * Given BitParser *bp, returns the next
 * start code as an unsigned integer. If bp
 * is not positioned at a start code, 
 * DVM_MPEG2_NOT_FOUND is returned. No bytes
 * are parsed.
 *-----------------------------------------*/

unsigned int
Mpeg2GetCurrStartCode(BitParser *bp) {
  unsigned int code;
  Bp_InByteAlign(bp);
  Bp_PeekInt(bp, code);

  if ((code & 0xffffff00) != 0x00000100) {
    return DVM_MPEG2_NOT_FOUND;
  }
  return code;
}

/*-----------------------------------------
 * Mpeg2GetCurrExtensionCode
 *
 * Given BitParser *bp, returns type code of
 * extension header. If bp is not positioned
 * at an extension header, DVM_MPEG2_INVALID_START_CODE
 * is returned. No bytes are parsed.
 *-----------------------------------------*/

unsigned int
Mpeg2GetCurrExtensionCode(BitParser *bp) 
{
  unsigned int ext_code, code;
  
  Bp_InByteAlign(bp);
  Bp_GetInt(bp, code);

  if (code != MPEG2_EXT_START_CODE) {
    Bp_RestoreInt(bp);
    return DVM_MPEG2_INVALID_START_CODE;
  }

  Bp_PeekBits(bp, 4, ext_code);
  Bp_InByteAlign(bp);
  Bp_RestoreInt(bp);
  
  return ext_code;
}

/*-----------------------------------------
 * Mpeg2SkipToNextStartCode
 *
 * Given BitParser *bp, parses until next
 * start code. Bp is positioned at start of
 * next start code. If successful, returns
 * total number of bytes parsed. If no start
 * code found, returns DVM_MPEG2_NOT_FOUND.
 * The only difference between this function
 * and Mpeg2AnyStartCodeFind is that if bp
 * is already positioned at the beginning of
 * a start code, this function will skip it and
 * find the next one.
 *-----------------------------------------*/

int
Mpeg2SkipToNextStartCode(BitParser *bp)
{
  int total;
  Bp_InByteAlign(bp);
  Bp_FlushByte(bp);

  total = Mpeg2AnyStartCodeFind(bp);

  if (total != DVM_MPEG2_NOT_FOUND) {
    return total + 1;
  }

  return DVM_MPEG2_NOT_FOUND;
}
