/*
  media-filesystem library, io routines
  tridge@samba.org, January 2001
  released under the Gnu GPL v2
*/

#include "mfs.h"
#include "tproto.h"

static int readahead_enabled;
static int need_bswap;
static int verbose;
static u32 total_size;
static char *dev_list;
extern char *mfs_dev;

void mfs_readahead(int set)
{
	readahead_enabled = set;
}


/* this holds the list of block devices in the tivo */
static struct {
	char *dev;
	unsigned long sectors;
	int fd;
} devs[MAX_DEVS];

/* read from the virtual tivo disk (ie. all partitions concatenated) */
void mfs_read_sectors(void *buf, u32 sec, u32 count)
{
	int i;
#ifdef __linux__
	loff_t start=0;
#else
	off_t start=0;
#endif
	
	sec = partition_remap(sec);

	for (i=0; devs[i].dev; i++) {
		if (sec < start + devs[i].sectors) break;
		start += devs[i].sectors;
	}
	if (!devs[i].dev) {
		fprintf(stderr,"Failed to map sector %d\n", sec);
		exit(1);
	}

	if (verbose) {
		printf("mapped %d to %s/%d\n", sec, devs[i].dev, (int)(sec-start));
	}

	read_sectors(devs[i].fd, buf, sec-start, count);

	if (need_bswap) {
		u16 *v = (u16 *)buf;
		for (i=0;i<count<<(SECTOR_SHIFT-1);i++) v[i] = ntohs(v[i]);
	}
}


/* write to the virtual tivo disk (ie. all partitions concatenated) */
void mfs_write_sectors(void *buf, u32 sec, u32 count)
{
	int i;
#ifdef __linux__
	loff_t start=0;
#else
	off_t start=0;
#endif
	
	sec = partition_remap(sec);

	for (i=0; devs[i].dev; i++) {
		if (sec < start + devs[i].sectors) break;
		start += devs[i].sectors;
	}
	if (!devs[i].dev) {
		fprintf(stderr,"Failed to map sector %d\n", sec);
		exit(1);
	}

	if (verbose) {
		printf("mapped %d to %s/%d\n", sec, devs[i].dev, (int)(sec-start));
	}

	write_sectors(devs[i].fd, buf, sec-start, count);
}


void mfs_zero_sectors(int sector, int count)
{
	int chunk_size=512;
	char buf[chunk_size*512];
	char buf1[chunk_size*512];
	int total=0;

	bzero(buf, chunk_size*512);

	printf("wiping %d sectors at %d\n", count, sector);

	while (total<count) {
		int n = MIN(chunk_size, count);
		mfs_read_sectors(buf1, sector, n);
		if (memcmp(buf, buf1, n*SECTOR_SIZE)) {
			mfs_write_sectors(buf, sector, n);
		}
		sector += n;
		total += n;
		printf("%3.1f%%\r", 100.0*total/count);
		fflush(stdout);
	}
	printf("%3.1f%%\n", 100.0*total/count);
}


/* initialise the devices list from the superblock devlist */
void load_devs(char *devlist)
{
	char *p;
	int i=0, len;

	if (dev_list) devlist = dev_list;

	total_size = 0;

	while (1) {
		p = strchr(devlist, ' ');
		if (!p) {
			len = strlen(devlist);
		} else {
			len = (int)(p-devlist);
		}
		devs[i].dev = (char *)malloc(len+1);
		strncpy(devs[i].dev, devlist, len);
		devs[i].dev[len] = 0;
		devs[i].fd = open(devs[i].dev, O_RDWR|O_LARGEFILE);
		if (devs[i].fd == -1) {
			printf("failed to open [%s]\n", devs[i].dev);
			break;
		}
		devs[i].sectors = llseek(devs[i].fd, 0, SEEK_END) >> SECTOR_SHIFT;
		if (devs[i].sectors == 0) {
			ioctl(devs[i].fd, BLKGETSIZE, &devs[i].sectors);
		}
		devs[i].sectors &= ~(MFS_BLOCK_ROUND-1);
		total_size += devs[i].sectors;
		i++;
		devlist += len+1;
		if (!p) break;
	}	
}

void add_dev_map(char *mapping)
{
	char *p;

	dev_list = strdup(mapping);
	mfs_dev = strdup(dev_list);
	if ((p = strchr(mfs_dev, ' '))) *p = 0;
}

void io_dev_info(void)
{
	int i;

	for (i=0; devs[i].dev; i++) {
		printf("%s has %ld sectors\n", 
		       devs[i].dev,
		       devs[i].sectors);
	}
}

void io_need_bswap(int set)
{
	need_bswap = set;
}

u32 io_total_size(void)
{
	if (partition_total_size()) {
		total_size = partition_total_size();
	}	
	return total_size;
}


/* read bytes from a sector, handling partial sectors */
void mfs_read_partial(void *buf, u32 sec, u32 size)
{
	char tmp[SECTOR_SIZE];
	if (size >= SECTOR_SIZE) {
		u32 n = size / SECTOR_SIZE;
		mfs_read_sectors(buf, sec, n);
		buf += n*SECTOR_SIZE;
		size -= n*SECTOR_SIZE;
		sec += n;
	}
	if (size == 0) return;
	mfs_read_sectors(tmp, sec, 1);
	memcpy(buf, tmp, size);
}
