/*
  media-filesystem library
  tridge@samba.org, January 2001
  released under the Gnu GPL v2
*/

#include "mfs.h"
#include "tproto.h"

static struct mfs_super super;
static struct mfs_zone_map *zones[MAX_ZONES];
static int num_zones;

static int little_endian;
static int total_inodes;

char *mfs_dev;

unsigned fsid_hash(unsigned fsid, unsigned size)
{
	return fsid*67289 % size;
}

/* useful debug routine */
void dump_sectors(void *d, int n)
{
	u16 *v = d;
	int i, j;
	char *p;
	n = SECTOR_SIZE*n/2;
	for (i=0;i<n;i+=8) {
		int i0 = i;
		while (i>8 && i < (n-8) && memcmp(&v[i-8], &v[i], 16) == 0) {
			if (i == i0) printf("*\n");
			i += 8;
		}
		printf("%07x: %04x %04x %04x %04x %04x %04x %04x %04x   ", 
		       i*2, 
		       ntohs(v[i]), ntohs(v[i+1]), ntohs(v[i+2]), ntohs(v[i+3]),
		       ntohs(v[i+4]), ntohs(v[i+5]), ntohs(v[i+6]), ntohs(v[i+7]));
		p = (char *)&v[i];
		for (j=0;j<16;j++) {
			printf("%c", isprint(p[j])?p[j]:'.');
		}
		printf("\n");
	}
}



/* load the mfs super block - assumes MFS_DEVICE is set */
static void load_super(void)
{
	int fd;

		if (!mfs_dev) mfs_dev = getenv("MFS_DEVICE");
		if (!mfs_dev) {
			mfs_dev = "/dev/hda10";
			printf("Assuming MFS_DEVICE=%s\n", mfs_dev);
		}

		fd = open(mfs_dev, O_RDONLY|O_LARGEFILE);
		read_sectors(fd, &super, 0, 1);
		close(fd);

	load_devs(super.devlist);

	if (*(u16 *)&super == 0x1492 || *(u16 *)&super == 0x9214) {
		if (*(u16 *)&super == 0x1492) io_need_bswap(1);
		partition_parse();
		mfs_read_partial(&super, 0, sizeof(super));
	}

	switch (super.magic) {
	case 0xabbafeed: /* normal tivo access */
		break;
	case 0xbaabedfe:
		io_need_bswap(1);
		break;
	case 0xedfebaab:
		little_endian = 1;
		break;
	case 0xfeedabba:
		little_endian = 1;
		io_need_bswap(1);
		break;
	default:
		fprintf(stderr,"Not a TiVo super block! (magic=0x%08x)\n", 
			super.magic);
		exit(1);
	}

	mfs_read_partial(&super, 0, sizeof(super));

	check_crc((void *)&super, sizeof(super), &super.crc);
	
	byte_swap(&super, "i9 b128 i87");

	if (super.magic != 0xabbafeed) {
		fprintf(stderr,"Failed to byte swap correctly\n");
		exit(1);
	}

	if (io_total_size() && io_total_size() != super.total_sectors) {
		printf("WARNING: total sectors doesn't match (total=%d sb=%d)\n",
		       io_total_size(), super.total_sectors);
	}
}

/* load the mfs zones - currently we only use the inode
   zone but might as well load the lot */
static void load_zones(void)
{
	u32 next = super.zonemap_ptr;
	u32 map_size = super.zonemap_size;
	total_inodes = 0;

	while (next) {
		zones[num_zones] = (struct mfs_zone_map *)malloc(SECTOR_SIZE*map_size);
		mfs_read_sectors(zones[num_zones], next, map_size);
		check_crc(zones[num_zones], map_size*SECTOR_SIZE, &zones[num_zones]->crc);
		byte_swap(zones[num_zones], "i18");
		if (next != zones[num_zones]->sector) {
			fprintf(stderr,"sector wrong in zone (%d %d)\n",
				next, zones[num_zones]->sector);
			exit(1);
		}
		if (zones[num_zones]->type == ZONE_INODE) {
			total_inodes += zones[num_zones]->zone_size/2;
		}
		next = zones[num_zones]->next_zonemap_ptr;
		map_size = zones[num_zones]->next_zonemap_size;
		num_zones++;
		if (num_zones == MAX_ZONES) {
			fprintf(stderr,"Too many zones\n");
			exit(1);
		}
	}
}

/* turn a hash into a zone sector */
static u32 zone_sector(u32 hash)
{
	int i;
	u32 start = 0;
	for (i=0;i<num_zones;i++) {
		u32 len;
		if (zones[i]->type != ZONE_INODE) continue;
		len = zones[i]->zone_size/2;
		if (hash < start + len) {
			return zones[i]->zone_start + (hash-start)*2;
		}
		start += len;
	}
	fprintf(stderr, "Didn't find hash %x in zones!\n", hash);
	exit(1);
}

/* load one inode by fsid - optimised to avoid repeats */
static void mfs_load_inode(int fsid, struct mfs_inode *inode)
{
	static struct mfs_inode in;
	static u32 last_fsid;
	unsigned hash, hash1;

	if (fsid == last_fsid) {
		*inode = in;
		return;
	}

	hash1 = hash = fsid_hash(fsid, total_inodes);
	do {
		mfs_read_partial(&in, zone_sector(hash), sizeof(in));
		byte_swap(&in, "i10 b2 s1 i4");
		if (in.num_runs) {
			byte_swap(&in.u.runs[0], "i48");
		}
		hash = (hash+1) % total_inodes;
	} while ((in.flags & MFS_FLAGS_CHAIN) && in.id != fsid && hash != hash1);

	if (in.id != fsid) {
		fprintf(stderr, "ERROR: Didn't find fsid=%d!\n", fsid);
		exit(1);
	}

	*inode = in;
	last_fsid = fsid;
}

/* must call this before anything else */
void mfs_init(void)
{
	char *p = getenv("MFS_DEVLIST");
	if (p) add_dev_map(p);
	load_super();
	load_zones();
}

/* dump some global info on the mfs */
void mfs_info(void)
{
	int i;
	u32 total_size = io_total_size();
	printf("Super:\n\tstate=%x magic=%x\n\tdevlist=%s\n\tzonemap_ptr=%d total_secs=%d next_fsid=%d\n",
	       super.state, super.magic, super.devlist, super.zonemap_ptr,
	       super.total_sectors, super.next_fsid);
	printf("\tbackup_zonemap_ptr=%x zonemap_size=%d\n",
	       super.backup_zonemap_ptr, super.zonemap_size);

	io_dev_info();

	for (i=0; i<num_zones; i++) {
		printf("zone(%d):\n\tsector=%d type=%d start=%d next_zonemap=%d\n\tsize=%d per_chunk=%d limit=%d zone_size=%d\n",
		       i,
		       zones[i]->sector, 
		       zones[i]->type, 
		       zones[i]->zone_start, 
		       zones[i]->next_zonemap_ptr, 
		       zones[i]->zone_size,
		       zones[i]->per_chunk,
		       zones[i]->zone_start+zones[i]->zone_size,
		       zones[i]->zone_size);
		printf("\tbackup_sector=%x zonemap_size=%d\n\tbackup_next_zonemap=%x next_zonemap_size=%d\n\tbuddy_size=%d\n",
		       zones[i]->backup_sector, zones[i]->zonemap_size, 
		       zones[i]->backup_next_zonemap_ptr, 
		       zones[i]->next_zonemap_size,
		       zones[i]->buddy_size);
		if (total_size &&
		    zones[i]->zone_start+zones[i]->zone_size > total_size) {
			printf("Warning: zone is out of range\n");
		}
	}
	
}

/* dome some details about a fsid */
void mfs_fsid_info(int fsid)
{
	struct mfs_inode inode;
	int i;
	u32 hash = fsid_hash(fsid, zones[0]->zone_size/2);

	mfs_load_inode(fsid, &inode);
	printf("id=%d type=%d/%s hash=%x sec=%d typexx=%d units=%d size=%d runs=%d\n", 
	       inode.id, inode.type, 
	       mfs_type_string(inode.type),
	       hash, zones[0]->zone_start+hash*2,
	       inode.typexx,
	       inode.units, inode.size, inode.num_runs);
	for (i=0; i<inode.num_runs; i++) {
		printf("run 0x%08x:0x%x\n", 
		       inode.u.runs[i].start,
		       inode.u.runs[i].len);
	}
	printf("fsid %d is a total of %lld bytes\n", 
	       fsid, mfs_fsid_size(fsid));
}


/* read count bytes from a mfs file at offset ofs,
   returning the number of bytes read 
   ofs must be on a sector boundary
*/
u32 mfs_fsid_pread(int fsid, void *buf, u64 ofs, u32 count)
{
	struct mfs_inode inode;
	int i, n, dontfree = 0, dontcopy = 0;
	u32 start, ocount;
	u32 ret=0;
	u32 sec = ofs >> SECTOR_SHIFT;
	u64 size;
	static u_char *sparebuf = NULL;

	mfs_load_inode(fsid, &inode);

	if (inode.num_runs == 0) {
		if (ofs >= inode.size) return 0;
		ret = count;
		if (ret > inode.size-ofs) {
			ret = inode.size-ofs;
		}
		memcpy(buf, inode.u.data, ret);
		return ret;
	}

	size = inode.size;
	if (inode.units == 0x20000) {
		size <<= 17;
	}

	if (ofs > size) return 0;

	ocount = count;
	if (ofs + count > size) {
		count = size-ofs;
	}

	/* allocate a static buffer for later re-use */
	if (sparebuf == NULL) {
		sparebuf = malloc(131072);
	}

#if 0
	mfs_fsid_info(fsid);
#endif

	while (count > 0) {
		void *buf2;
		u32 n2;
		start = 0;
		for (i=0; i<inode.num_runs; i++) {
			if (sec < start + inode.u.runs[i].len) break;
			start += inode.u.runs[i].len;
		}
		if (i == inode.num_runs) return ret;
		n = (count+(SECTOR_SIZE-1)) >> SECTOR_SHIFT;
		if (n > inode.u.runs[i].len - (sec-start)) {
			n = inode.u.runs[i].len - (sec-start);
		}
		if ((n << SECTOR_SHIFT) == 131072) {
			buf2 = sparebuf;
			dontfree = 1;
		} else {
			buf2 = malloc(n << SECTOR_SHIFT);
			dontfree = 0;
		}
		if (ocount == count) {
			mfs_read_sectors(buf, inode.u.runs[i].start+(sec-start), n);
			dontcopy = 1;
		} else {
			mfs_read_sectors(buf2, inode.u.runs[i].start+(sec-start), n);
		}
		n2 = n << SECTOR_SHIFT;
		if (n2 > count) n2 = count;
		if (dontcopy == 1) {
			dontcopy = 0;
		} else {
			memcpy(buf, buf2, n2);
		}
		if (dontfree == 0)
			free(buf2);
		buf += n2;
		sec += n;
		count -= n2;
		ret += n2;
	}
	return ret;
}

#if 0
/* I haven't finished this yet - Warren */
/* write 128K bytes to a mfs file at chunk chk,
   returning 1 if ok, 0 if failure.
   ofs must be on a sector boundary
*/
u32 mfs_fsid_pwrite(int fsid, void *buf, u32 chk)
{
	struct mfs_inode inode;
	int i, n;
	u32 start;
	u32 ret=0;
	u32 sec = chk;
	u64 size;

	mfs_load_inode(fsid, &inode);

	if (inode.num_runs == 0) return(0); 
	if (inode.units != 0x20000) return(0);

	size = inode.size;
	if (chk >= size) return 0;

	/* Cross-check that every run is 128K in length */
	for (i=0; i<inode.num_runs; i++) {
		if (inode.u.runs[i].len!= 131072) return(0);
	}
	mfs_write_sectors(buf, inode.u.runs[chk].start, 131072);
	return(1);
}
#endif

/* return the type of a fsid */
int mfs_fsid_type(int fsid)
{
	struct mfs_inode inode;
	mfs_load_inode(fsid, &inode);
	return inode.type;
}

/* return the number of bytes used by a fsid */
u64 mfs_fsid_size(int fsid)
{
	struct mfs_inode inode;

	if (fsid == 0) return 0;
	mfs_load_inode(fsid, &inode);

	// mfs_fsid_info(fsid);
	switch (inode.units) {
	case 0: return inode.size;
	case 0x20000: return ((u64)inode.size) << 17;
	}
	fprintf(stderr, "ERROR: fsid=%d Unknown units %d\n", 
		fsid, inode.units);
	exit(1);
	return inode.size;
}

/* return the number of chunks in a fsid */
u32 mfs_fsid_numchunks(int fsid)
{
	struct mfs_inode inode;

	if (fsid == 0) return 0;
	mfs_load_inode(fsid, &inode);

	switch (inode.units) {
	  case 0x20000: return (inode.size);
	}
	fprintf(stderr, "ERROR: fsid=%d Unknown units %d\n", 
		fsid, inode.units);
	exit(1);
}

/* list a mfs directory - make sure you free with mfs_dir_free() */
struct mfs_dirent *mfs_dir(int fsid, u32 *count)
{
	u32 *buf, *p;
	int n=0, i;
	int size = mfs_fsid_size(fsid);
	int dsize, dflags;
	struct mfs_dirent *ret;

	*count = 0;

	if (size < 4) return NULL;

	if (mfs_fsid_type(fsid) != MFS_TYPE_DIR) {
		fprintf(stderr,"fsid %d is not a directory\n", fsid);
		// mfs_fsid_info(fsid);
		return NULL;
	}

	buf = (u32 *)malloc(size);
	mfs_fsid_pread(fsid, buf, 0, size);
	dsize = ntohl(buf[0]) >> 16;
	dflags = ntohl(buf[0]) & 0xFFFF;
	p = buf + 1;
	while ((int)(p-buf) < dsize/4) {
		u8 *s = ((char *)p)+4;
		p += s[0]/4;
		n++;
	}
	ret = malloc((n+1)*sizeof(*ret));
	p = buf + 1;
	for (i=0;i<n;i++) {
		u8 *s = ((char *)p)+4;
		ret[i].name = strdup(s+2);
		ret[i].type = s[1];
		ret[i].fsid = ntohl(p[0]);
		p += s[0]/4;
	}	
	ret[n].name = NULL;
	free(buf);
	*count = n;

	/* handle meta-directories. These are just directories which are
	   lists of other directories. All we need to do is recursively read
	   the other directories and piece together the top level directory */
	if (dflags == 0x200) {
		struct mfs_dirent *meta_dir = NULL;
		int meta_size=0;

		*count = 0;

		for (i=0;i<n;i++) {
			struct mfs_dirent *d2;
			int n2;
			if (ret[i].type != MFS_TYPE_DIR) {
				printf("ERROR: non dir %d/%s in meta-dir %d!\n", 
				       ret[i].type, ret[i].name, fsid);
				continue;
			}
			d2 = mfs_dir(ret[i].fsid, &n2);
			if (!d2 || n2 == 0) continue;
			meta_dir = realloc(meta_dir, sizeof(ret[0])*(meta_size + n2 + 1));
			memcpy(meta_dir+meta_size, d2, n2*sizeof(ret[0]));
			meta_size += n2;
			free(d2);
		}
		mfs_dir_free(ret);
		if (meta_dir) meta_dir[meta_size].name = NULL;
		*count = meta_size;
		return meta_dir;
	}


	return ret;
}

/* free a dir from mfs_dir */
void mfs_dir_free(struct mfs_dirent *dir)
{
	int i;
	for (i=0; dir[i].name; i++) {
		free(dir[i].name);
		dir[i].name = NULL;
	}
	free(dir);
}

/* return a string identifier for a tivo file type */
char *mfs_type_string(int type)
{
	
	switch (type) {
	case 0: return "NULL";
	case MFS_TYPE_DIR: return "tyDir";
	case MFS_TYPE_OBJ: return "tyDb";
	case MFS_TYPE_STREAM: return "tyStream";
	case MFS_TYPE_FILE: return "tyFile";
	default: 
		fprintf(stderr,"ERROR: Unknown file type %d!\n", type);
		exit(1);
		break;
	}
}

/* resolve a path to a fsid */
u32 mfs_resolve(char *path)
{
	char *p0, *tok, *r;
	u32 fsid;
	struct mfs_dirent *dir = NULL;

	if (path[0] != '/') {
		return atoi(path);
	}

	fsid = MFS_ROOT_FSID;
	p0 = strdup(path);
	path = p0;
	for (tok=strtok_r(path,"/", &r); tok; tok=strtok_r(NULL,"/", &r)) {
		u32 count;
		int i;
		dir = mfs_dir(fsid, &count);
		if (!dir) {
			fprintf(stderr,"resolve failed for fsid=%d\n",
				fsid);
			return 0;
		}
		for (i=0;i<count;i++) {
			if (strcmp(tok, dir[i].name) == 0) break;
		}
		if (i == count) {
			fsid = 0;
			goto done;
		}
		fsid = dir[i].fsid;
		if (dir[i].type != MFS_TYPE_DIR) {
			if (strtok_r(NULL, "/", &r)) {
				fprintf(stderr,"not a directory %s\n",tok);
				fsid = 0;
				goto done;
			}
			goto done;
		}
		mfs_dir_free(dir);
		dir = NULL;
	}

 done:
	if (dir) mfs_dir_free(dir);
	if (p0) free(p0);
	return fsid;
}


/* loop over all inodes calling fn on each one */
void mfs_all_inodes(void (*fn)(struct mfs_inode *))
{
	int i, z;
	u32 start_hash=0;
	struct mfs_inode inode;

	for (z=0; z<num_zones;z++) {
		if (zones[z]->type != ZONE_INODE) continue;
		for (i=0; i<zones[z]->zone_size; i+= 2) {
			u32 hash = start_hash + i/2;
			mfs_read_partial(&inode, zone_sector(hash), 
					 sizeof(inode));
			byte_swap(&inode, "i10 b2 s1 i4");
			if (inode.num_runs) {
				byte_swap(&inode.u.runs[0], "i48");
			}
			if (inode.id != 0) {
				if (inode.id > super.next_fsid) {
					printf("invalid fsid %d (next=%d)\n",
					       inode.id, super.next_fsid);
					exit(1);
				}
				fn(&inode);
			}
		}
		start_hash += zones[z]->zone_size/2;
	}
}


struct bitmap *mfs_zone_bitmap(int zone, u64 limit)
{
	struct bitmap *bm;
	int num_blocks;

	void bitmap_fn(struct mfs_inode *inode) {
		int i;
		//printf("inode %d runs=%d\n", inode->id, inode->num_runs);
		if (limit && mfs_fsid_size(inode->id) > limit) {
			printf("skipping inode %d of size=%lld\n", 
			       inode->id, mfs_fsid_size(inode->id));
			return;
		}
		// if (bitmap_excluded(inode->id)) return;
		for (i=0;i<inode->num_runs;i++) {
			u32 start, len;
			start = inode->u.runs[i].start;
			len = inode->u.runs[i].len;
			if (start < zones[zone]->zone_start ||
			    start >= zones[zone]->zone_start+
			    zones[zone]->zone_size) {
				continue;
			}
			if (len % zones[zone]->per_chunk) {
				printf("Not a multiple of per-chunk? fsid=%d\n", inode->id);
				exit(1);
			}
			bitmap_set(bm, 
				   (start-zones[zone]->zone_start)/
				   zones[zone]->per_chunk,
				   len/zones[zone]->per_chunk);
		}
	}

	num_blocks = zones[zone]->zone_size / zones[zone]->per_chunk;
	bm = bitmap_allocate(num_blocks);
	
	mfs_all_inodes(bitmap_fn);

	return bm;
}


void mfs_purge_zone(int zone, u32 limit)
{
	struct bitmap *bm;
	int i, n;
	int units = zones[zone]->per_chunk;

	bm = mfs_zone_bitmap(zone, limit);
	
	for (i=0;i<bm->n;) {
		if (bitmap_query(bm, i)) {
			i++; continue;
		}
		n = 1;
		while (bitmap_query(bm, i+n) == 0 && 
		       i+n < bm->n) n++;
		mfs_zero_sectors(zones[zone]->zone_start+i*units,
				 n*units);
		i += n;
	}
}

void mfs_purge_all(u64 limit)
{
	int z;

	for (z=0; z<num_zones;z++) {
		if (zones[z]->type != ZONE_STREAM &&
		    zones[z]->type != ZONE_FILE) continue;
		printf("Purging zone %d with size limit %lld\n", z, limit);
		mfs_purge_zone(z, limit);
	}
}

/* zone size in sectors */
u32 mfs_zone_size(int zone)
{
	return zones[zone]->zone_size;
}
