/*
 * Copyright (C) 2002, 2003  John Barrett (MrBassMan) <johnbarretthcc@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: processor.cpp,v 1.24 2003/03/30 04:41:43 olaf_be Exp $ */

#include "processor.h"

#include <assert.h>
#include <string.h>
#include <time.h>
#if !defined(WIN32)
#include <ctype.h>
#endif

#include "tymalloc.h"
#include <qmessagebox.h>

static char demux_progress[ 26 ];
static char transcode_progress[ 26 ];
static char mplex_progress[ 26 ];
static time_t start_time = 0;
static time_t last_progress_time = 0;
static char last_progress[ 256 ] = {0};
static bool bAbort = false;
static bool bErrorSet = false;

/*
* You can now fetch images and intit the image just as before the fact that
tystream->remote_holder isn't null takes care of that. Hence this
tystream->process is
totally transparent.

* When the user is finished and wants to process - you will need to close the
remote recording but NOT the free or close the remote server or so!! I.e.
just do a tydemux_close_remote_tystream(remote_holder) and free the tystream
- this will NOT free the remote_holder present in the tystream structure.

* Now you will need to pass the remote holder pointer and the fsid as
parametes to typrossess/tydemux. NOTE!!!!!!!: This is a bit fuzzy on my side
- it's a absolute must the the remote_holder in the demux parameters is NULL
if we don't do any remote stuff - PLEASE check typrocess.cpp and main.c under
typrocess to make sure of that! This is the thing I mentioned above.

* tydemux/tytranscode/tymplex will now do it's job as normal and when they are
finished we will need to close the server connection this is done in
tydemux_close_tyserver(remote_holder) - we can now free the remote holder this is done with free_remote_holder(remote_holder). NOTE - we will probably
need to do something to atexit in tydemux if we exit abnormally - if a remote
session is in place we will need to close the remote tystream with
tydemux_close_remote_tystream if we exit abnormally.

*/

int logger_hook(int level, const char *module, const char *msg )
{
	if( level == log_progress ) {
		/* do not output progress messages to the log file, used stderr instead */
		bErrorSet = false;

		/* Check the module name and copy to the appropriate buffer */
		if( strcmp( module, "demux" ) == 0 ) {
			strncpy( demux_progress, msg, 25 );
			/* trim trailing whitespace */
			while( isspace( demux_progress[ strlen(demux_progress) - 1] ) ) {
				demux_progress[ strlen(demux_progress) - 1] = 0;
			}
		} else if( strcmp( module, "transcode" ) == 0 ) {
			strncpy( transcode_progress, msg, 25 );
			/* trim trailing whitespace */
			while( isspace( transcode_progress[ strlen(transcode_progress) - 1] ) ) {
				transcode_progress[ strlen(transcode_progress) - 1] = 0;
			}
		} else if( strcmp( module, "mplex" ) == 0 ) {
			strncpy( mplex_progress, msg, 25 );
			/* trim trailing whitespace */
			while( isspace( mplex_progress[ strlen(mplex_progress) - 1] ) ) {
				mplex_progress[ strlen(mplex_progress) - 1] = 0;
			}
		}

		if( last_progress_time < time(0) ) {
			sprintf( last_progress, "TyDemux: %-25.25s  TyTranscode: %-25.25s   TyMplex: %-25.25s",
				demux_progress, transcode_progress, mplex_progress );
			last_progress_time = time(0);
		}

		return 0;
	} else if( level == log_error ) {
		sprintf( last_progress, "ERROR: %s", msg );
		last_progress_time = time(0);
		bErrorSet = true;
	} else if( level == log_fatal ) {
		sprintf( last_progress, "FATAL: %s", msg );
		last_progress_time = time(0);
		bErrorSet = true;
		bAbort = true;
	}

	return 1;
}

/* Background processor task */
int processor_thread( void *user_data )
{
	processor_thread_params_t *params;

	/* First progress message so initialise everything */
	strcpy( demux_progress, "Not Running");
	strcpy( transcode_progress, "Not Running");
	strcpy( mplex_progress, "Not Running");
	sprintf( last_progress, "TyDemux: %-25.25s  TyTranscode: %-25.25s   TyMplex: %-25.25s",
		demux_progress, transcode_progress, mplex_progress );
	start_time = time(0);

	params = (processor_thread_params_t *) user_data;
	params->progress = 0;
	params->complete = 0;
	params->result = 0;

#ifdef _WIN32
	params->result = TyProcessDll( &(params->typrocess_params) );
#else
	params->result = typrocess( &(params->typrocess_params) );
#endif
	if( params->result == 0 && (bAbort || bErrorSet) ){
		params->result = -1;
	}
	params->progress = 100;
	params->complete = 1;

	return 0;
}

CDemuxProcessTask::CDemuxProcessTask( const char *tyFile, const char *outfile, FILE *logFile )
: m_th(0)
{
	bAbort = false;

	memset( &m_Params, 0, sizeof(processor_thread_params_t) );
	m_Params.complete = 0;
	m_Params.result = 0;
	m_Params.progress = 0;

	if( tyFile && tyFile[0] ) {
		strcpy( m_Params.typrocess_params.in_ty_file, tyFile );
	}
	strcpy( m_Params.typrocess_params.out_mpeg_file, outfile );
	strcpy( m_Params.typrocess_params.out_video_file, outfile );
	m_Params.typrocess_params.demux = 1;
	m_Params.typrocess_params.mplex = 1;

	m_Params.typrocess_params.progress = &(m_Params.progress);

	/* Set a log file */
	m_Params.typrocess_params.logfile = logFile;
	m_Params.typrocess_params.loglevel = log_info;
	m_Params.typrocess_params.loghookfn = logger_hook;

}

CDemuxProcessTask::~CDemuxProcessTask()
{
	int i;
	CancelJob();

	if( m_Params.typrocess_params.demux_params.cutlist ) {
		while( m_Params.typrocess_params.demux_params.cutcount ) {
			free( m_Params.typrocess_params.demux_params.cutlist[
					--(m_Params.typrocess_params.demux_params.cutcount) ] );
		}
	free( m_Params.typrocess_params.demux_params.cutlist );
	}

	// Free additional command line parameters
	for( i = 0; i < m_Params.typrocess_params.mplex_params.argc; ++i ) {
		free( m_Params.typrocess_params.mplex_params.argv[i] );
	}
	if( m_Params.typrocess_params.mplex_params.argv ) {
		free( m_Params.typrocess_params.mplex_params.argv );
	}
	// Free additional command line parameters
	for( i = 0; i < m_Params.typrocess_params.transcode_params.argc; ++i ) {
		free( m_Params.typrocess_params.transcode_params.argv[i] );
	}
	if( m_Params.typrocess_params.transcode_params.argv ) {
		free( m_Params.typrocess_params.transcode_params.argv );
	}
	// Free additional command line parameters
	for( i = 0; i < m_Params.typrocess_params.demux_params.argc; ++i ) {
		free( m_Params.typrocess_params.demux_params.argv[i] );
	}
	if( m_Params.typrocess_params.demux_params.argv ) {
		free( m_Params.typrocess_params.demux_params.argv );
	}

}

void CDemuxProcessTask::SetRemoteFile( remote_holder_t *pRemote, int fsid )
{
	m_Params.typrocess_params.demux_params.fsid = fsid;
	m_Params.typrocess_params.demux_params.remote_holder = pRemote;
}

// Set up per process command line parameters
void CDemuxProcessTask::SetModuleOptions( const char *demux, const char *transcode, const char *mplex )
{
	char strTemp[256];
	int args;
	char *ptr;

	if( demux && demux[0] ) {
		strcpy( strTemp, demux );
		// 64 command line parameters should be more than enough
		m_Params.typrocess_params.demux_params.argv = (char **)malloc( sizeof(char *) * 64 );

		m_Params.typrocess_params.demux_params.argc = 1;
		m_Params.typrocess_params.demux_params.argv[0] = strdup( "tyeditor" );
		ptr = strtok( strTemp, " \t\n" );

		while( ptr ) {
			m_Params.typrocess_params.demux_params.argv[m_Params.typrocess_params.demux_params.argc++] = strdup( ptr );
			ptr = strtok( NULL, " \t\n" );
			assert( m_Params.typrocess_params.demux_params.argc < 64 );
		}
	}

	if( transcode && transcode[0] ) {
		strcpy( strTemp, transcode );
		// 64 command line parameters should be more than enough
		m_Params.typrocess_params.transcode_params.argv = (char **)malloc( sizeof(char *) * 64 );

		m_Params.typrocess_params.transcode_params.argc = 1;
		m_Params.typrocess_params.transcode_params.argv[0] = strdup( "tyeditor" );
		ptr = strtok( strTemp, " \t\n" );

		while( ptr ) {
			m_Params.typrocess_params.transcode_params.argv[m_Params.typrocess_params.transcode_params.argc++] = strdup( ptr );
			ptr = strtok( NULL, " \t\n" );
			assert( m_Params.typrocess_params.transcode_params.argc < 64 );
		}
	}

	if( mplex && mplex[0] ) {
		strcpy( strTemp, mplex );
		// 64 command line parameters should be more than enough
		m_Params.typrocess_params.mplex_params.argv = (char **)malloc( sizeof(char *) * 64 );

		m_Params.typrocess_params.mplex_params.argc = 1;
		m_Params.typrocess_params.mplex_params.argv[0] = strdup( "tyeditor" );
		ptr = strtok( strTemp, " \t\n" );

		while( ptr ) {
			args = m_Params.typrocess_params.mplex_params.argc++;	
			m_Params.typrocess_params.mplex_params.argv[args] = strdup( ptr );
			ptr = strtok( NULL, " \t\n" );
			assert( m_Params.typrocess_params.mplex_params.argc < 64 );
		}
	}

}

void CDemuxProcessTask::AddCut( const char *CutString )
{
	/* Resize the cut entry list after every 64 entries */
	if( !(m_Params.typrocess_params.demux_params.cutlist) ) {
		m_Params.typrocess_params.demux_params.cutlist = (char **)malloc( sizeof( char * ) * 64 );
	} else if ( ( (m_Params.typrocess_params.demux_params.cutcount)%64 ) == 0 ) {
		int newSize = sizeof( char * ) * 64;
		newSize *= ( ( (m_Params.typrocess_params.demux_params.cutcount)%64 ) + 1);
		m_Params.typrocess_params.demux_params.cutlist =
			(char **) realloc( m_Params.typrocess_params.demux_params.cutlist, newSize );
	}

	m_Params.typrocess_params.demux_params.cutlist[	m_Params.typrocess_params.demux_params.cutcount ] = strdup( CutString );
	++(m_Params.typrocess_params.demux_params.cutcount);
}

void CDemuxProcessTask::SetAudioFormat( eAudioFormat format, int rate, int bitrate )
{
	switch( format ) {
	case AF_AC3:
		m_Params.typrocess_params.transcode_params.format = 1;
		break;
	case AF_MPEG:
	default:
		m_Params.typrocess_params.transcode_params.format = 0;
		break;
	}

	m_Params.typrocess_params.transcode_params.rate = rate;
	m_Params.typrocess_params.transcode_params.bitrate = bitrate;
	m_Params.typrocess_params.transcode_params.dbadjust = 255; //RLM DEFAULT_DB_ADJUST;

	m_Params.typrocess_params.transcode = 1;
}

void CDemuxProcessTask::SetVideoFormat( eVideoFormat format )
{

	/* NOTE: The mplex_params.format is the mplex -f X + 1 */

	switch( format ) {
	case VF_DVD:
		m_Params.typrocess_params.mplex_params.format = 9;
		m_Params.typrocess_params.mplex = 1;
		break;
	case VF_SVCD:
		m_Params.typrocess_params.mplex_params.format = 6;
		m_Params.typrocess_params.mplex = 1;
		break;
	/* We can't support VCD/MPEG1 since there is no mpeg1 Tivo
	if the user enables this we will get into big trouble */
	case VF_VCD:
//		m_Params.typrocess_params.mplex_params.format = 3;
//		m_Params.typrocess_params.mplex = 1;
		break;
	case VF_MPEG1:
//		m_Params.typrocess_params.mplex_params.format = 1;
//		m_Params.typrocess_params.mplex = 1;
		break;
	case VF_MPEG2:
		m_Params.typrocess_params.mplex_params.format = 4;
		m_Params.typrocess_params.mplex = 1;
		break;
	case VF_ELEMENTAL:
		m_Params.typrocess_params.mplex = 0;
		break;
	}
}

void CDemuxProcessTask::SetAudioFile( const char *fileName )
{
	strcpy( m_Params.typrocess_params.out_audio_file, fileName );
}

const char *CDemuxProcessTask::GetVideoFile()
{
	return( m_Params.typrocess_params.out_video_file );
}

bool CDemuxProcessTask::StartJob()
{
	assert( !m_th );

	m_th = thread_start( processor_thread, &m_Params );

	return true;
}

bool CDemuxProcessTask::CancelJob()
{
	if( m_th ) {
		thread_stop( m_th, -1 );
		m_th = 0;
		m_Params.result = -1;
		return true;
	}

	return false;
}

bool CDemuxProcessTask::IsCompleted()
{
	if( !m_th ) {
		return true;
	}

	if( bAbort )
	{
		return true;
	} else if( m_Params.complete ) {
		return true;
	} else {
		return false;
	}
}

int CDemuxProcessTask::GetResult()
{
	if( m_Params.result == 0 && (bAbort || bErrorSet) ){
		m_Params.result = -1;
	}

	return( m_Params.result );
}

int  CDemuxProcessTask::GetProgress()
{
	return( m_Params.progress );
}

const char *CDemuxProcessTask::GetProgressText()
{
	return( last_progress );
}
