/*
 * Copyright (C) 2002, 2003  John Barrett (MrBassMan) <johnbarretthcc@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: tydemuxclass.cpp,v 1.30 2003/03/29 13:29:29 mrbassman Exp $ */

#include "tydemuxclass.h"

#include <assert.h>
#include <string.h>

#include "tymalloc.h"

#define FILE_READ_BUFFER_SIZE 1024

void TimeAsString( int64_t ticks, char *timeString )
{
	int64_t hours, mins, secs;
	int64_t msecs = ticks/TICKS_PER_MILLISECOND;

	hours = ((msecs/1000)/60)/60;
	if( hours < 0 ) hours = 0;
	msecs -= hours * 60 * 60 * 1000;
	mins = (msecs/1000)/60;
	if( mins < 0 ) mins = 0;
	msecs -= mins * 60 * 1000;
	secs = msecs/1000;
	if( secs < 0 ) secs = 0;
	msecs -= secs * 1000;
	if( msecs < 0 ) msecs = 0;

	sprintf( timeString, "%02d:%02d:%02d.%03d", (int)hours, (int)mins, (int)secs, (int)msecs );
}

void FrameAsString( int64_t frame, char *frameString )
{
	sprintf( frameString, "%010lld", frame );
}

CDemuxCutPoint::CDemuxCutPoint()
{
	m_Type = unknown;
	m_Index = NULL;
	frame = 0;
	time = 0;
	ticks = 0;
}

void CDemuxCutPoint::GetTimeAsString( char *TimeString )
{

	TimeAsString( time, TimeString );
}


void CDemuxCutPoint::GetCutAsString( char *CutString )
{
	char t[15], c[15], r[15];

	FrameAsString( m_Index->time_of_iframe, t );
	FrameAsString( m_Index->seq_rec_nr, r );
	FrameAsString( m_Index->chunk_number_seq, c );

	sprintf( CutString, "@%s %s %s@", t, c, r );
}

CDemuxInterface::CDemuxInterface()
{
	m_Index = NULL;
	m_tyStream = NULL;
	m_init_image = NULL;
	m_last_image = NULL;
	m_currentGop = NULL;
	m_CurrentStreamTime = 0;
	m_CurrentStreamFrame = 0;
	m_AvgTicksPerGOP = 0;
	m_pRemote = NULL;
	m_fsid = 0;
	strcpy( m_tyFile, "UNKNOWN" );
	strcpy( m_tivoHost, "tivo" );
}

CDemuxInterface::~CDemuxInterface()
{
	if( m_last_image ) {
		free( m_last_image->data_buffer );
		free(m_last_image);
	}

	if( m_init_image ) {
		free( m_init_image->data_buffer );
		free(m_init_image);
	}

	if( m_tyStream ) {
		free_tystream(m_tyStream);
	}

	closeRemote();
}

/* Load the ty file and get ready for a new editing session */
bool CDemuxInterface::LoadTY( const char *fileName, int TivoVersion )
{
	int in_file;
	stat64_t file_info;
	off64_t in_file_size;

	// Parameter validation
	assert( !m_tyStream );
	assert( TivoVersion == 1 || TivoVersion == 2 );

	 /* Open the TY file */
	in_file = open(fileName, OPEN_READONLY_FLAGS | READ_PERMISSIONS );
	if(in_file == -1) {
		LOG_ERROR1( "Failed to open the ty file %s\n", fileName );
		return( false );
	}
	fstat64(in_file, &file_info);
	in_file_size = file_info.st_size;

	if(in_file_size < 5242880) {
		LOG_ERROR("TyStream is to small to demux - We need at least 40 chunks\n");
		return( false );
	}

	m_tyStream = new_tystream(DEMUX);

	tystream_set_in_file(m_tyStream, in_file, in_file_size);
	m_tyStream->tivo_version = (tivo_version_t) TivoVersion;
	m_tyStream->std_alone = 1;

	if(!std_probe_tystream(m_tyStream)) {
		LOG_ERROR("Error in probe\n");
		free_tystream(m_tyStream);
		m_tyStream = NULL;
		return (false);
	}

	if(m_tyStream->miss_alinged) {
		LOG_ERROR("Your stream is not aligned - please use -b to specify byte offset\n");
		free_tystream(m_tyStream);
		m_tyStream = NULL;
		exit(tydemux_return(-1000));
	}

	print_probe(m_tyStream);
	get_start_chunk(m_tyStream);

	m_CurrentStreamTime = 0;
	m_CurrentStreamFrame = 0;
	strcpy( m_tyFile, fileName );

	return ( true );
}

void CDemuxInterface::closeRemote()
{
	if( m_pRemote ) {
		tydemux_close_remote_tystream(m_pRemote );
		tydemux_close_tyserver(m_pRemote );
		free_remote_holder( m_pRemote );
	}
	m_pRemote = NULL;
}

#define check_strncpy(d,s,l) { if(s) strncpy(d,s,l); }

CTivoFile *CDemuxInterface::GetTivoFileList( const char *tivohostname )
{
	CTivoFile *pList = NULL;
	CTivoFile *pCurrent = NULL;

	closeRemote();

	strcpy( m_tivoHost, tivohostname );

	m_pRemote = new_remote_holder( m_tivoHost );

	if( m_pRemote && tydemux_init_remote( m_pRemote ) ) {
		fsid_index_t *pIndex = m_pRemote->fsid_index;

		while( pIndex ) {
			if( !pList ) {
				pList = new CTivoFile;
				pCurrent = pList;
			} else {
				pCurrent->pNext = new CTivoFile;
				pCurrent = pCurrent->pNext;
			}

			memset( pCurrent, 0, sizeof( CTivoFile ) );

			pCurrent->fsid = pIndex->fsid;
			pCurrent->state = pIndex->state;
			/* From John to Olaf - logic seems reversed here? From Olaf to John - fixed :)*/
			pCurrent->indexed = (pIndex->index_avalible == 0) ? false:true;
			check_strncpy( pCurrent->tystream, pIndex->tystream, 256 );
			check_strncpy( pCurrent->year, pIndex->year, 25 );
			check_strncpy( pCurrent->air, pIndex->air, 25 );
			check_strncpy( pCurrent->day, pIndex->day, 25 );
			check_strncpy( pCurrent->date, pIndex->date, 25 );
			check_strncpy( pCurrent->rectime, pIndex->rectime, 25 );
			check_strncpy( pCurrent->duration, pIndex->duration, 25 );
			check_strncpy( pCurrent->title, pIndex->title, 256 );
			if( !pIndex->episode || !pIndex->episode[0] ) {
				strcpy( pCurrent->episode, "Not an Episode" );
			} else {
				check_strncpy( pCurrent->episode, pIndex->episode, 256 );
			}
			check_strncpy( pCurrent->episodenr, pIndex->episodenr, 25 );
			check_strncpy( pCurrent->description, pIndex->description, 256 );

			pIndex = pIndex->next;
		}
	}

	return pList;
}

bool CDemuxInterface::OpenRemoteTivoFile( int fsid )
{
	if( !m_pRemote )
		return false;
	assert( !m_tyStream );

	m_tyStream = tydemux_open_probe_remote( m_pRemote, fsid );
	m_fsid = fsid;

	if( !m_tyStream ) {
		closeRemote();
		return false;
	}

	m_tyStream->remote_holder = m_pRemote;
	m_tyStream->fsid = fsid;

	return true;
}

CDemuxIndexTask *CDemuxInterface::CreateIndexer(FILE *logfile)
{
	assert( m_tyStream );

	return( new CDemuxIndexTask( m_tyStream, logfile ) );

}

CDemuxCopyTask *CDemuxInterface::CreateCopier( int fsid, const char *dest, FILE *logfile )
{
	assert( m_pRemote );

	return( new CDemuxCopyTask( m_pRemote, fsid, dest, logfile ) );

}

int	CDemuxInterface::GetFrameRate()
{
	return( m_tyStream->frame_rate );
}

gop_index_t *CDemuxInterface::get_CurrentGopIndex()
{
	if( m_currentGop )
		return m_currentGop;

	// First time of calling - get the first GOP index
	assert( m_tyStream->index ); // Indexing has not been completed

	if( !m_init_image ) {
		/* First time - get a seq module */
		m_init_image = init_image( m_tyStream);
		if(!m_init_image) {
			assert(m_init_image); //We faild to get a seq even this func is recursive
		}
	}
	m_currentGop = m_tyStream->index->real_start_gop_index;

	/* Get an estimate of the number of ticks per GOP */
	m_AvgTicksPerGOP = ( m_tyStream->index->real_end_gop_index->display_time -
                       m_tyStream->index->real_start_gop_index->display_time )/
                       ( m_tyStream->index->nr_of_gops );

	assert( m_AvgTicksPerGOP );
	return ( m_currentGop );
}

void CDemuxInterface::setCurrentGOP( gop_index_t *index )
{
	if( m_currentGop != index ) {
		/* Save the new stream frame position */
		if( m_last_image ) {
			/* Free the current image */
			free(m_last_image->data_buffer);
			free(m_last_image);
			m_last_image = NULL;
		}

		m_currentGop = index;
		m_CurrentStreamFrame = index->chunk_number_i_frame_pes;
	}
}

bool CDemuxInterface::GetCurrentImage( ImagePtr_t &pImage, int &size )
{
	gop_index_t *pIndex = get_CurrentGopIndex();
	if( !pIndex )
		return false;

	if( !m_last_image ) {
		m_last_image = get_image( m_tyStream, pIndex, m_init_image );
		if( !m_last_image )
			return false;
	}

	/* Pass the image back to the user */
	size = m_last_image->buffer_size;
	pImage = m_last_image->data_buffer;

	return true;
}

/* Work with the currently stream time pointer in GOPS */
void CDemuxInterface::SetStreamGopRelative( int64_t GOPOffset )
{
	gop_index_t *index = get_CurrentGopIndex();

	if( !index )
		return;

	if( GOPOffset > 0 ) {
		/* Search forwards */
		while( index->next && (GOPOffset--) != 0 ) {
			index = index->next;
		}
	} else {
		/* Search backwards */
		while( index->previous && (GOPOffset++) != 0  ) {
			index = index->previous;
		}
	}

	setCurrentGOP( index );
}

void CDemuxInterface::set_GopIndexFromTime()
{
	gop_index_t *index = get_CurrentGopIndex();

	if( !index )
		return;

	ticks_t ticksCurrent = m_CurrentStreamTime;

	/* Find the fastest place to search from - start, current or end */
	ticks_t diff_current = llabs( ticksCurrent - index->display_time );
	ticks_t diff_start = llabs( ticksCurrent - m_tyStream->index->real_start_gop_index->display_time );
	ticks_t diff_last = llabs( ticksCurrent - m_tyStream->index->real_end_gop_index->display_time );
	if( diff_current > diff_start ) {
		index = m_tyStream->index->real_start_gop_index;
		diff_current = diff_start;
	}
	if( diff_current > diff_last ) {
		index = m_tyStream->index->real_end_gop_index;
		diff_current = diff_last;
	}

	/* First do an index based search to roughly find our position */
	int64_t indexcount = diff_current/m_AvgTicksPerGOP;
	if( indexcount > 1 ) {
		indexcount -= 1; // Deliberately undershoot

		if( index->display_time < m_CurrentStreamTime ) {
			/* Search forwards */
			while( index->next && indexcount ) {
				index = index->next;
				--indexcount;
			}
		} else {
			/* Search backwards */
			while( index->previous && indexcount ) {
				index = index->previous;
				--indexcount;
			}
		}
	}

	assert( index );

	/* Now do the slower search which gets the index time for each frame it looks at */
	while( 1 ) {
		ticks_t tPrevious, tNext;

		/* Check special cases first */
		if( index->display_time == m_CurrentStreamTime ) {
			break;
		}

		/* Get the time of the next GOP */
		if( index->next ) {
			tNext = index->next->display_time;
		} else {
			tNext = index->display_time;
		}
		/* Get the time of the previous GOP */
		if( index->previous ) {
			tPrevious = index->previous->display_time;
		} else {
			tPrevious = index->display_time;
		}

		/* Check to see if we need to move forward/back a whole GOP */
		if( tNext <= m_CurrentStreamTime ) {
			if( index->next ) {
				index = index->next;
				continue;
			}
			break;
		}
		if( tPrevious >= m_CurrentStreamTime ) {
			if( index->previous ) {
				index = index->previous;
				continue;
			}
			break;
		}

		/* To get here the time is somewhere between the previous and next GOPs */
		/* See which one we are closest to, previous, current or next */
		ticks_t diffCurrent = llabs( m_CurrentStreamTime - index->display_time );
		if( index->display_time > m_CurrentStreamTime ) {
			ticks_t diffPrevious = m_CurrentStreamTime - tPrevious;
			if( diffPrevious < diffCurrent ) {
				index = index->previous;
			}
		} else {
			ticks_t diffNext = tNext - m_CurrentStreamTime;
			if( diffNext < diffCurrent ) {
				index = index->next;
			}
		}
		break;
	}

	setCurrentGOP( index );
}

void CDemuxInterface::set_GopIndexFromFrame()
{
	gop_index_t *index = get_CurrentGopIndex();

	if( !index )
		return;

	/* Find the fastest place to search from - start, current or end */
	int64_t diff_current = llabs((m_CurrentStreamFrame - index->chunk_number_i_frame_pes) );
	int64_t diff_start = llabs((m_CurrentStreamFrame - m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes) );
	int64_t diff_last = llabs((m_CurrentStreamFrame - m_tyStream->index->real_end_gop_index->chunk_number_i_frame_pes) );
	if( diff_current > diff_start ) {
		index = m_tyStream->index->real_start_gop_index;
		diff_current = diff_start;
	}
	if( diff_current > diff_last ) {
		index = m_tyStream->index->real_end_gop_index;
		diff_current = diff_last;
	}

	if( index->chunk_number_i_frame_pes < m_CurrentStreamFrame ) {
		/* Search forwards */
		while( index->next && index->chunk_number_i_frame_pes < m_CurrentStreamFrame ) {
			index = index->next;
		}
	} else {
		/* Search backwards */
		while( index->previous && index->chunk_number_i_frame_pes > m_CurrentStreamFrame ) {
			index = index->previous;
		}
	}

	setCurrentGOP( index );
	m_CurrentStreamTime = index->display_time;
}

int64_t CDemuxInterface::GetStreamTime()
{
	ticks_t totalticks = m_tyStream->index->real_end_gop_index->display_time - m_tyStream->index->real_start_gop_index->display_time;

	return( totalticks );
}

int64_t CDemuxInterface::GetStreamCurrentTime(bool noOffset)
{
	if( noOffset ) {
		return( m_CurrentStreamTime );
	} else {
		return( m_CurrentStreamTime - m_tyStream->index->real_start_gop_index->display_time );
	}
}

void CDemuxInterface::SetStreamTime( const char *szTime, bool bSetTimeToExactGOP )
{
	SetStreamTime( cutpoint_convert_time_to_ticks( szTime), bSetTimeToExactGOP );
}

void CDemuxInterface::SetStreamTime( int64_t streamTime, bool bSetTimeToExactGOP, bool noOffset )
{
	m_CurrentStreamTime = streamTime;

	if( !noOffset ) {
		m_CurrentStreamTime += m_tyStream->index->real_start_gop_index->display_time;
	}
	if( m_CurrentStreamTime < m_tyStream->index->real_start_gop_index->display_time ) {
		m_CurrentStreamTime = m_tyStream->index->real_start_gop_index->display_time;
	}
	if( m_CurrentStreamTime > m_tyStream->index->real_end_gop_index->display_time ) {
		m_CurrentStreamTime = m_tyStream->index->real_end_gop_index->display_time;
	}
	set_GopIndexFromTime();

	if(bSetTimeToExactGOP) {
		m_CurrentStreamTime = get_CurrentGopIndex()->display_time;
	}
}

void CDemuxInterface::SetStreamTimeRelative( int64_t streamTimeOffset, bool bSetTimeToExactGOP )
{
	m_CurrentStreamTime += streamTimeOffset;
	if( m_CurrentStreamTime < m_tyStream->index->real_start_gop_index->display_time ) {
		m_CurrentStreamTime = m_tyStream->index->real_start_gop_index->display_time;
	}
	if( m_CurrentStreamTime > m_tyStream->index->real_end_gop_index->display_time ) {
		m_CurrentStreamTime = m_tyStream->index->real_end_gop_index->display_time;
	}

	set_GopIndexFromTime();

	if(bSetTimeToExactGOP) {
		m_CurrentStreamTime = get_CurrentGopIndex()->display_time;
	}
}

/* Work with the currently stream pointer in frames */
int64_t CDemuxInterface::GetStreamFrames()
{
	return( (m_tyStream->index->real_end_gop_index->chunk_number_i_frame_pes) -
			(m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes) );
}

int64_t CDemuxInterface::GetStreamCurrentFrame()
{
	return m_CurrentStreamFrame - m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes;
}

void CDemuxInterface::SetStreamCurrentFrame( int64_t frame )
{
	m_CurrentStreamFrame = m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes  + frame;
	if( m_CurrentStreamFrame < m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes )
		m_CurrentStreamFrame = m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes;
	if( m_CurrentStreamFrame > m_tyStream->index->real_end_gop_index->chunk_number_i_frame_pes )
		m_CurrentStreamFrame = m_tyStream->index->real_end_gop_index->chunk_number_i_frame_pes;

	set_GopIndexFromFrame();
	m_CurrentStreamTime = m_currentGop->display_time;
}

void CDemuxInterface::SetStreamCurrentFrameRelative( int64_t frameOffset )
{
	m_CurrentStreamFrame += frameOffset;
	if( m_CurrentStreamFrame < m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes )
		m_CurrentStreamFrame = m_tyStream->index->real_start_gop_index->chunk_number_i_frame_pes;
	if( m_CurrentStreamFrame > m_tyStream->index->real_end_gop_index->chunk_number_i_frame_pes )
		m_CurrentStreamFrame = m_tyStream->index->real_end_gop_index->chunk_number_i_frame_pes;
	set_GopIndexFromFrame();
	m_CurrentStreamTime = m_currentGop->display_time;
}

/* Cutpoint Management Functions */
CDemuxCutPoint CDemuxInterface::GetCurrentCut( CDemuxCutPoint::eCutType type )
{
	CDemuxCutPoint Cut;

	Cut.m_Type = type;
	Cut.m_Index = m_currentGop;
	Cut.time = GetStreamCurrentTime(false);
	Cut.ticks = GetStreamCurrentTime(true);
	Cut.frame = GetStreamCurrentFrame();
	return ( Cut );
}

bool CDemuxInterface::IsSystemStartCutNeeded()
{
	if( m_tyStream->index->gop_index == m_tyStream->index->real_start_gop_index ) {
		return false;
	} else {
		return true;
	}
}

CDemuxCutPoint CDemuxInterface::GetSystemStartCutStart()
{
	CDemuxCutPoint Cut;
	Cut.time = -1;
	Cut.ticks = -1;
	Cut.frame = -1;

	if( m_tyStream->index->gop_index == m_tyStream->index->real_start_gop_index ) {
		// No reset GOPs at the start than need cutting
		Cut.m_Type = CDemuxCutPoint::unknown;
		Cut.m_Index = NULL;
	} else {
		Cut.m_Type = CDemuxCutPoint::cutsystem;
		Cut.m_Index = m_tyStream->index->gop_index;
	}
	return ( Cut );
}

CDemuxCutPoint CDemuxInterface::GetSystemStartCutStop()
{
	CDemuxCutPoint Cut;
	Cut.time = -1;
	Cut.ticks = -1;
	Cut.frame = -1;

	if( m_tyStream->index->gop_index == m_tyStream->index->real_start_gop_index ) {
		// No reset GOPs at the start than need cutting
		Cut.m_Type = CDemuxCutPoint::unknown;
		Cut.m_Index = NULL;
	} else {
		/* This will not work if we have a rest - we need to add the cut
		point at ->next->next->next instead to be sure to stay clear */
		Cut.m_Type = CDemuxCutPoint::cutsystem;
		Cut.m_Index = m_tyStream->index->real_start_gop_index->next;
	}
	return ( Cut );
}

bool CDemuxInterface::IsSystemEndCutNeeded()
{
	if( m_tyStream->index->last_gop_index == m_tyStream->index->real_end_gop_index ) {
		return false;
	} else {
		return true;
	}
}

CDemuxCutPoint CDemuxInterface::GetSystemEndCutStart()
{
	CDemuxCutPoint Cut;
	Cut.time = -1;
	Cut.ticks = -1;
	Cut.frame = -1;

	if( m_tyStream->index->last_gop_index == m_tyStream->index->real_end_gop_index ) {
		// No reset GOPs at the start than need cutting
		Cut.m_Type = CDemuxCutPoint::unknown;
		Cut.m_Index = NULL;
	} else {
		/* This will not work if we have a rest - we need to add the cut
		at ->previous->previous->previous to be sure to stay clear */
		Cut.m_Type = CDemuxCutPoint::cutsystem;
		Cut.m_Index = m_tyStream->index->real_end_gop_index->previous;
	}
	return ( Cut );
}

CDemuxCutPoint CDemuxInterface::GetSystemEndCutStop()
{
	CDemuxCutPoint Cut;
	Cut.time = -1;
	Cut.ticks = -1;
	Cut.frame = -1;

	if( m_tyStream->index->last_gop_index == m_tyStream->index->real_end_gop_index ) {
		// No reset GOPs at the start than need cutting
		Cut.m_Type = CDemuxCutPoint::unknown;
		Cut.m_Index = NULL;
	} else {
		Cut.m_Type = CDemuxCutPoint::cutsystem;
		Cut.m_Index = m_tyStream->index->last_gop_index;
	}
	return ( Cut );
}

void CDemuxInterface::JumpToCut( CDemuxCutPoint Cut )
{
	assert( Cut.m_Index );
	setCurrentGOP( Cut.m_Index );
	m_CurrentStreamTime = m_currentGop->display_time;
}

void CDemuxInterface::SetupProcessor( CDemuxProcessTask *processor )
{
	if( m_pRemote ) {
		processor->SetRemoteFile( m_pRemote, m_fsid );
		tydemux_close_remote_tystream(m_pRemote);
		m_pRemote = NULL;  // We no longer have control over this
	}
}

bool CDemuxInterface::IsStreamRemote()
{
	if( m_tyStream->remote_holder ) {
		return true;
	}

	return false;
}

/* Background index creation task */
int indexer_thread( void *user_data )
{
	indexer_thread_params_t *params;

	params = (indexer_thread_params_t *) user_data;

	FILE *logfile = stderr;

	/* Initialise the logger for this thread */
	if( params->logfile ) {
		logfile = params->logfile;
	}
	logger_init( log_warn, logfile, NULL, 1 );

	strcpy( params->progress_text, "Indexing...");
	if( params->tyStream->remote_holder ) {
		params->tyStream->index = tydemux_index_remote(params->tyStream, params->tyStream->remote_holder,
			&params->progress, params->progress_text);
	} else {
		params->tyStream->index = create_index(params->tyStream, 0, &params->progress, params->progress_text );
	}

	logger_free();

	if( !params->tyStream->index )
		params->result = -1;

	params->progress = 100;
	params->complete = 1;
	params->result = 0;
	return 0;
}

CDemuxIndexTask::CDemuxIndexTask( tystream_holder_t *tyStream, FILE *logfile )
: m_th(0)
{
	m_Params.complete = 0;
	m_Params.result = 0;
	m_Params.progress = 0;
	m_Params.tyStream = tyStream;
	m_Params.logfile = logfile;

	strcpy( m_Params.progress_text, "Startup up");
}

CDemuxIndexTask::~CDemuxIndexTask()
{
	CancelJob();
}


bool CDemuxIndexTask::StartJob()
{
	assert( !m_th );
	assert( m_Params.tyStream );

	m_th = thread_start( indexer_thread, &m_Params );

	return true;
}

bool CDemuxIndexTask::CancelJob()
{
	if( m_th ) {
		thread_stop( m_th, -1 );
		m_th = 0;
		m_Params.result = -1;
		return true;
	}

	return false;
}

bool CDemuxIndexTask::IsCompleted()
{
	if( m_th && m_Params.complete ) {
		return true;
	} else {
		return false;
	}
}

int CDemuxIndexTask::GetResult()
{
	return( m_Params.result );
}

int  CDemuxIndexTask::GetProgress()
{
	return( m_Params.progress );
}

const char *CDemuxIndexTask::GetProgressText()
{
	return( m_Params.progress_text );
}


/* Background TY file copier task */
int copier_thread( void *user_data )
{
	tystream_holder_t * tystream = NULL;
	vstream_t * vstream = NULL;
	int64_t i;
	int video_file;

	copier_thread_params_t *params;
	params = (copier_thread_params_t *) user_data;
	params->complete = 0;
	params->result = -1;

	FILE *logfile = stderr;

	/* Initialise the logger for this thread */
	if( params->logfile ) {
		logfile = params->logfile;
	}
	logger_init( log_warn, logfile, NULL, 1 );

	tystream = tydemux_open_probe_remote(params->remote_holder, params->fsid);

	if(!tystream) {
		LOG_ERROR1( "Failed to do remote probe on fsid %d\n", params->fsid );
		params->result = -1;
		return 0;
	}

	video_file = open(params->dest, O_WRONLY|O_CREAT|O_TRUNC|OS_FLAGS, READWRITE_PERMISSIONS);

	if(video_file == -1) {
		LOG_ERROR1( "Failed to open video output file: %s\n", params->dest );
		free_tystream(tystream);
		params->result = -1;
		return 0;
	}

	for(i = 0; i < tystream->nr_of_remote_chunks; i++) {

		vstream = tydemux_get_remote_chunk(params->remote_holder, tystream, i);
		if(!vstream) {
			LOG_ERROR( "Error fetching remote chunk\n");
			free_vstream(vstream);
			vstream = NULL;
			continue;
		}

		write(video_file, vstream->start_stream, (unsigned int)(vstream->size));
		free_vstream(vstream);
		vstream = NULL;

		params->progress = (int)((i * 100)/tystream->nr_of_remote_chunks);
	}
	//tydemux_get_remote_stream(tystream, params->progress, params->remote_holder, video_file);

	close(video_file);

	free_tystream(tystream);

	logger_free();

	params->progress = 100;
	params->complete = 1;
	params->result = 0;

	return 0;
}

CDemuxCopyTask::CDemuxCopyTask( remote_holder_t *pRemote, int fsid, const char *dest, FILE *logfile )
: m_th(0)
{
	m_Params.complete = 0;
	m_Params.result = 0;
	m_Params.progress = 0;
	m_Params.remote_holder = pRemote;
	m_Params.fsid = fsid;
	m_Params.dest = dest;
	m_Params.logfile = logfile;
}

CDemuxCopyTask::~CDemuxCopyTask()
{
	CancelJob();
}

bool CDemuxCopyTask::StartJob()
{
	assert( !m_th );
	assert( m_Params.remote_holder );

	m_th = thread_start( copier_thread, &m_Params );

	return true;
}

bool CDemuxCopyTask::CancelJob()
{
	if( m_th ) {
		thread_stop( m_th, -1 );
		m_th = 0;
		m_Params.result = -1;
		return true;
	}

	return false;
}

bool CDemuxCopyTask::IsCompleted()
{
	if( m_th && m_Params.complete ) {
		return true;
	} else {
		return false;
	}
}

int CDemuxCopyTask::GetResult()
{
	return( m_Params.result );
}

int  CDemuxCopyTask::GetProgress()
{
	return( m_Params.progress );
}

