/*
 * Copyright (C) 2002, 2003  John Barrett (MrBassMan) <johnbarretthcc@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: tydemuxclass.h,v 1.16 2003/03/26 20:38:49 mrbassman Exp $ */

#ifndef __TYDEMUXCLASS_H__
#define __TYDEMUXCLASS_H__

#define LOG_MODULE "tyeditor"

#include "tycommon.h"
#include "tylogging.h"
#include "../tydemux/tydemux.h"
#include "processor.h"

#define TICKS_PER_MILLISECOND 90

/* Background task handlers */
class CDemuxIndexTask;
class CDemuxCopyTask;
class CDemuxProcessTask;

typedef uint8_t * ImagePtr_t;

extern void TimeAsString( int64_t ticks, char *timeString );
extern void FrameAsString( int64_t frame, char *frameString );

#define MAX_CUTSTRING_SIZE 256

class CDemuxCutPoint
{
public:
	CDemuxCutPoint();

	typedef enum { cutstart, cutstop, chapter, cutsystem, unknown } eCutType;

	/* Need at least MAX_CUTSTRING_SIZE characters */
	void GetCutAsString( char *CutString );
	void GetTimeAsString( char *TimeString );

	gop_index_t *m_Index;
	eCutType	m_Type;

	int64_t frame;
	int64_t time;
	int64_t ticks;

	CDemuxCutPoint & operator = (CDemuxCutPoint &src) {
		m_Index = src.m_Index; m_Type = src.m_Type;
		frame = src.frame; time = src.time; ticks = src.ticks;
		return *this;
	};
	bool operator == (CDemuxCutPoint &comp) {
		return( m_Index->chunk_number_i_frame_pes == comp.m_Index->chunk_number_i_frame_pes );
	};
	bool operator < (CDemuxCutPoint &comp) {
		return( m_Index->chunk_number_i_frame_pes < comp.m_Index->chunk_number_i_frame_pes );
	};
	bool operator > (CDemuxCutPoint &comp) {
		return( m_Index->chunk_number_i_frame_pes > comp.m_Index->chunk_number_i_frame_pes );
	};
};

class CTivoFile
{
public:
	
	int fsid;
	bool indexed;
	char tystream[256];
	int state;
	char year[25];
	char air[25];
	char day[25];
	char date[25];
	char rectime[25];
	char duration[25];
	char title[256];
	char episode[256];
	char episodenr[25];
	char description[256];

	CTivoFile *pNext;
};

class CDemuxInterface
{
	friend class CDemuxIndexTask;
	friend class CDemuxCopyTask;

public:

	CDemuxInterface();
	~CDemuxInterface();

	/* Load the ty file and get ready for a new editing session */
	bool LoadTY( const char *fileName, int TivoVersion = 2 );

	/* Background indexing task - use delete when you've finished with it */
	/* Do NOT delete the interface object while you still have tasks created */
	CDemuxIndexTask *CreateIndexer(FILE *logfile);
	CDemuxCopyTask *CreateCopier( int fsid, const char *dest, FILE *logfile );

	void SetupProcessor( CDemuxProcessTask *processor );

	/* Work with the currently stream time pointer in GOPS */
	void SetStreamGopRelative( int64_t GOPOffset );

	/* Work with the currently stream time pointer in milliseconds */
	int64_t GetStreamTime();
	int64_t GetStreamCurrentTime(bool noOffset = false);
	void SetStreamTime( const char *szTime, bool bSetTimeToExactGOP );
	void SetStreamTime( int64_t streamTime, bool bSetTimeToExactGOP, bool bNoOffset = false );
	void SetStreamTimeRelative( int64_t streamTimeOffset, bool bSetTimeToExactGOP );

	/* Work with the currently stream pointer in frames */
	int64_t GetStreamFrames();
	int64_t GetStreamCurrentFrame();
	void SetStreamCurrentFrame( int64_t frame );
	void SetStreamCurrentFrameRelative( int64_t frameOffset );
	void SetStreamCurrentFrameToEnd();

	bool GetCurrentImage( ImagePtr_t &pImage, int &size );
	int	 GetFrameRate();

	/* Cut point management */
	bool IsSystemStartCutNeeded();
	CDemuxCutPoint GetSystemStartCutStart();
	CDemuxCutPoint GetSystemStartCutStop();
	bool IsSystemEndCutNeeded();
	CDemuxCutPoint GetSystemEndCutStart();
	CDemuxCutPoint GetSystemEndCutStop();
	CDemuxCutPoint GetCurrentCut( CDemuxCutPoint::eCutType type );
	void JumpToCut( CDemuxCutPoint Cut );

	/* Static function to get a list of files on a remote TIVO */
	bool OpenRemoteTivoFile( int fsid );
	CTivoFile *GetTivoFileList( const char *tivohostname );
	bool IsStreamRemote();

protected:

private:
	index_t	*m_Index;
	remote_holder_t *m_pRemote;
	int m_fsid;
	tystream_holder_t *m_tyStream;
	module_t *m_init_image;
	module_t *m_last_image;
	gop_index_t *m_currentGop;
	int64_t m_CurrentStreamTime;
	int64_t m_CurrentStreamFrame;
	int64_t m_AvgTicksPerGOP;
	char m_tyFile[ MAXPATHLEN + 1];
	char m_tivoHost[ MAXPATHLEN + 1];

	void setCurrentGOP( gop_index_t *index );
	gop_index_t *get_CurrentGopIndex();
	void set_GopIndexFromTime();
	void set_GopIndexFromFrame();
	void closeRemote();

};

/* Background index creation task */
typedef struct {
	tystream_holder_t *tyStream;
	int progress;
	int result;
	int complete;
	FILE *logfile;
	char progress_text[256];
} indexer_thread_params_t;

class CDemuxIndexTask
{
public:
	CDemuxIndexTask( tystream_holder_t *tyStream, FILE *logfile );
	~CDemuxIndexTask();

	bool StartJob();
	bool IsCompleted();
	bool CancelJob();
	int  GetResult();
	int  GetProgress();  // 0 - 100 (percent)
	const char *GetProgressText();

protected:
	indexer_thread_params_t m_Params;

private:
	thread_handle_t m_th;
};

typedef struct {
	remote_holder_t *remote_holder;
	int progress;
	int result;
	int complete;
	int fsid;
	const char *dest;
	FILE *logfile;
} copier_thread_params_t;

class CDemuxCopyTask
{
public:
	CDemuxCopyTask( remote_holder_t *pRemote, int fsid, const char *dest, FILE *logfile );
	~CDemuxCopyTask();

	bool StartJob();
	bool IsCompleted();
	bool CancelJob();
	int  GetResult();
	int  GetProgress();  // 0 - 100 (percent)

protected:
	copier_thread_params_t m_Params;

private:
	thread_handle_t m_th;
};

#endif //__TYDEMUXCLASS_H__
