/*
 * Copyright (C) 2002  John Barrett <johnbarretthcc@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* "$Id: tymplex.cpp,v 1.19 2003/03/30 12:24:55 mrbassman Exp $" */

/* TyMplex - a wrapper program for a threaded version of the mjpeg mplex code */

#ifdef _WIN32
#include <windows.h>
#include <io.h>

#define sleep(s) Sleep(s * 1000)

#else  /* Unix */
#include <inttypes.h>
#include <unistd.h>

#endif

#include <tygetopt.h>

#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>

/* Our library functions */
#include "../libs/threadlib.h"
#define LOG_MODULE "mplex"
#include "tylogging.h"

/* Our own header files */
#include "tymplex.h"

/* mplex includes */
#include "mplexlib.h"

/* Our version of the mplex PS_Stream object
 * We support pipes as well as file based output */
class outputstream : public PS_Stream
{
public:
    outputstream( pipe_handle_t ph, unsigned mpeg, unsigned int sector_size,
                  off_t max_segment_size ) 
		: PS_Stream( mpeg, sector_size, "NULL", max_segment_size )
		, m_ph( ph )
		{};

	/* Over-rides for base functions to allow pipe output */
    virtual bool FileLimReached( );
    virtual void NextFile();
	virtual void RawWrite(uint8_t *data, unsigned int len);
    virtual void Close();
    virtual int  Open( );

private:
	pipe_handle_t m_ph;
};

/* Our version of the mplex MultiplexJob object
 * We support pipes as well as file based input/output */
class tymultiplexjob : public MultiplexJob
{
public:
	tymultiplexjob() :m_streamout(0) {};
	virtual ~tymultiplexjob();

	/* Over-ride some virtual functions in the MultiplexJob object */
	virtual void SetFromCmdLine(unsigned int argc, char *argv[]);
	virtual void Usage(char *str);
	virtual PS_Stream *GetOutputStream(
			   unsigned mpeg,
               unsigned int sector_size,
               const char *filename_pat,
               off_t max_segment_size // 0 = No Limit
			   );

	void SetAudioOffset( int offset );
	void add_video_input( pipe_handle_t ph );
	void add_audio_input( pipe_handle_t ph );
	void add_piped_output( pipe_handle_t ph );
	void set_video_params();

private:
	outputstream *m_streamout;
	pipe_handle_t m_ph_out;
};

/* Our version of the mplex IBitStream object that supports pipes */
class IPipeBitStream : public IBitStream
{
public:
 	IPipeBitStream( pipe_handle_t pipe_handle, 
					unsigned int buf_size = BUFFER_SIZE);

private:
	pipe_handle_t ph;
	virtual size_t ReadStreamBytes( uint8_t *buf, size_t number ) {
			//printf("tymplex - read pipe\n");
			return thread_pipe_read( ph, number, (char *) buf ); 
			//printf("tymplex - read pipe - done \n");
		}
	virtual bool EndOfStream() {
		return ( thread_pipe_iseof(ph) ? true : false );
	}
	
};

bool outputstream::FileLimReached()
{
	/* Pipe limit is never reached */
	return false;
}

void outputstream::NextFile( )
{
	/* Only one pipe - do nothing */
}

/* Implement our version of the mplex PS_Stream object */
void outputstream::RawWrite(uint8_t *data, unsigned int len)
{
	static int last_progress = -1;
	int writtenmb;

	pipe_stat_t bytes_read;
	pipe_stat_t bytes_written;

	if( thread_pipe_write( m_ph, len, (const char *) data ) != len ) {
        mjpeg_error_exit1( "Failed write to pipe" );
    }

	thread_pipe_stats( m_ph, &bytes_read, &bytes_written );
	writtenmb = bytes_written/1024000;
	if( writtenmb != last_progress ) {
		last_progress = writtenmb;
		LOG_PROGRESS1("Written: %dMB\r", writtenmb );
	}
}

int outputstream::Open()
{
	/* Do nothing - the pipe will already be open */
	return (0);
}

void outputstream::Close()
{
	/* Mark EOF on our output pipe */
	thread_pipe_eof( m_ph, 1 );
}

/* Implement our version of the mplex IBitStream object */
IPipeBitStream::IPipeBitStream( pipe_handle_t pipe_handle, unsigned int buf_size )
{
	ph = pipe_handle;

	eobs = false;
	byteidx = 0;

    SetBufSize(buf_size);

	if (!ReadIntoBuffer(buf_size))
	{
		if (buffered==0)
		{
			mjpeg_info( "Unable to read from pipe." );
		}
	}

}

tymultiplexjob::~tymultiplexjob()
{
	if( m_streamout )
		delete m_streamout;
}

PS_Stream *tymultiplexjob::GetOutputStream(
			   unsigned mpeg, unsigned int sector_size,
               const char *, off_t max_segment_size )
{
	assert( m_streamout == NULL );
	assert( m_ph_out );

	m_streamout = new outputstream( m_ph_out, mpeg, sector_size, max_segment_size );

	return( m_streamout );
}


/* Add a pipe to the list of outputs */
void tymultiplexjob::add_piped_output( pipe_handle_t ph )
{
	m_ph_out = ph;
}


/* Implement our version of the mplex MultiplexJob object */
/* Add a pipe to the list of inputs */
void tymultiplexjob::add_audio_input( pipe_handle_t ph )
{
    IPipeBitStream *bs;
    IBitStreamUndo undo;
    bool bad_pipe = true;
    
    bs = new IPipeBitStream( ph, 4 * 1024 );

    /* Remember the streams initial state */
    bs->PrepareUndo( undo);

    //printf("Create audio input\n");

	/* Start probing the stream for recognisable formats */
    if( MPAStream::Probe( *bs ) ) {
        LOG_USERDIAG ("Pipe looks like an MPEG Audio stream.\n" );
		bs->UndoChanges( undo );
		streams.push_back( new JobStream( bs, MPEG_AUDIO) );
        ++audio_tracks;
	    bad_pipe = false;
    }

    if( bad_pipe ) {
	    bs->UndoChanges( undo);
		if( AC3Stream::Probe( *bs ) ) {
			LOG_USERDIAG ("Pipe looks like an AC3 Audio stream.\n" );
			bs->UndoChanges( undo);
            streams.push_back( new JobStream( bs, AC3_AUDIO) );
	        ++audio_tracks;
			bad_pipe = false;
		}
    }

    if( bad_pipe ) {
	    bs->UndoChanges( undo);
		if( DTSStream::Probe( *bs ) ) {
			LOG_USERDIAG ("Pipe looks like a dts Audio stream.\n" );
			bs->UndoChanges( undo);
            streams.push_back( new JobStream( bs, DTS_AUDIO) );
	        ++audio_tracks;
			bad_pipe = false;
		}
    }

    if( bad_pipe ) {
	    bs->UndoChanges( undo);
		if( LPCMStream::Probe( *bs ) ) {
			LOG_USERDIAG ("Pipe looks like an LPCM Audio stream.\n" );
			bs->UndoChanges( undo);
            streams.push_back( new JobStream( bs,  LPCM_AUDIO) );
	        ++audio_tracks;
            ++lpcm_tracks;
			bad_pipe = false;
		}
    }

    if( bad_pipe )
    {
		delete bs;
		LOG_ERROR ("Pipe data is unrecogniseable!\n" );
    }
    //printf("Done creating audio input\n");

}


/* Implement our version of the mplex MultiplexJob object */
/* Add a pipe to the list of inputs */
void tymultiplexjob::add_video_input( pipe_handle_t ph )
{
    IPipeBitStream *bs;
    IBitStreamUndo undo;
    bool bad_pipe = true;
    
    bs = new IPipeBitStream( ph );

    /* Remember the streams initial state */
    bs->PrepareUndo( undo);

    if( bad_pipe ) {
//	    bs->UndoChanges( undo);
		if( VideoStream::Probe( *bs ) ) {
			LOG_USERDIAG ("Pipe looks like an MPEG Video stream.\n" );
			bs->UndoChanges( undo);
            streams.push_back( new JobStream( bs, MPEG_VIDEO) );
			bad_pipe = false;
			++video_tracks;
		}
    }

    if( bad_pipe ) {
		delete bs;
		LOG_ERROR ("Pipe data is unrecogniseable!\n" );
    }
}

void tymultiplexjob::set_video_params()
{
	unsigned int i;

	for( i = video_param.size(); i < video_tracks; ++i )
	{
		video_param.push_back(VideoParams::Default( mux_format ));
	}
	for( i = lpcm_param.size(); i < lpcm_tracks; ++i )
	{
		lpcm_param.push_back(LpcmParams::Default(mux_format));
	}

	//
	// Set standard values if the selected profile implies this...
	//
	for( i = 0; i <video_tracks; ++i )
	{
		if( video_param[i]->Force(mux_format) )
		{
			mjpeg_info( "Video stream %d: profile %d selected - ignoring non-standard options!", i, mux_format );
		}
	}

	mjpeg_info( "Found %d audio streams and %d video streams",
                audio_tracks,
				video_tracks
        );
}

void tymultiplexjob::SetAudioOffset( int offset )
{
	video_offset = offset;
	if( video_offset < 0 )
	{
		audio_offset = - video_offset;
		video_offset = 0;
	}
}

void tymultiplexjob::Usage(char *str)
{
    fprintf( stderr,
	"tyStudio tymplex; a port of mjpegtools mplex-2 version " VERSION "\n"
	"Usage: %s [params] -o <output filename pattern> <input file>... \n"
	"         %%d in the output file name is by segment count\n"
	"  where possible params are:\n"
	"--verbose|-v num\n"
    "  Level of verbosity. 0 = quiet, 1 = normal 2 = verbose/debug\n"
	"--format|-f fmt\n"
    "  Set defaults for particular MPEG profiles\n"
	"  [0 = Generic MPEG1, 1 = VCD, 2 = user-rate VCD, 3 = Generic MPEG2,\n"
    "   4 = SVCD, 5 = user-rate SVCD\n"
	"   6 = VCD Stills, 7 = SVCD Stills, 8 = DVD with NAV sectors, 9 = DVD]\n"
    "--mux-bitrate|-r num\n"
    "  Specify data rate of output stream in kbit/sec\n"
	"    (default 0=Compute from source streams)\n"
	"--video-buffer|-b num [, num...] \n"
    "  Specifies decoder buffers size in kB.  [ 20...2000]\n"
    "--lpcm-params | -L samppersec:chan:bits [, samppersec:chan:bits]\n"
	"--mux-limit|-l num\n"
    "  Multiplex only num seconds of material (default 0=multiplex all)\n"
	"--sync-offset|-O num\n"
    "  Specify offset of timestamps (video-audio) in mSec\n"
	"--sector-size|-s num\n"
    "  Specify sector size in bytes for generic formats [256..16384]\n"
    "--vbr|-V\n"
    "  Multiplex variable bit-rate video\n"
	"--packets-per-pack|-p num\n"
    "  Number of packets per pack generic formats [1..100]\n"
	"--system-headers|-h\n"
    "  Create System header in every pack in generic formats\n"
	"--max-segment-size|-S size\n"
    "  Maximum size of output file(s) in Mbyte (default: 2000) (0 = no limit)\n"
	"--split-segment|-M\n"
    "  Simply split a sequence across files rather than building run-out/run-in\n"
	"--help|-?\n"
    "  Print this lot out!\n", str);

	thread_exit (1);
}

static const char short_options[] = "o:b:r:O:v:m:f:l:s:S:q:p:L:VXMeh";

#if defined(HAVE_GETLONG)
static struct option long_options[] = 
{
     { "verbose",           1, 0, 'v' },
     { "format",            1, 0, 'f' },
     { "mux-bitrate",       1, 0, 'r' },
     { "video-buffer",      1, 0, 'b' },
     { "lpcm-params",       1, 0, 'L' },
     { "output",            1, 0, 'o' },
     { "sync-offset",    	1, 0, 'O' },
     { "vbr",      	        1, 0, 'V' },
     { "system-headers",    1, 0, 'h' },
     { "split-segment",     0, 0, 'M' },
     { "max-segment-size",  1, 0, 'S' },
     { "mux-upto",   	    1, 0, 'l' },
     { "packets-per-pack",  1, 0, 'p' },
     { "sector-size",       1, 0, 's' },
     { "help",              0, 0, '?' },
     { 0,                   0, 0, 0   }
 };
#endif

void tymultiplexjob::SetFromCmdLine(unsigned int argc, char *argv[])
{
    int n;
    outfile_pattern = NULL;

	tygetopt_lock();

#if defined(HAVE_GETLONG)
	while( (n=getlong(argc,argv,short_options,long_options, NULL)) != -1 )
#else
    while( (n=getopt(argc,argv,short_options)) != -1 )
#endif
	{
		switch(n)
		{
        case 0 :
            break;
		case 'o' :
			outfile_pattern = optarg;
			break;
		case 'm' :
			mpeg = atoi(optarg);
			if( mpeg < 1 || mpeg > 2 ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}

			break;
		case 'v' :
			verbose = atoi(optarg);
			if( verbose < 0 || verbose > 2 ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			break;

		case 'V' :
			VBR = true;
			break;
	  
		case 'h' :
			always_system_headers = true;
			break;

		case 'b' :
            if( ! ParseVideoOpt( optarg ) )
            {
                mjpeg_error( "Illegal video decoder buffer size(s): %s", 
                             optarg );
				tygetopt_unlock();
                Usage(argv[0]);
            }
            break;
        case 'L':
            if( ! ParseLpcmOpt( optarg ) )
            {
                mjpeg_error( "Illegal LPCM option(s): %s", optarg );
				tygetopt_unlock();
                Usage(argv[0]);
            }
            break;

		case 'r':
			data_rate = atoi(optarg);
			if( data_rate < 0 ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			/* Convert from kbit/sec (user spec) to 50B/sec units... */
			data_rate = (( data_rate * 1000 / 8 + 49) / 50 ) * 50;
			break;

		case 'O':
			video_offset = atoi(optarg);
			if( video_offset < 0 )
			{
				audio_offset = - video_offset;
				video_offset = 0;
			}
			break;
          
		case 'l' : 
 			max_PTS = atoi(optarg);
			if( max_PTS < 1  ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			break;
		
		case 'p' : 
			packets_per_pack = atoi(optarg);
			if( packets_per_pack < 1 || packets_per_pack > 100  ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			break;
		
	  
		case 'f' :
			mux_format = atoi(optarg);
			if( mux_format < MPEG_FORMAT_MPEG1 || mux_format > MPEG_FORMAT_LAST ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			break;
		case 's' :
			sector_size = atoi(optarg);
			if( sector_size < 256 || sector_size > 16384 ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			break;
		case 'S' :
			max_segment_size = atoi(optarg);
			if( max_segment_size < 0  ) {
				tygetopt_unlock();
				Usage(argv[0]);
			}
			max_segment_size *= 1024*1024; 
			break;
		case 'M' :
			multifile_segment = true;
			break;
		case 'e' :
			emul_vcdmplex = 1;
			break;
		case '?' :
		default :
			tygetopt_unlock();
			Usage(argv[0]);
			break;
		}
	}

	tygetopt_unlock();

	(void)mjpeg_default_handler_verbosity(verbose);
	mjpeg_info( "mplex version %s (%s)",MPLEX_VER,MPLEX_DATE );

	(void)mjpeg_default_handler_verbosity(verbose);
}

void our_mjpeg_log_handler(log_level_t level, const char message[])
{
	switch(level) {
		case 0:
			/* No Logging */
		break;
		case 1:
			LOG_DEVDIAG1( "%s\n", message);
		break;
		case 2:
			LOG_USERDIAG1( "%s\n", message);
		break;
		case 3:
			LOG_WARN1( "%s\n", message);
		break;
		case 4:
			LOG_ERROR1( "%s\n", message);
		break;
		default:
			LOG_FATAL( message );
	}
}

static mutex_handle_t process_lock = 0;
static pipe_handle_t write_pipe = 0;

/* Failsafe to ensure the caller is not permanently blocked if we exit prematurely */
void mplex_exit_handler( void )
{
	if( write_pipe ) {
		LOG_FATAL("mplex finished prematurely and did not exit gracefully\n");
		thread_pipe_eof( write_pipe, 0 );
		write_pipe = 0;
	}
	if( process_lock ) {
		thread_mutex_unlock( process_lock );
		process_lock = 0;
	}
}

/* Entry point for the mplex thread */
int mplex_thread( void *user_data )
{
	mplex_start_params_t *params;
	tymultiplexjob job;
	FILE *logfile = stderr;

	atexit( mplex_exit_handler );

	if( !user_data )
		return -1;   /* We need some parameters, something has gone horribly wrong  */
	params = ( mplex_start_params_t *) user_data;
	write_pipe = params->output_pipe;
	//printf("In tymplex\n");

	/* Initialise the logger */
	if( params->logfile ) {
		logfile = params->logfile;
	}
	logger_init( params->loglevel, logfile, params->loghookfn, 1 );

	/* Switch mjpeg logging to our handler */
	/* And map our log levels onto those used by mjpeg */
	mjpeg_log_set_handler( our_mjpeg_log_handler );
	switch( logger_get_log_level() ) {
	case log_fatal:
		mjpeg_default_handler_verbosity(0);
		break;
	case log_error:
		mjpeg_default_handler_verbosity(4);
		break;
	case log_warn:
		mjpeg_default_handler_verbosity(3);
		break;
	case log_progress:
	case log_info:
		mjpeg_default_handler_verbosity(2);
		break;
	case log_function:
	case log_diag:
	case log_data:
		mjpeg_default_handler_verbosity(1);
		break;
	}

	/* Process mplex commands */
	job.SetFromCmdLine(params->argc, params->argv);

	/* Mux format over-ride */
	if( params->format ) {
		job.mux_format = params->format - 1;
	}

	/* Block waiting for demux to calculate the audio offset */
	if( params->wait_for_audio_offset ) {
		while( params->wait_for_audio_offset ) {
			sleep(0);  // Wait for demux to work out the audio offset
		}
		job.SetAudioOffset( params->audio_offset );
	}

	/* Add our pipe for the output stream */
	if( !params->output_pipe ) {
		LOG_FATAL( "Output pipe not specified\n" );
	}

	job.add_piped_output( params->output_pipe );

	/* Add our pipes to the list of inputs */
	//printf("Add video\n");
	job.add_video_input( params->video_pipe );
	//printf("Done add video\n");
	//printf("Add audio\n");
	job.add_audio_input( params->audio_pipe );
	//printf("Done audio\n");
	job.set_video_params();

	if( params->thread_lock ) {
		/* Lock the main process while we are running */
		thread_mutex_lock( params->thread_lock );
	}
	process_lock = params->thread_lock;

	/* Tell the main program we are running and that it should block itself on the thread_lock mutex */
	params->running = true;
	Multiplexor mux(job);
	mux.Multiplex();

	/* Mark EOF on our output pipe */
	thread_pipe_eof( params->output_pipe, 1 );
	write_pipe = 0;

	logger_free();

	if( params->thread_lock ) {
		/* All done so release the main process */
		thread_mutex_unlock( params->thread_lock );
	}
	process_lock = 0;

	return (0); 
}

