/*
 * Copyright (C) 2002  John Barrett <johnbarretthcc@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* "$Id: main.cpp,v 1.15 2003/03/29 21:44:16 mrbassman Exp $" */

/* TyProcess - Main control process to process command line args
 * and fire off the library version of typrocess */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

/* Logging */
#define LOG_MODULE "typrocess"
#include "tylogging.h"

#include "tycommon.h"
#include "../libs/about.c"

#include "typrocess.h"

#define TYPROCESS_VERSION "0.5.0"

static char demux_progress[ 26 ];
static char transcode_progress[ 26 ];
static char mplex_progress[ 26 ];
static time_t last_progress_time = 0;

static const char *default_mplex_args[] = 
{
	"typrocess",
	"-V"
};

int logger_hook(int level, const char *module, const char *msg )
{
	if( level == log_progress ) {
		/* do not output progress messages to the log file, used stderr instead */

		if( !last_progress_time ) {
			/* First progress message so initialise everything */
			strcpy( demux_progress, "Not Running");
			strcpy( transcode_progress, "Not Running");
			strcpy( mplex_progress, "Not Running");
			fprintf( stderr,"demux                    transcode                mplex\n" );
		}

		/* Check the module name and copy to the appropriate buffer */
		if( strcmp( module, "demux" ) == 0 ) {
			strncpy( demux_progress, msg, 25 );
			/* trim trailing whitespace */
			while( isspace( demux_progress[ strlen(demux_progress) - 1] ) ) {
				demux_progress[ strlen(demux_progress) - 1] = 0;
			}
		} else if( strcmp( module, "transcode" ) == 0 ) {
			strncpy( transcode_progress, msg, 25 );
			/* trim trailing whitespace */
			while( isspace( transcode_progress[ strlen(transcode_progress) - 1] ) ) {
				transcode_progress[ strlen(transcode_progress) - 1] = 0;
			}
		} else if( strcmp( module, "mplex" ) == 0 ) {
			strncpy( mplex_progress, msg, 25 );
			/* trim trailing whitespace */
			while( isspace( mplex_progress[ strlen(mplex_progress) - 1] ) ) {
				mplex_progress[ strlen(mplex_progress) - 1] = 0;
			}
		}

		/* Once a second, update the screen */
		if( last_progress_time < time(0) ) {
			fprintf( stderr, "%-25.25s%-25.25s%-25.25s\r", demux_progress, transcode_progress, mplex_progress );
			last_progress_time = time(0);
		}

		return 0;
	}

	return 1;
}

void Usage(char *str)
{
    fprintf( stderr,
	"typrocess command usage: %s [params]\n"
	"  Possible commands are:\n"
	"-mplex [{ mplex params }]\n"
	"-tcode [{ transcode params }]\n"
	"-demux [{ transcode params }]\n"
	"-d	<loglevel>\n"
	"-l	<log file>\n"
	"  Log to file. By default, all logging goes to stderr\n"
	"-la <log file>\n"
	"  As above but the log file is appended to\n"
	"-pipe\n"
	"  Input and output from pipes, implies -demux and -mplex\n"
	"-i	<input ty file>\n"
	"-o <output mpeg file>\n"
	"-it <input ty file>\n"
	"-ia <input audio file>\n"
	"-iv <input video file>\n"
	"-oa <output audio file>\n"
	"-ov <output video file>\n"
	"-om <output mpeg file>\n", str );
   fprintf(stderr,
	"typrocess "TYPROCESS_VERSION" %s", about);
	exit( -1 );
}

/* Main program entry point
 * analyses the command line and fires off the typrocess */
int main( int argc, char **argv )
{
	typrocess_params_t params;
	int i, j;
	int errorcode = 0;
	char buf[15];
	char *logfile = NULL;
	int logappend = 0;

	memset( &params, 0, sizeof( typrocess_params_t ) );

	if( argc == 1 ) {
		Usage(argv[0]);
	}

	/* Set up default command line parameters */
	params.mplex_params.argc = sizeof( default_mplex_args ) / sizeof( char *);
	params.mplex_params.argv = (char **) default_mplex_args;

	for( i = 1; i < argc; ++i ) {
		if( strcmp( argv[i], "-demux" ) == 0 ) {
			params.demux = 1;

			if( i+1 < argc && strcmp( argv[i+1], "{" ) == 0 ) {

				/* Arguments for demux */
				++i;

				while( strcmp( argv[i + params.demux_params.argc], "}" ) != 0 && i < argc ) {
					++params.demux_params.argc;
				}
				/* Make sure we saw a close parenthesis */
				if( strcmp( argv[i + params.demux_params.argc], "}" ) != 0 ) {
					Usage(argv[0]);
				}

				params.demux_params.argv = (char **) malloc( sizeof( char *) * (params.demux_params.argc+1) );
				params.demux_params.argv[0] = strdup( "-demux" );
				++params.demux_params.argc;

				/* Copy the demux args */
				for( j = 1; j < params.demux_params.argc; ++j ) {
					params.demux_params.argv[j] = strdup( argv[i++] );
				}
			}
		} else if( strcmp( argv[i], "-mplex" ) == 0 ) {
			params.mplex = 1;
			if( i+1 < argc && strcmp( argv[i+1], "{" ) == 0 ) {

				/* Arguments for mplex */
				++i;
				params.mplex_params.argc = 0;

				while( strcmp( argv[i + params.mplex_params.argc], "}" ) != 0 && i < argc ) {
					++params.mplex_params.argc;
				}
				/* Make sure we saw a close parenthesis */
				if( strcmp( argv[i + params.mplex_params.argc], "}" ) != 0 ) {
					Usage(argv[0]);
				}

				params.mplex_params.argv = (char **) malloc( sizeof( char *) * (params.mplex_params.argc+1) );
				params.mplex_params.argv[0] = strdup( "-mplex" );
				++params.mplex_params.argc;

				/* Copy the mplex args */
				for( j = 1; j < params.mplex_params.argc; ++j ) {
					params.mplex_params.argv[j] = strdup( argv[i++] );
				}
			}
		} else if( strcmp( argv[i], "-tcode" ) == 0 || strcmp( argv[i], "-transcode" ) == 0 ) {
			params.transcode = 1;
			if( i+1 < argc && strcmp( argv[i+1], "{" ) == 0 ) {

				/* Arguments for Transcode */
				++i;

				while( strcmp( argv[i + params.transcode_params.argc], "}" ) != 0 && i < argc ) {
					++params.transcode_params.argc;
				}
				/* Make sure we saw a close parenthesis */
				if( strcmp( argv[i + params.transcode_params.argc], "}" ) != 0 ) {
					Usage(argv[0]);
				}

				params.transcode_params.argv = (char **) malloc( sizeof( char *) * (params.transcode_params.argc+8) );
				params.transcode_params.argv[0] = strdup( "-tcode" );
				++params.transcode_params.argc;
        sprintf( buf, "-%c", params.transcode_params.format ? 'a' : 'm' );
				params.transcode_params.argv[1] = strdup( buf );
				++params.transcode_params.argc;
				params.transcode_params.argv[2] = strdup( "-s" );
				++params.transcode_params.argc;
        sprintf( buf, "%d", params.transcode_params.rate );
				params.transcode_params.argv[3] = strdup( buf );
				++params.transcode_params.argc;
				params.transcode_params.argv[4] = strdup( "-b" );
				++params.transcode_params.argc;
        sprintf( buf, "%d", params.transcode_params.bitrate );
				params.transcode_params.argv[5] = strdup( buf );
				++params.transcode_params.argc;
				params.transcode_params.argv[6] = strdup( "-l" );
				++params.transcode_params.argc;
        sprintf( buf, "%d", params.transcode_params.dbadjust );
				params.transcode_params.argv[7] = strdup( buf );
				++params.transcode_params.argc;

				/* Copy the transcode args */
				for( j = 6; j < params.transcode_params.argc; ++j ) {
					params.transcode_params.argv[j] = strdup( argv[i++] );
				}
			}
		} else if( strcmp( argv[i], "-pipe" ) == 0 ) {
			params.use_pipes = 1;
			params.mplex = 1;
			params.demux = 1;
		} else if( strcmp( argv[i], "-d" ) == 0 ) {
			params.loglevel = (loglevel_t)atoi( argv[++i] );
		} else if( strcmp( argv[i], "-l" ) == 0 ) {
			logfile = argv[++i];
			logappend = 0;
		} else if( strcmp( argv[i], "-la" ) == 0 ) {
			logfile = argv[++i];
			logappend = 1;
		} else if( strcmp( argv[i], "-i" ) == 0 ) {
			strcpy( params.in_ty_file, argv[++i] );
		} else if( strcmp( argv[i], "-it" ) == 0 ) {
			strcpy( params.in_ty_file, argv[++i] );
		} else if( strcmp( argv[i], "-iv" ) == 0 ) {
			strcpy( params.in_video_file, argv[++i] );
		} else if( strcmp( argv[i], "-ia" ) == 0 ) {
			strcpy( params.in_audio_file, argv[++i] );
		} else if( strcmp( argv[i], "-o" ) == 0 ) {
			strcpy( params.out_mpeg_file, argv[++i] );
		} else if( strcmp( argv[i], "-om" ) == 0 ) {
			strcpy( params.out_mpeg_file, argv[++i] );
		} else if( strcmp( argv[i], "-oa" ) == 0 ) {
			strcpy( params.out_audio_file, argv[++i] );
		} else if( strcmp( argv[i], "-ov" ) == 0 ) {
			strcpy( params.out_video_file, argv[++i] );
		} else {
			Usage(argv[0]);
		}
	}

	/* Redirect progress logging */
	params.loghookfn = logger_hook;

	/* Initialise the logger */
	if( logfile && logfile[0] ) {
		/* Create the log file in append or over-write mode */
		params.logfile = fsopen( logfile, logappend ? APPEND_MODE : WRITE_MODE, SHARE_DENY_NONE );

		if( !params.logfile ) {
			fprintf( stderr, "Failed to create the log file: %s, Error %d\n", logfile, errno );
			return ( -1 );
		}
	}
	else
	{
		params.logfile = stderr;
	}

	if( logger_init( (loglevel_t) params.loglevel, params.logfile, params.loghookfn, 0 ) ) {
		fprintf( stderr, "Failed to initialise the logger\n" );
		return ( -1 );
	}

	/* Run the process */
	errorcode = typrocess( &params );

	/* Free memory used to passed command line args to the modules */
	if( params.demux_params.argv ) {
		for( j = 0; j < params.demux_params.argc; ++j ) {
			free ( params.demux_params.argv[j] );
		}
		free ( params.demux_params.argv );
	}
	if( params.transcode_params.argv ) {
		for( j = 0; j < params.transcode_params.argc; ++j ) {
			free ( params.transcode_params.argv[j] );
		}
		free ( params.transcode_params.argv );
	}
	if( params.mplex_params.argv ) {
		for( j = 0; j < params.mplex_params.argc; ++j ) {
			free ( params.mplex_params.argv[j] );
		}
		free ( params.mplex_params.argv );
	}

	if( errorcode == 1 ) {
		Usage(argv[0]);
	}

	logger_free();

	if( params.logfile && params.logfile != stderr ) {
		fclose( params.logfile );
	}

	return errorcode;
}
