/*
 * Copyright (C) 2002  John Barrett <johnbarretthcc@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* "$Id: typrocess.cpp,v 1.32 2003/03/30 02:12:23 mrbassman Exp $" */

/* TyProcess - Main control process to process command line args
 * create inter-module threads and fire off the modules on threads */

#include "tycommon.h"
#include "tymalloc.h"

/* Logging */
#define LOG_MODULE "typrocess"
#include "tylogging.h"

/* Our threading and pipe library */
#include "../libs/threadlib.h"

/* Header files for each module's parameters */
#include "../tymplex/tymplex.h"
#include "../tytranscode/tytranscode.h"
#include "../tydemux/tydemux.h"
#include "typrocess.h"

#define DEFAULT_PIPE_SIZE (100 * CHUNK_SIZE)

/* Main program entry point */
int typrocess( typrocess_params_t *params )
{
	int errorcode = 0;
 	pipe_handle_t pipes2free[15];
	int npipes = 0;

	mutex_handle_t main_process_lock = 0;

	demux_start_params_t demux_params;
	transcode_start_params_t transcode_params;
	mplex_start_params_t mplex_params;

	thread_handle_t mplex_th = 0;
	thread_handle_t transcode_th = 0;
	thread_handle_t demux_th = 0;

	memset( &demux_params, 0, sizeof(demux_start_params_t ) );
	memset( &transcode_params, 0, sizeof(transcode_start_params_t ) );
	memset( &mplex_params, 0, sizeof(mplex_start_params_t ) );

	/* Create a mutex to lock the main process until the last module has finished */
	main_process_lock = thread_mutex_create();

	/* Set up all the pipes and parameters for the modules */

	if( params->demux ) {
		/* Create a thread for the demux job to run in */

		/* Set up demux_params */
		demux_params.running = 0;
		demux_params.argc = params->demux_params.argc;
		demux_params.argv = params->demux_params.argv;
		demux_params.progress = params->progress;
		demux_params.loghookfn = params->loghookfn;
		demux_params.loglevel = params->loglevel;
		demux_params.logfile = params->logfile;

		/* Note the reliance on the cutlist structures being in the same format!!!*/
		demux_params.cutlist = params->demux_params.cutlist;
		demux_params.cutcount = params->demux_params.cutcount;


		if( !params->transcode && !params->mplex ) {
			/* Last module locks the main process until it has finished */
			demux_params.thread_lock = main_process_lock;
		}

		if( params->demux_params.remote_holder && params->demux_params.fsid ) {
			/* Remote TIVO file */
			demux_params.remote_holder = (remote_holder_t *)params->demux_params.remote_holder;
			demux_params.fsid = params->demux_params.fsid;
		}

		if ( !demux_params.input_pipe && !demux_params.remote_holder) {  /* Probe created either of them */
			if( params->use_pipes ) {
				/* Get our input from stdin */
				demux_params.input_pipe = thread_pipe_create_with_handle( stdin );
				demux_params.remote_holder = NULL;
				pipes2free[npipes++] = demux_params.input_pipe;
			} else {
				/* Get our input from a file */
				if( !params->in_ty_file[0] ) {
					return 1;
				}
				demux_params.input_pipe = thread_pipe_create_from_file( params->in_ty_file );
				demux_params.remote_holder = NULL;
				pipes2free[npipes++] = demux_params.input_pipe;
			}

		}


		if( !params->transcode && !params->mplex ) {
			/* Create an audio output pipe to a file */
			if( !params->out_audio_file[0] ) {
				return ( 1 );
			}
			demux_params.audio_pipe = thread_pipe_create_to_file( params->out_audio_file, 1 );
			pipes2free[npipes++] = demux_params.audio_pipe;
		} else {
			//printf("typrocess - creating audio pipe\n");
			/* Create an ordinary pipe for audio output */
			demux_params.audio_pipe = thread_pipe_create( DEFAULT_PIPE_SIZE );
			pipes2free[npipes++] = demux_params.audio_pipe;
		}

		if( !params->mplex ) {
			/* Create a video output pipe to a file */
			if( !params->out_video_file[0] ) {
				return ( 1 );
			}
			demux_params.video_pipe = thread_pipe_create_to_file( params->out_video_file, 1 );
			pipes2free[npipes++] = demux_params.video_pipe;
		} else {
			/* Create an ordinary pipe for video output */
			demux_params.video_pipe = thread_pipe_create( DEFAULT_PIPE_SIZE );
			pipes2free[npipes++] = demux_params.video_pipe;

			/* Link demux and mplex to allow mplex to be blocked waiting for the audio offset */
			demux_params.wait_for_audio_offset = &(mplex_params.wait_for_audio_offset);
			demux_params.audio_offset = &(mplex_params.audio_offset);
			mplex_params.wait_for_audio_offset = 1;
		}
	}

	if( params->transcode ) {

		transcode_params.running = 0;
		transcode_params.argc = params->transcode_params.argc;
		transcode_params.argv = params->transcode_params.argv;
		transcode_params.format = params->transcode_params.format;
		transcode_params.rate = params->transcode_params.rate;
		transcode_params.bitrate = params->transcode_params.bitrate;
		transcode_params.dbadjust = params->transcode_params.dbadjust;

		transcode_params.loghookfn = params->loghookfn;
		transcode_params.loglevel = params->loglevel;
		transcode_params.logfile = params->logfile;

		//printf("typrocess - creating audio transcode\n");

		if( !params->mplex ) {
			/* Last module locks the main process until it has finished */
			transcode_params.thread_lock = main_process_lock;
		}

		if( !params->demux ) {
			/* If no demux, get our input from a file */
			if( !params->in_audio_file[0] ) {
				return ( 1 );
			}
			transcode_params.input_pipe = thread_pipe_create_from_file( params->in_audio_file );
			pipes2free[npipes++] = transcode_params.input_pipe;
		} else {
			/* Use demux's audio output pipe as our input */
			transcode_params.input_pipe = demux_params.audio_pipe;
		}

		if( !params->mplex ) {
			/* If no mplex, put our output to a file */
			if( !params->out_audio_file[0] ) {
				return ( 1 );
			}
			transcode_params.output_pipe = thread_pipe_create_to_file( params->out_audio_file, 1 );
			pipes2free[npipes++] = transcode_params.output_pipe;
		} else {
			/* Create an ordinary pipe for audio output */
			transcode_params.output_pipe = thread_pipe_create( DEFAULT_PIPE_SIZE );
			pipes2free[npipes++] = transcode_params.output_pipe;
		}
	}

	if( params->mplex ) {
		mplex_params.running = 0;
		mplex_params.thread_lock = main_process_lock;
		mplex_params.argc = params->mplex_params.argc;
		mplex_params.argv = params->mplex_params.argv;
		mplex_params.format = params->mplex_params.format;

		mplex_params.loghookfn = params->loghookfn;
		mplex_params.loglevel = params->loglevel;
		mplex_params.logfile = params->logfile;

		if ( params->demux ) {
			/* Get our video input from demux's video output pipe */
			mplex_params.video_pipe = demux_params.video_pipe;
		} else {
			/* Get our video input from a file */
			if( !params->in_video_file[0] ) {
				return ( 1 );
			}
			mplex_params.video_pipe = thread_pipe_create_from_file( params->in_video_file );
			pipes2free[npipes++] = mplex_params.video_pipe;
		}

		if( params->transcode ) {
			/* Get our audio input from transcode's output pipe */
			mplex_params.audio_pipe = transcode_params.output_pipe;
		} else if ( params->demux ) {
			/* Get our audio input from demux's audio output pipe */
			mplex_params.audio_pipe = demux_params.audio_pipe;
		} else {

			/* Get our audio input from a file */
			if( !params->in_audio_file[0] ) {
				return ( 1 );
			}
			mplex_params.audio_pipe = thread_pipe_create_from_file( params->in_audio_file );
			pipes2free[npipes++] = mplex_params.audio_pipe;
		}

		if( params->use_pipes ) {
			/* Put our output to stdout */
			mplex_params.output_pipe = thread_pipe_create_with_handle( stdout );
			pipes2free[npipes++] = mplex_params.output_pipe;
		} else {
			/* Set our output to a file */
			if( !params->out_mpeg_file[0] ) {
				return ( 1 );
			}
			mplex_params.output_pipe = thread_pipe_create_to_file( params->out_mpeg_file, 1 );
			pipes2free[npipes++] = mplex_params.output_pipe;
		}

	}

	/* OK everything has been set up, now we can start the threads */
	if( !errorcode && params->demux ) {
		demux_th = thread_start(tydemux_thread, &demux_params );
		if( !demux_th ) {
			LOG_ERROR( "Failed to start demux on a thread." );
			errorcode = -1;
		}
	}
	if( !errorcode && params->transcode ) {
		transcode_th = thread_start( transcode_thread, &transcode_params );
		if( !transcode_th ) {
			LOG_ERROR( "Failed to start transcode on a thread." );
			errorcode = -1;
		}
	}
	if( !errorcode && params->mplex ) {
		mplex_th = thread_start(mplex_thread, &mplex_params );
		if( !mplex_th ) {
			LOG_ERROR( "Failed to start mplex on a thread." );
			errorcode = -1;
		}
	}

	if( !errorcode ) {

		/* Wait for the demux thread to start up */
		while( params->demux && !(demux_params.running) )
			sleep(0);
		/* Wait for the transcode thread to start up */
		while( params->transcode && !(transcode_params.running) )
			sleep(0);
		/* Wait for the mplex thread to start up */
		while( params->mplex && !(mplex_params.running) )
			sleep(0);

		/* Now it's all running, the last module will have grabbed the thread lock, */
		/* we will wait for the thread lock to be released.*/
		/* This is more efficient that doing the if(check) sleep() that we did above */
		thread_mutex_lock( main_process_lock );

		/* To get here we must have been given the mutex, therefore the last module */
		/* must have finished - release the mutex again */
		thread_mutex_unlock( main_process_lock );
	}

	/* Make sure all the threads are dead */
	if( demux_th ) {
		thread_stop( demux_th, -1 );
	}
	if( transcode_th ) {
		thread_stop( transcode_th, -1 );
	}
	if( mplex_th ) {
		thread_stop( mplex_th, -1 );
	}

	if( params->demux_params.remote_holder ) {
		tydemux_close_tyserver((remote_holder_t *)(params->demux_params.remote_holder) );
		free_remote_holder((remote_holder_t *)(params->demux_params.remote_holder));
	}

	/* Free all the pipes */
	do {
 		thread_pipe_free( pipes2free[--npipes] );
	} while( npipes );

	thread_mutex_free(main_process_lock);

	return 0;
}
