/*
 * Copyright (C) 2002  Rowan McFarland <tytranscode@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "audio_convert_structs.h"
#include "tytranscode.h"

// globals
const int32_t m_SampleRates[3] = { 44100, 48000, 32000 };
const int32_t m_MpegBitRates[5][16] = {
  {-1, 32000, 64000, 96000, 128000, 160000, 192000, 224000, 256000, 288000, 320000, 352000, 384000, 416000, 448000, -1}, // V1, L1
  {-1, 32000, 48000, 56000,  64000,  80000,  96000, 112000, 128000, 160000, 192000, 224000, 256000, 320000, 384000, -1}, // V1, L2
  {-1, 32000, 40000, 48000,  56000,  64000,  80000,  96000, 112000, 128000, 160000, 192000, 224000, 256000, 320000, -1}, // V1, L3
  {-1, 32000, 48000, 56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 176000, 192000, 224000, 256000, -1}, // V2, L1
  {-1,  8000, 16000, 24000,  32000,  40000,  48000,  56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, -1}  // V2, L2 & L3
};
const int     m_MpegChannels[4] = { 2, 2, 2, 1 };
const int     m_AC3EncChannels[6]  = { 1, 2, 3, 4, 5, 6 };
const int     m_AC3DecChannels[8]  = { 2, 1, 2, 3, 3, 4, 4, 5 };
const int32_t m_AC3BitRates[]   = {
32000,  40000,  48000,  56000,  64000,  80000,  96000, 112000,
128000, 160000, 192000, 224000, 256000, 320000, 384000, 448000,
512000, 576000, 640000
};

// external data
extern int8_t   m_RawOutput;
extern int32_t  m_RawInputSampleRate;

// local functions
int  FindFrame( PAUDIO_CONVERT_TYPE audioConvert );

// external functions

// ---------------------------------------------------------------------
int AudioInitConverter( PAUDIO_CONVERT_TYPE audioConvert, PENCODE_SETTINGS encodeSettings, int dbAdjust )
{
  int i;

  // set defaults
  audioConvert->Decode.MpegAttenuation = 0;
  audioConvert->Decode.Ac3Attenuation  = 0;

  // init a52 decoder library
//RLM - fix this someday, we need to detect proper mmx setting
  audioConvert->Decode.A52State = a52_init( 0/*MM_ACCEL_DJBFFT*/ );
  if ( !audioConvert->Decode.A52State ) {
    return AC_A52_LIB_INIT_FAILED;
  }

  // init transcoder
  TranscodeOpen( audioConvert );

  // save the encode settings
  audioConvert->Encode.EncodeSettings = *encodeSettings;


  switch ( encodeSettings->FrameSync ) {
    case MPEG_FRAME_SYNC:
      // set default for ac3->mpeg
      audioConvert->Decode.Ac3Attenuation  = 10;

      // set output frame type
      audioConvert->Encode.FrameHdr.FrameSync = MPEG_FRAME_SYNC;

      // range check and save new sample rate
      for ( i = 0; i < (int)(sizeof( m_SampleRates ) / sizeof( int32_t )); i++ ) {
        if ( m_SampleRates[i] == (int32_t)encodeSettings->u.mpeg.SampleRate ) {
          audioConvert->Encode.FrameHdr.u.mpeg.SampleRate = encodeSettings->u.mpeg.SampleRate;
          break;
        }
      }
      if ( i >= (int)(sizeof( m_SampleRates ) / sizeof( int32_t )) ) {
        return AC_INVALID_SAMPLE_RATE;
      }

      // range check and save new bitrate
      for ( i = 1; i < 15; i++ ) {
        if ( m_MpegBitRates[1][i] == encodeSettings->u.mpeg.BitRate ) {
          audioConvert->Encode.FrameHdr.u.mpeg.BitRate = encodeSettings->u.mpeg.BitRate;
          break;
        }
      }
      if ( i >= 15 ) {
        return AC_INVALID_BIT_RATE;
      }

      // range check and save mode
      if ( ((int)(encodeSettings->u.mpeg.ChannelMode) < CHANNEL_MODE_STEREO) || (encodeSettings->u.mpeg.ChannelMode > CHANNEL_MODE_SINGLE_MONO) ) {
        encodeSettings->u.mpeg.ChannelMode = CHANNEL_MODE_STEREO;
      }
      audioConvert->Encode.FrameHdr.u.mpeg.ChannelMode = encodeSettings->u.mpeg.ChannelMode;
      audioConvert->Encode.Channels = m_MpegChannels[audioConvert->Encode.FrameHdr.u.mpeg.ChannelMode];

      // range check and save psychoModel
      if ( ((int)(encodeSettings->u.mpeg.PsychoModel) < PSYCO_MODEL_NONE) || (encodeSettings->u.mpeg.PsychoModel > PSYCO_MODEL_2) ) {
        encodeSettings->u.mpeg.PsychoModel = PSYCO_MODEL_1;
      }

      // init encoder
      LayerIIEncodeInit( audioConvert );
      break;

    case AC3_FRAME_SYNC:
      // set default for mpeg->ac3
      audioConvert->Decode.MpegAttenuation = 5;

      // set output frame type
      audioConvert->Encode.FrameHdr.FrameSync = AC3_FRAME_SYNC;

      // range check and save new sample rate
      for ( i = 0; i < (int)(sizeof( m_SampleRates ) / sizeof( int32_t )); i++ ) {
        if ( m_SampleRates[i] == (int32_t)(encodeSettings->u.ac3.SampleRate) ) {
          audioConvert->Encode.FrameHdr.u.ac3.SampleRate = encodeSettings->u.ac3.SampleRate;
          break;
        }
      }
      if ( i >= (int)(sizeof( m_SampleRates ) / sizeof( int32_t )) ) {
        return AC_INVALID_SAMPLE_RATE;
      }

      // range check and save new bitrate
      for ( i = 0; i < (int)(sizeof( m_AC3BitRates ) / sizeof( int32_t )); i++ ) {
        if ( m_AC3BitRates[i] == encodeSettings->u.ac3.BitRate ) {
          audioConvert->Encode.FrameHdr.u.ac3.BitRate = encodeSettings->u.ac3.BitRate;
          break;
        }
      }
      if ( i >= (int)(sizeof( m_AC3BitRates ) / sizeof( int32_t )) ) {
        return AC_INVALID_BIT_RATE;
      }

      // this signals that we use the number of source channels
      audioConvert->Encode.Channels = 0;  // use whatever the source provides
      break;

    default:
      return AC_INVALID_OUTPUT_FORMAT;
  }

  // has user asked to override
  if ( dbAdjust != DEFAULT_DB_ADJUST ) {
    // mpge db adjust setup
    if ( dbAdjust > 18 ) {
      dbAdjust = 18;
    }
    else if ( dbAdjust < -175 ) {
      dbAdjust = -175;
    }
    audioConvert->Decode.MpegAttenuation = dbAdjust;

    // ac3 db adjust setup
    if ( dbAdjust > 96 ) {
      dbAdjust = 96;
    }
    else if ( dbAdjust < -96 ) {
      dbAdjust = -96;
    }
    audioConvert->Decode.Ac3Attenuation = dbAdjust;
  }

  return AC_OK;
}

// ---------------------------------------------------------------------
void AudioShutDownConverter( PAUDIO_CONVERT_TYPE audioConvert )
{
  a52_free( audioConvert->Decode.A52State );
  AC3_encode_close( audioConvert );
  TranscodeClose( audioConvert );
}

// ---------------------------------------------------------------------
static __inline int blah( int32_t i )
{
  if ( i > 0x43c07fff ) {
    return 32767;
  }
  else if ( i < 0x43bf8000 ) {
    return -32768;
  }
  return i - 0x43c00000;
}

// ---------------------------------------------------------------------
int AudioFindFrameHeader( PAUDIO_CONVERT_TYPE audioConvert, PFRAME_HEADER_INFO hdr )
{
  if ( !m_RawInputSampleRate ) {
    // try to find the frame header
    if ( FindFrame( audioConvert ) ) {
      memcpy( hdr, &audioConvert->Decode.FrameHdr, sizeof( FRAME_HEADER_INFO ) );
      return AC_OK;
    }
  }
  else {
    // with raw PCM everything is OK
    return AC_OK;
  }
  return AC_NO_FRAME_FOUND;
}

// ---------------------------------------------------------------------
int AudioProcessFrame( PAUDIO_CONVERT_TYPE audioConvert )
{
  int       flags;
  sample_t  level = 1;
  sample_t  gain = 1;
  int32_t   c;
  int       s;
  int32_t   len;
  int32_t   *f; // XXX assumes IEEE float format
  int       i;
  uint8_t   frameBuf[MAX_FRAME_SIZE];
  int16_t  *sptr;

  if ( !m_RawInputSampleRate ) {
    // decode frame
    switch ( audioConvert->Decode.FrameHdr.FrameSync ) {
      case MPEG_FRAME_SYNC:
        // create a buffer that has the complete frame
        if ( PipePeekBytes( &audioConvert->InputStream, frameBuf, audioConvert->Decode.FrameHdr.u.mpeg.FrameLen ) ==
                             audioConvert->Decode.FrameHdr.u.mpeg.FrameLen ) {

          // setup bit stream
          StreamInit( &audioConvert->Decode.BitStream, frameBuf, audioConvert->Decode.FrameHdr.u.mpeg.FrameLen );

          // decode frame
          if ( LayerIIDecode( audioConvert ) ) {

            // synthesis PCM data from frame
            if ( audioConvert->Encode.FrameHdr.FrameSync == AC3_FRAME_SYNC ) {
            }
            SynthFrame( audioConvert );

            // compute number of channels to use
            audioConvert->Decode.OutputChannels = m_MpegChannels[audioConvert->Decode.FrameHdr.u.mpeg.ChannelMode];

            // eat the proper number of bytes, since we just peek and are now done with them
            PipeSkipBytes( &audioConvert->InputStream, audioConvert->Decode.FrameHdr.u.mpeg.FrameLen );
          }
          else {
            return AC_DECODE_FAILED;
          }
        }
        else {
          return AC_RAN_OUT_OF_INPUT_BYTES;
        }
        break;

      case AC3_FRAME_SYNC:
        switch ( audioConvert->Encode.FrameHdr.FrameSync ) {
          default:
          case MPEG_FRAME_SYNC:
            // setup output format to what input format is
            switch ( audioConvert->Decode.OutputChannels ) {
              case 1:
                flags = A52_MONO;
                break;

              case 2:
                flags = A52_STEREO;
                break;

              case 3:
                flags = A52_3F;
                break;

              case 4:
                flags = A52_2F2R;
                break;

              case 5:
                flags = A52_3F2R;
                break;

              case 6:
                flags = A52_3F2R | A52_LFE;
                break;

              default:
                flags = audioConvert->Decode.FrameHdr.u.ac3.Flags;
                break;
            }

            // check if db level has been requested to be adjusted
            if ( audioConvert->Decode.Ac3Attenuation ) {
              gain = (sample_t)pow( 2, ((float)audioConvert->Decode.Ac3Attenuation) / 6 );
              level *= gain;
            }
            else {
              flags |= A52_ADJUST_LEVEL;
            }
            break;

          case AC3_FRAME_SYNC:
            // we are going from AC3 to AC3 we keep all the channels
            flags = audioConvert->Decode.FrameHdr.u.ac3.Flags;
            break;
        }

        // create a buffer that has the complete AC3 frame
        if ( PipePeekBytes( &audioConvert->InputStream, frameBuf, audioConvert->Decode.FrameHdr.u.ac3.FrameLen ) ==
             audioConvert->Decode.FrameHdr.u.ac3.FrameLen ) {

          if ( a52_frame( audioConvert->Decode.A52State, frameBuf, &flags, &level, 384 ) ) {
            return AC_A52_FRAME_FAILED;
          }

          // compute number of channels to use
          audioConvert->Decode.OutputChannels = m_AC3DecChannels[flags & 7];
          if ( flags & A52_LFE ) {
            audioConvert->Decode.OutputChannels++;
          }

          // put samples into the decode buffer
          for ( i = 0; i < 6; i++ ) {
            if ( a52_block( audioConvert->Decode.A52State ) ) {
              LOG_TY_USERDIAG(  "\nBad a52 frame found, creating silence frame to fill\n" );
              memset( &audioConvert->Decode.DecodedSamples, 0, sizeof( audioConvert->Decode.DecodedSamples ) );
  //            return AC_A52_BLOCK_FAILED;
            }
            else {
              f = (int32_t *) a52_samples( audioConvert->Decode.A52State ); // XXX assumes IEEE float format

              for ( c = 0; c < audioConvert->Decode.OutputChannels; c++ ) {
                for ( s = 0; s < 256; s++ ) {
                  audioConvert->Decode.DecodedSamples[c][(i * 256) + s] = blah( *f++ );
                }
              }
            }
          }
          audioConvert->Decode.Synth.Pcm.Length = 6 * 256;

          // eat the proper number of bytes, since we just peek and are now done with them
          PipeSkipBytes( &audioConvert->InputStream, audioConvert->Decode.FrameHdr.u.ac3.FrameLen );
        }
        else {
          return AC_RAN_OUT_OF_INPUT_BYTES;
        }
        break;
    }
  }
  else {
    // just make us look like we came from an AC3 input
    audioConvert->Decode.FrameHdr.FrameSync         = AC3_FRAME_SYNC;
    audioConvert->Decode.FrameHdr.u.ac3.SampleRate  = m_RawInputSampleRate;
    audioConvert->Decode.OutputChannels             = 2;

    // create a buffer that has the complete AC3 frame
    len = PipePeekBytes( &audioConvert->InputStream, frameBuf, sizeof( frameBuf ) );

    // compute number of samples per channel
    audioConvert->Decode.Synth.Pcm.Length = (int16_t)(len / 4);

    if ( audioConvert->Decode.Synth.Pcm.Length ) {
      sptr = (int16_t *)frameBuf;
      for ( s = 0; s < audioConvert->Decode.Synth.Pcm.Length; s++ ) {
        audioConvert->Decode.DecodedSamples[0][s] = *sptr++;
        audioConvert->Decode.DecodedSamples[1][s] = *sptr++;
      }

      // eat the proper number of bytes, since we just peek and are now done with them
      PipeSkipBytes( &audioConvert->InputStream, len );
    }
    else {
      return AC_OK;
    }
  }

  // transcode
  if ( !TranscodeChangeRate( audioConvert ) ) {
    // transcode error
    return AC_TRANSCODE_FAILED;
  }
  return AC_OK;
}

// ---------------------------------------------------------------------
void *AudioGetOutputFrame( PAUDIO_CONVERT_TYPE audioConvert, int *frameLength )
{
  void    *ptr;
  int     c;
  int16_t *sptr;
  int16_t  channels;
  int32_t  frameSize = (audioConvert->Encode.FrameHdr.FrameSync == MPEG_FRAME_SYNC) ? MPEG_FRAME_SIZE : AC3_FRAME_SIZE;

  if ( m_RawOutput ) {
    frameSize = (sizeof( audioConvert->Encode.Frame ) / 2) / 2; // bytes to shorts to channels
  }

  *frameLength = 0;

  // while we have a full buffer
  ptr = TranscodeGetFullFrame( audioConvert, frameSize, &channels );
  if ( ptr ) {

    if ( m_RawOutput ) {
      // fill in FrameLength
      *frameLength = frameSize * 2 * 2;

      // copy into proper buffer
      sptr = (int16_t *)audioConvert->Encode.Frame;
      for ( c = 0; c < frameSize; c++ ) {
        *sptr++ = audioConvert->Transcode.Samples.Samples[0][c];
        *sptr++ = audioConvert->Transcode.Samples.Samples[1][c];
      }
    }
    else {
      // copy into proper buffer
      memcpy( &audioConvert->Transcode.TranscodedOutput, ptr, sizeof( audioConvert->Transcode.TranscodedOutput ) );

      if ( audioConvert->Encode.EncodeSettings.FrameSync == AC3_FRAME_SYNC ) {

        // do we need to reinit
        if ( channels != audioConvert->Encode.Channels ) {

          // save new number of channels
          audioConvert->Encode.Channels = channels;

          // init encoder
          if ( AC3_encode_init( audioConvert ) == -1 ) {
            LOG_TY_FATAL( "AC3 Init Failed\n" );
          }
        }

        // encode the frame
        *frameLength = AC3_encode_frame( audioConvert );
      }
      else {
        // encode frame
        LayerIIEncodeFrame( audioConvert );

        // fill in FrameLength
        *frameLength = audioConvert->Encode.BitStream.StreamPos;
      }
    }

    // cleanup current buffer
    TranscodeDoneWithCurrentFrame( audioConvert, frameSize );

    // return pointer to frame
    return audioConvert->Encode.Frame;
  }
  return NULL;
}

// ---------------------------------------------------------------------
void *AudioGetLastFrame( PAUDIO_CONVERT_TYPE audioConvert, int *frameLength )
{
  void   *ptr;
  int     c;
  int16_t *sptr;
  int     samples;
  int16_t channels;

  *frameLength = 0;

  // decode last block if needed, the last part of the buffer has silence so we should be OK
  ptr = TranscodeGetParitalFrame( audioConvert, &samples, &channels );
  if ( ptr ) {

    if ( m_RawOutput ) {
      // fill in FrameLength
      *frameLength = samples * 2 * 2;

      // copy into proper buffer
      sptr = (int16_t *)audioConvert->Encode.Frame;
      for ( c = 0; c < samples; c++ ) {
        *sptr++ = audioConvert->Transcode.Samples.Samples[0][c];
        *sptr++ = audioConvert->Transcode.Samples.Samples[1][c];
      }
    }
    else {
      // copy into proper buffer
      memcpy( &audioConvert->Transcode.TranscodedOutput, ptr, sizeof( audioConvert->Transcode.TranscodedOutput ) );

      if ( audioConvert->Encode.EncodeSettings.FrameSync == AC3_FRAME_SYNC ) {

        // do we need to reinit
        if ( channels != audioConvert->Encode.Channels ) {

          // save new number of channels
          audioConvert->Encode.Channels = channels;

          // init encoder
          if ( AC3_encode_init( audioConvert ) == -1 ) {
            LOG_TY_FATAL( "AC3 Init Failed\n" );
          }
        }

        // encode frame
        *frameLength = AC3_encode_frame( audioConvert );
      }
      else {
        // encode frame
        LayerIIEncodeFrame( audioConvert );

        // fill in FrameLength
        *frameLength = audioConvert->Encode.BitStream.StreamPos;
      }
    }

    // cleanup current buffer
    TranscodeDoneWithCurrentFrame( audioConvert, samples );

    // return pointer to frame
    return audioConvert->Encode.Frame;
  }
  return NULL;
}

// ---------------------------------------------------------------------
void *AudioCreateSilenceFrame( PAUDIO_CONVERT_TYPE audioConvert, uint8_t *masterFrame, int *silenceFrameBytes )
{
  //RLM - we need the input size and output size
  *silenceFrameBytes = 0;
  return NULL;
}

// ---------------------------------------------------------------------
int FindFrame( PAUDIO_CONVERT_TYPE audioConvert )
{
//  int            hdrlen = 0;
  int            length;
  int            flags;
  int            sampleRate;
  int            bitRate;
  unsigned char  fptr[7];
  PAUDIO_DECODE  frame = &audioConvert->Decode;

  // clear out frame info
  memset( &frame->FrameHdr, 0, sizeof( FRAME_HEADER_INFO ) );
  memset( &frame->SbSample, 0, sizeof( frame->SbSample ) );
  memset( &frame->FrameDuration, 0, sizeof( frame->FrameDuration ) );

  // we need at least 4 bytes for mpeg and 7 bytes for AC3
  while ( PipePeekBytes( &(audioConvert->InputStream), fptr, 2 ) == 2 ) {

    // check for mpeg header sig
    if ( (fptr[0] == 0xFF) && ((fptr[1] & 0xE0) == 0xE0) ) {

      // we need upto 6 bytes for the header, get them now
      if ( PipePeekBytes( &(audioConvert->InputStream), fptr, 6 ) == 6 ) {

        // we have found a posible header, lets validate it
        // 0 layer is bad, f for bitsreate index is bad, 3 for sample rate is bad
        if ( ((fptr[1] & 0x6) != 0) && ((fptr[2] & 0xF0) != 0xF0) && ((fptr[2] & 0xC) != 0xC) ) {

          // init bit stream
          StreamInit( &frame->BitStream, fptr, 6 );

          // check for valid header, it will be filled in if valid hdr
          length = LayerIIHdrCheck( audioConvert );
          if ( length ) {

            // caculate frame duration
            timer_set( &frame->FrameDuration, 0, 32 * (frame->FrameHdr.u.mpeg.LayerDescrition == MPEG_LAYER1 ? 12 : 36), frame->FrameHdr.u.mpeg.SampleRate );

            // add frame to total time
            timer_add( &frame->Stats.TotalPlayTime, frame->FrameDuration );
            return 1;
          }
        }
      }
    }
    else {
      // check for AC3 header
      if ( (fptr[0] == 0xB) && (fptr[1] & 0x77) ) {

        // we need a min of 7 bytes for the header, get them now
        if ( PipePeekBytes( &(audioConvert->InputStream), fptr, 7 ) == 7 ) {

          // decode header
          length = a52_syncinfo( fptr, &flags, &sampleRate, &bitRate );
          if ( length ) {
            frame->FrameHdr.FrameSync        = AC3_FRAME_SYNC;
            frame->FrameHdr.u.ac3.SampleRate = sampleRate;
            frame->FrameHdr.u.ac3.BitRate    = bitRate;
            frame->FrameHdr.u.ac3.Flags      = flags;
            frame->FrameHdr.u.ac3.FrameLen   = length;
            frame->FrameHdr.u.ac3.Channels   = m_AC3DecChannels[flags & 7];
            if ( flags & A52_LFE ) {
              frame->FrameHdr.u.ac3.Channels++;
            }

            // caculate frame duration
            timer_set( &frame->FrameDuration, 0, AC3_FRAME_SIZE, frame->FrameHdr.u.ac3.SampleRate );

            // add frame to total time
            timer_add( &frame->Stats.TotalPlayTime, frame->FrameDuration );
            return 1;
          }
        }
        else {
          break;
        }
      }
    }

    // move to next byte
    if ( !PipeSkipBytes( &audioConvert->InputStream, 1 ) ) {
      break;
    }
  }
  return 0;
}

