/*
 * Copyright (C) 2002  Rowan McFarland <tytranscode@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "audio_convert_structs.h"

// global data
static const int m_BitPos[8]   = { 0x80,  0x40, 0x20, 0x10, 0x8,  0x4,  0x2,  0x1  };
static const int m_BitClr[8]   = { 0x7f,  0xbf, 0xdf, 0xef, 0xf7, 0xfb, 0xfd, 0xfe };
static const int m_BitShift[8] = { 7,     6,    5,    4,    3,    2,    1,    0 };
static const int m_BitsLeft[8] = { 8,     7,    6,    5,    4,    3,    2,    1 };
static const int m_BitMask[32] = { 0x1,       0x2,       0x4,       0x8,       0x10,       0x20,       0x40,       0x80,
                                   0x100,     0x200,     0x400,     0x800,     0x1000,     0x2000,     0x4000,     0x8000,
                                   0x10000,   0x20000,   0x40000,   0x80000,   0x100000,   0x200000,   0x400000,   0x800000,
                                   0x1000000, 0x2000000, 0x4000000, 0x8000000, 0x10000000, 0x20000000, 0x40000000, 0x80000000
                                 };


/* Calculate the number of bytes in the circular buffer */
static unsigned long bytes_in_buffer( PPIPE_STREAM_TYPE stream )
{
  unsigned long nbytes = 0;

  if( stream->StreamPos < stream->BufferHighWater ) {
    /* Stream data in the buffer is contiguous */
    nbytes = (stream->BufferHighWater - stream->StreamPos);
  }

  if( stream->StreamPos > stream->BufferHighWater ) {
    /* Stream data in the buffer has wrapped */
    nbytes = stream->BufferSize - stream->StreamPos;
    nbytes += (stream->BufferHighWater+1);
  }

  return nbytes;
}

/* Ensure at least nbytes are in the buffer. refilling if necessary */
static void refill_buffer( PPIPE_STREAM_TYPE stream, unsigned long nbytes )
{
  unsigned long bytesread;

  assert( nbytes <= stream->BufferSize );

  if( bytes_in_buffer( stream ) < nbytes ) {
    /* Need to refill the buffer */

    if( stream->BufferHighWater == stream->StreamPos ) {
      /* Buffer is empty so fill from the start */
      bytesread = thread_pipe_read( stream->PipeHandle,
                      stream->BufferSize,
                      stream->BufferPtr );
      stream->BufferHighWater = bytesread-1;
      stream->StreamPos = 0;
    } else if( stream->BufferHighWater < stream->StreamPos ) {
      /* HWM below current position so fill up to the current position */
      bytesread = thread_pipe_read( stream->PipeHandle,
                      (stream->StreamPos - stream->BufferHighWater) - 1,
                      &( stream->BufferPtr[ stream->BufferHighWater + 1]) );
      stream->BufferHighWater += bytesread;

    }else if( stream->BufferHighWater < (stream->BufferSize-1) ) {
      /* HWM above current position, fill to the end of the buffer */
      bytesread = thread_pipe_read( stream->PipeHandle,
                      stream->BufferSize - stream->BufferHighWater - 1,
                      &( stream->BufferPtr[ stream->BufferHighWater + 1]) );
      stream->BufferHighWater += bytesread;
    }

    if( (stream->BufferHighWater == (stream->BufferSize - 1) ) &&
      (stream->StreamPos > 1) ) {
      /* Buffer will wrap so fill from the start of the stream to the current buffer position */

      bytesread = thread_pipe_read( stream->PipeHandle,
                      stream->StreamPos,
                      stream->BufferPtr );
      stream->BufferHighWater = bytesread - 1;
    }
  }
}

// ---------------------------------------------------------------------
void PipeInit( PPIPE_STREAM_TYPE stream, pipe_handle_t  pipehandle, int write, unsigned long bufferlen )
{
  memset( stream, 0, sizeof( PIPE_STREAM_TYPE ) );
  stream->PipeHandle  = pipehandle;
  stream->BufferPtr = (unsigned char *) malloc( bufferlen );
  stream->BufferSize = bufferlen;
  stream->WriteBuffer = write;
}

// ---------------------------------------------------------------------
unsigned long  PipePeekBytes( PPIPE_STREAM_TYPE stream, unsigned char *buffer, unsigned long nbytes )
{
  unsigned long returned = 0;
  unsigned long index;

  assert( !stream->WriteBuffer );

  if ( PipeCheckAvailBytes( stream, nbytes ) ) {
    index = stream->StreamPos;

    while( returned < nbytes ) {
      /* Check for end of buffer */
      if( index == stream->BufferHighWater ) {
        break; /* End of buffer found, no more bytes can be returned */
      }

      buffer[ returned++ ] = stream->BufferPtr[ index++ ];

      /* Check for buffer wrap-around */
      if( index == stream->BufferSize ) {
        index = 0;
      }
    }
  }

  return returned;
}

// ---------------------------------------------------------------------
int PipeCheckAvailBytes( PPIPE_STREAM_TYPE stream, unsigned long byts )
{
  if( stream->WriteBuffer ) {
    return 0;
  }

  /* Ensure at least nbytes are in the buffer. refilling if necessary */
  refill_buffer( stream, byts );

  /* Check there is enough in the buffer */
  if( bytes_in_buffer( stream ) < byts ) {
    return 0;
  }

    return 1;
}

// ---------------------------------------------------------------------
int PipeSkipBytes( PPIPE_STREAM_TYPE stream, unsigned long byts )
{
  assert( !stream->WriteBuffer );
  assert( byts < stream->BufferSize );

  /* Ensure sufficient bytes are in the buffer */
  if( !PipeCheckAvailBytes( stream, byts ) ) {
    return 0;
  }

  if( stream->StreamPos + byts >= stream->BufferSize ) {
    /* Wrap the buffer */
    stream->StreamPos = ( stream->StreamPos + byts ) - stream->BufferSize;

  } else {
	  if( byts >= ( stream->BufferHighWater - stream->StreamPos ) ) {
		  stream->StreamPos = stream->BufferHighWater;
		  return 0; /* Don't overlap the HWM */
	  }
    stream->StreamPos += byts;
  }
  stream->BytCounter += byts;

  stream->StreamBitPos = 0;

  return 1;
}

// ---------------------------------------------------------------------
unsigned long PipeGetBits( PPIPE_STREAM_TYPE stream, int bits )
{
  unsigned long  rc = 0;
  unsigned long bytes_left;

  assert( !stream->WriteBuffer );

  // range check bits
  if ( bits > 32 ) {
    bits = 32;
  }

  /* Ensure enough bytes are in the buffer */
  refill_buffer( stream, bits/8 );
  bytes_left = bytes_in_buffer( stream );

  // make sure we have enough free bits/bytes left
  if ( bits > m_BitsLeft[stream->StreamBitPos] ) {

    // convert bits needed to bytes and make sure we have that may bytes left
    if ( (unsigned long)((bits - m_BitsLeft[stream->StreamBitPos]) / 8) > (bytes_left-1) ) {

      // convert number of bytes left to bits
      bits = m_BitsLeft[stream->StreamBitPos] + ( (bytes_left-1) * 8 );
    }
  }

  // extract the needed bits
  while ( bits ) {
    // make room for the bit
    rc <<= 1;

    // or in the bit
    rc |= ( stream->BufferPtr[stream->StreamPos] & m_BitPos[stream->StreamBitPos] ) >> m_BitShift[stream->StreamBitPos];
    stream->StreamBitPos++;
    if ( stream->StreamBitPos >= 8 ) {
      stream->StreamBitPos = 0;
      stream->StreamPos++;
      stream->BytCounter++;

      if( stream->StreamPos >= stream->BufferSize ) {
        /* Wrap the buffer */
        stream->StreamPos = 0;
      }
    }
    bits--;
  }
  return( rc );
}

// ---------------------------------------------------------------------
void PipeFlushWrite( PPIPE_STREAM_TYPE stream )
{
  assert( stream->WriteBuffer );

  /* Flush from current position to the HWM */
  if( stream->StreamPos < stream->BufferHighWater ) {
    thread_pipe_write(  stream->PipeHandle,
              stream->BufferHighWater - stream->StreamPos,
              &( stream->BufferPtr[ stream->StreamPos ] ) );
    stream->StreamPos = 0;
    stream->BufferHighWater = 0;
    stream->StreamBitPos = 0;
  }
}

// ---------------------------------------------------------------------
void PipePutBits( PPIPE_STREAM_TYPE stream, int bits, unsigned long val )
{
  int bitPos;

  assert( stream->WriteBuffer );

  // range check bits
  if ( bits > 32 ) {
    bits = 32;
  }

  bitPos  = bits - 1;

  // add the needed bits
  while ( bits ) {

    // check bit
    if ( (val & m_BitMask[bitPos--]) ) {
      // set bit
      stream->BufferPtr[stream->BufferHighWater] |= m_BitPos[stream->StreamBitPos];
    }
    else {
      // clear bit
      stream->BufferPtr[stream->BufferHighWater] &= m_BitClr[stream->StreamBitPos];
    }

    // adjust bit postion
    stream->StreamBitPos++;
    if ( stream->StreamBitPos >= 8 ) {
      stream->StreamBitPos = 0;
      stream->BufferHighWater++;

    /* Check for buffer flush required */
    if( stream->BufferHighWater == stream->BufferSize )
    PipeFlushWrite( stream );
    }
    bits--;
  }
}

// ---------------------------------------------------------------------
void StreamInit( PBIT_STREAM_TYPE stream, void *memPtr, unsigned long len )
{
  memset( stream, 0, sizeof( BIT_STREAM_TYPE ) );
  stream->StreamPtr  = (unsigned char *)memPtr;
  stream->StreamSize = len;
}

// ---------------------------------------------------------------------
unsigned char *StreamGetBytePtr( PBIT_STREAM_TYPE stream )
{
  return stream->StreamPtr;
}

// ---------------------------------------------------------------------
int StreamCheckAvailBytes( PBIT_STREAM_TYPE stream, int byts )
{
  // zero means unlimited size
  if ( !stream->StreamSize ) {
    return 1;
  }

  // check if we have byts left in stream
  if ( (stream->StreamPos + byts) < stream->StreamSize ) {
    return 1;
  }
  return 0;
}

// ---------------------------------------------------------------------
int StreamSkipBytes( PBIT_STREAM_TYPE stream, int byts )
{
  // make syre we have bytes to skip
  if ( StreamCheckAvailBytes( stream, byts ) ) {
    stream->StreamPos   += byts;
    stream->StreamBitPos = 0;
    return 1;
  }
  return 0;
}

// ---------------------------------------------------------------------
unsigned long StreamGetBits( PBIT_STREAM_TYPE stream, int bits )
{
  unsigned long  rc = 0;

  // range check bits
  if ( bits > 32 ) {
    bits = 32;
  }

  // make sure we have enough free bits/bytes left
  if ( bits > m_BitsLeft[stream->StreamBitPos] ) {

    // convert bits needed to bytes and make sure we have that may bytes left
    if ( (unsigned long)((bits - m_BitsLeft[stream->StreamBitPos]) / 8) > (stream->StreamSize - (stream->StreamPos + 1)) ) {

      // convert number of bytes left to bits
      bits = (stream->StreamSize - (stream->StreamPos + 1)) * 8;
    }
  }

  // extract the needed bits
  while ( bits ) {
    // make room for the bit
    rc <<= 1;

    // or in the bit
    rc |= (*(stream->StreamPtr + stream->StreamPos) & (m_BitPos[stream->StreamBitPos])) >> m_BitShift[stream->StreamBitPos];
    stream->StreamBitPos++;
    if ( stream->StreamBitPos >= 8 ) {
      stream->StreamBitPos = 0;
      stream->StreamPos++;
    }
    bits--;
  }
  return( rc );
}

// ---------------------------------------------------------------------
void StreamPutBits( PBIT_STREAM_TYPE stream, int bits, unsigned long val )
{
  int bitPos;

  // range check bits
  if ( bits > 32 ) {
    bits = 32;
  }

  bitPos  = bits - 1;

  // add the needed bits
  while ( bits ) {

    // check bit
    if ( (val & m_BitMask[bitPos--]) ) {
      // set bit
      *(stream->StreamPtr + stream->StreamPos) |= m_BitPos[stream->StreamBitPos];
    }
    else {
      // clear bit
      *(stream->StreamPtr + stream->StreamPos) &= m_BitClr[stream->StreamBitPos];
    }

    // adjust bit postion
    stream->StreamBitPos++;
    if ( stream->StreamBitPos >= 8 ) {
      stream->StreamBitPos = 0;
      stream->StreamPos++;
    }
    bits--;
  }
}

// ---------------------------------------------------------------------
// we just flush the rest of the bits to make a total byte
void StreamFlush( PBIT_STREAM_TYPE stream )
{
  // compute number of bits it will take to make us exactly 8
  if ( stream->StreamBitPos ) {
    StreamPutBits( stream, 8 - stream->StreamBitPos, 0 );
  }
}

// ---------------------------------------------------------------------
void StreamPadBytes( PBIT_STREAM_TYPE stream, int byts )
{
  while ( byts-- ) {
    StreamPutBits( stream, 8, 0 );
  }
}
