/*
 * mad - MPEG audio decoder
 * Copyright (C) 2000-2001 Robert Leslie
 *
 *  Modifyed by Rowan McFarland <tytranscode@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "audio_convert_structs.h"

/*
 * scalefactor table
 * used in both Layer I and Layer II decoding
 */
static mad_fixed_t const sf_table[63] = {
# include "sf_table.dat"
};

/* --- Layer II ------------------------------------------------------------ */

/* possible quantization per subband table */
static struct {
  unsigned int sblimit;
  unsigned char offsets[30];
} const sbquant_table[5] = {
  /* ISO/IEC 11172-3 Table B.2a */
  { 27, { 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 3, 3, 3, 3, 3, /* 0 */
    3, 3, 3, 3, 3, 3, 3, 0, 0, 0, 0 } },
  /* ISO/IEC 11172-3 Table B.2b */
  { 30, { 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 3, 3, 3, 3, 3, /* 1 */
    3, 3, 3, 3, 3, 3, 3, 0, 0, 0, 0, 0, 0, 0 } },
  /* ISO/IEC 11172-3 Table B.2c */
  {  8, { 5, 5, 2, 2, 2, 2, 2, 2 } },       /* 2 */
  /* ISO/IEC 11172-3 Table B.2d */
  { 12, { 5, 5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2 } },   /* 3 */
  /* ISO/IEC 13818-3 Table B.1 */
  { 30, { 4, 4, 4, 4, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, /* 4 */
    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 } }
};

/* bit allocation table */
static struct {
  unsigned short nbal;
  unsigned short offset;
} const bitalloc_table[8] = {
  { 2, 0 },  /* 0 */
  { 2, 3 },  /* 1 */
  { 3, 3 },  /* 2 */
  { 3, 1 },  /* 3 */
  { 4, 2 },  /* 4 */
  { 4, 3 },  /* 5 */
  { 4, 4 },  /* 6 */
  { 4, 5 }   /* 7 */
};

/* offsets into quantization class table */
static unsigned char const offset_table[6][15] = {
  { 0, 1, 16                                             },  /* 0 */
  { 0, 1,  2, 3, 4, 5, 16                                },  /* 1 */
  { 0, 1,  2, 3, 4, 5,  6, 7,  8,  9, 10, 11, 12, 13, 14 },  /* 2 */
  { 0, 1,  3, 4, 5, 6,  7, 8,  9, 10, 11, 12, 13, 14, 15 },  /* 3 */
  { 0, 1,  2, 3, 4, 5,  6, 7,  8,  9, 10, 11, 12, 13, 16 },  /* 4 */
  { 0, 2,  4, 5, 6, 7,  8, 9, 10, 11, 12, 13, 14, 15, 16 }   /* 5 */
};

/* quantization class table */
static struct quantclass {
  unsigned short nlevels;
  unsigned char group;
  unsigned char bits;
  mad_fixed_t C;
  mad_fixed_t D;
} const qc_table[17] = {
# include "qc_table.dat"
};

// external data
extern const int32_t m_SampleRates[3];
extern const int32_t m_MpegBitRates[5][16];


/*
 * NAME:  II_samples()
 * DESCRIPTION: decode three requantized Layer II samples from a bitstream
 */
static void II_samples( struct quantclass const *quantclass, mad_fixed_t output[3], PAUDIO_CONVERT_TYPE audioConvert )
{
  unsigned int nb, s, sample[3];

  if ((nb = quantclass->group)) {
    unsigned int c, nlevels;

    /* degrouping */
    c = StreamGetBits( &audioConvert->Decode.BitStream, quantclass->bits );
    nlevels = quantclass->nlevels;

    for (s = 0; s < 3; ++s) {
      sample[s] = c % nlevels;
      c /= nlevels;
    }
  }
  else {
    nb = quantclass->bits;

    for (s = 0; s < 3; ++s)
      sample[s] = StreamGetBits( &audioConvert->Decode.BitStream, nb );
  }

  for (s = 0; s < 3; ++s) {
    mad_fixed_t requantized;

    /* invert most significant bit, extend sign, then scale to fixed format */

    requantized  = sample[s] ^ (1 << (nb - 1));
    requantized |= -(requantized & (1 << (nb - 1)));

    requantized <<= F_FRACBITS - (nb - 1);

    /* requantize the sample */

    /* s'' = C * (s''' + D) */
    output[s] = f_mul( requantized + quantclass->D, quantclass->C );

    /* s' = factor * s'' */
    /* (to be performed by caller) */
  }
}

// ---------------------------------------------------------------------
int LayerIIDecode( PAUDIO_CONVERT_TYPE audioConvert )
{
  unsigned int index, sblimit, nbal, nch, bound, gr, ch, s, sb;
  unsigned char const *offsets;
  unsigned char allocation[2][32], scfsi[2][32], scalefactor[2][32][3];
  mad_fixed_t samples[3];
//  mad_fixed_t bandfactor;

  // re-due header
  if ( !LayerIIHdrCheck( audioConvert ) ) {
    return 0;
  }

  nch = (audioConvert->Decode.FrameHdr.u.mpeg.ChannelMode != CHANNEL_MODE_SINGLE_MONO) ? 2 : 1;

  switch (nch == 2 ? audioConvert->Decode.FrameHdr.u.mpeg.BitRate / 2 : audioConvert->Decode.FrameHdr.u.mpeg.BitRate) {
  case 32000:
  case 48000:
    index = (audioConvert->Decode.FrameHdr.u.mpeg.SampleRate == 32000) ? 3 : 2;
    break;

  case 56000:
  case 64000:
  case 80000:
    index = 0;
    break;

  default:
    index = (audioConvert->Decode.FrameHdr.u.mpeg.SampleRate == 48000) ? 0 : 1;
  }

  sblimit = sbquant_table[index].sblimit;
  offsets = sbquant_table[index].offsets;

  bound = 32;
  if ( audioConvert->Decode.FrameHdr.u.mpeg.ChannelMode == CHANNEL_MODE_JOINT_STEREO ) {
    bound = 4 + audioConvert->Decode.FrameHdr.u.mpeg.ModeExtension * 4;
  }

  if (bound > sblimit)
    bound = sblimit;

  // decode bit allocations
  for (sb = 0; sb < bound; ++sb) {
    nbal = bitalloc_table[offsets[sb]].nbal;

    for (ch = 0; ch < nch; ++ch)
      allocation[ch][sb] = (unsigned char)StreamGetBits( &audioConvert->Decode.BitStream, nbal );
  }

  for (sb = bound; sb < sblimit; ++sb) {
    nbal = bitalloc_table[offsets[sb]].nbal;

    allocation[0][sb] =
    allocation[1][sb] = (unsigned char)StreamGetBits( &audioConvert->Decode.BitStream, nbal );
  }

  // decode scalefactor selection info
  for (sb = 0; sb < sblimit; ++sb) {
    for (ch = 0; ch < nch; ++ch) {
      if (allocation[ch][sb])
        scfsi[ch][sb] = (unsigned char)StreamGetBits( &audioConvert->Decode.BitStream, 2 );
    }
  }

  /* check CRC word */
  // for now we don't validate CRC

  // decode scalefactors
  for (sb = 0; sb < sblimit; ++sb) {
    for (ch = 0; ch < nch; ++ch) {
      if (allocation[ch][sb]) {
        scalefactor[ch][sb][0] = (unsigned char)StreamGetBits( &audioConvert->Decode.BitStream, 6 );

          switch (scfsi[ch][sb]) {
          case 2:
            scalefactor[ch][sb][2] =
            scalefactor[ch][sb][1] =
            scalefactor[ch][sb][0];
            break;

          case 0:
            scalefactor[ch][sb][1] = (unsigned char)StreamGetBits( &audioConvert->Decode.BitStream, 6 );
            /* fall through */

          case 1:
          case 3:
            scalefactor[ch][sb][2] = (unsigned char)StreamGetBits( &audioConvert->Decode.BitStream, 6 );
          }

          if (scfsi[ch][sb] & 1)
            scalefactor[ch][sb][1] = scalefactor[ch][sb][scfsi[ch][sb] - 1];

          if (scalefactor[ch][sb][0] == 63 ||
              scalefactor[ch][sb][1] == 63 ||
              scalefactor[ch][sb][2] == 63) {
            // error - bad scale factor
            return 0;
          }
      }
    }
  }

  // decode samples
  for (gr = 0; gr < 12; ++gr) {
    for (sb = 0; sb < bound; ++sb) {
      for (ch = 0; ch < nch; ++ch) {
        if ((index = allocation[ch][sb])) {
          index = offset_table[bitalloc_table[offsets[sb]].offset][index - 1];

          II_samples( &qc_table[index], samples, audioConvert );

          for (s = 0; s < 3; ++s) {
            audioConvert->Decode.SbSample[ch][3 * gr + s][sb] =
              f_mul(samples[s], sf_table[scalefactor[ch][sb][gr / 4]]);
          }
        }
        else {
          for (s = 0; s < 3; ++s)
            audioConvert->Decode.SbSample[ch][3 * gr + s][sb] = 0;
        }
      }
    }

    for (sb = bound; sb < sblimit; ++sb) {
      if ((index = allocation[0][sb])) {
        index = offset_table[bitalloc_table[offsets[sb]].offset][index - 1];

        II_samples( &qc_table[index], samples, audioConvert );

        for (ch = 0; ch < nch; ++ch) {
          for (s = 0; s < 3; ++s) {
            audioConvert->Decode.SbSample[ch][3 * gr + s][sb] =
              f_mul(samples[s], sf_table[scalefactor[ch][sb][gr / 4]]);
          }
        }
      }
      else {
        for (ch = 0; ch < nch; ++ch) {
          for (s = 0; s < 3; ++s)
            audioConvert->Decode.SbSample[ch][3 * gr + s][sb] = 0;
        }
      }
    }

    for (ch = 0; ch < nch; ++ch) {
      for (s = 0; s < 3; ++s) {
        for (sb = sblimit; sb < 32; ++sb)
          audioConvert->Decode.SbSample[ch][3 * gr + s][sb] = 0;
      }
    }
  }
  return 1;
}

// ---------------------------------------------------------------------
int LayerIIHdrCheck( PAUDIO_CONVERT_TYPE audioConvert )
{
  int rc = 0;

  // OK we have a header lets extract it
  audioConvert->Decode.FrameHdr.FrameSync              = (unsigned short)StreamGetBits( &audioConvert->Decode.BitStream, 11 );
  audioConvert->Decode.FrameHdr.u.mpeg.AudioVer        = (MPEG_VERS)StreamGetBits( &audioConvert->Decode.BitStream, 2 );
  audioConvert->Decode.FrameHdr.u.mpeg.LayerDescrition = (MPEG_LAYERS)StreamGetBits( &audioConvert->Decode.BitStream, 2 );
  audioConvert->Decode.FrameHdr.u.mpeg.Protection      = StreamGetBits( &audioConvert->Decode.BitStream, 1 );
  audioConvert->Decode.FrameHdr.u.mpeg.BitrateIndex    = StreamGetBits( &audioConvert->Decode.BitStream, 4 );
  audioConvert->Decode.FrameHdr.u.mpeg.SampleRateIndex = StreamGetBits( &audioConvert->Decode.BitStream, 2 );
  audioConvert->Decode.FrameHdr.u.mpeg.Padding         = StreamGetBits( &audioConvert->Decode.BitStream, 1 );
  audioConvert->Decode.FrameHdr.u.mpeg.Private         = StreamGetBits( &audioConvert->Decode.BitStream, 1 );
  audioConvert->Decode.FrameHdr.u.mpeg.ChannelMode     = (MPEG_CHANNEL_MODE)StreamGetBits( &audioConvert->Decode.BitStream, 2 );
  audioConvert->Decode.FrameHdr.u.mpeg.ModeExtension   = StreamGetBits( &audioConvert->Decode.BitStream, 2 );
  audioConvert->Decode.FrameHdr.u.mpeg.Copyright       = StreamGetBits( &audioConvert->Decode.BitStream, 1 );
  audioConvert->Decode.FrameHdr.u.mpeg.Original        = StreamGetBits( &audioConvert->Decode.BitStream, 1 );
  audioConvert->Decode.FrameHdr.u.mpeg.Emphasis        = StreamGetBits( &audioConvert->Decode.BitStream, 2 );

  // read CRC if needed
  if ( !audioConvert->Decode.FrameHdr.u.mpeg.Protection ) {
    audioConvert->Decode.FrameHdr.u.mpeg.HdrCrc = (unsigned int)StreamGetBits( &audioConvert->Decode.BitStream, 16 );
  }

  // compute frame length
  switch ( audioConvert->Decode.FrameHdr.u.mpeg.AudioVer ) {
    case MPEG_1_VER:
      audioConvert->Decode.FrameHdr.u.mpeg.SampleRate = m_SampleRates[audioConvert->Decode.FrameHdr.u.mpeg.SampleRateIndex];
      switch ( audioConvert->Decode.FrameHdr.u.mpeg.LayerDescrition ) {
        default:
        case MPEG_LAYER1:
        case MPEG_LAYER3:
          // we don't support this
          break;

        case MPEG_LAYER2:
          audioConvert->Decode.FrameHdr.u.mpeg.BitRate  = m_MpegBitRates[1][audioConvert->Decode.FrameHdr.u.mpeg.BitrateIndex];
          audioConvert->Decode.FrameHdr.u.mpeg.FrameLen = (144 * audioConvert->Decode.FrameHdr.u.mpeg.BitRate /
                                                           audioConvert->Decode.FrameHdr.u.mpeg.SampleRate) +
                                                           audioConvert->Decode.FrameHdr.u.mpeg.Padding;
          rc = audioConvert->Decode.FrameHdr.u.mpeg.FrameLen;
          break;
      }
      break;

    default:
    case MPEG_2_VER:
      // we don't support this
      break;
  }
  return rc;
}
