/*
 * Copyright (C) 2002  Rowan McFarland <tytranscode@hotmail.com>
 *
 *  functions were taken from: sox (http://home.sprynet.com/~cbagwell/sox.html))
 * by Alexander Oelzant <aoe@mars.tuwien.ac.at>
 * Copyright 1991 Lance Norskog And Sundry Contributors
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "audio_convert_structs.h"
#include "resampl.h"

// local defines & typedefs
#define IzeroEPSILON 1E-21               // Max error acceptable in Izero
#define NQMAX        511      // largest factor for which exact-coefficients upsampling will be used
#define BUFFSIZE     8192 /*16384*/  // Total I/O buffer size
#define M_PI         3.14159265358979323846
#define ISCALE       0x10000

// local functions
int       ValidRate( int rate );
void      StereoToMono( int16_t *dstBuf, uint32_t dstLen, int16_t *channel1, int16_t *channel2, uint32_t srcShorts );

int       RateSetRateChange( prate_filter_t rate, int inRate, int outRate );
int32_t   RateChangeRate( prate_filter_t rate, int16_t *inputBuf, int32_t isamp, int16_t *outputBuf );

void      SetChannelMap( PAUDIO_CONVERT_TYPE audioConvert, int16_t channels, uint32_t nitems );
uint32_t  InQueueBlk( PQUEUE_TYPE queue, int16_t *items, uint32_t nitems );
uint32_t  OutQueueBlk( PQUEUE_TYPE queue, int16_t *items, uint32_t nitems );
uint32_t  QueuePeak( PQUEUE_TYPE queue, int16_t *items, uint32_t nitems );
void      QueueEat( PQUEUE_TYPE queue, uint32_t nitems );
void      QueueFlush( PQUEUE_TYPE queue );
uint32_t  QueueFreeItems( PQUEUE_TYPE queue );
uint32_t  QueueUsedItems( PQUEUE_TYPE queue );


// ---------------------------------------------------------------------
void TranscodeOpen( PAUDIO_CONVERT_TYPE audioConvert )
{
  audioConvert->Transcode.RateConvertIdx = 0;
  memset( &audioConvert->Transcode.RateConvert, 0, sizeof( audioConvert->Transcode.RateConvert ) );
  memset( &audioConvert->Transcode.TranscodedSamples, 0, sizeof( audioConvert->Transcode.TranscodedSamples ) );
  memset( &audioConvert->Transcode.Samples, 0, sizeof( audioConvert->Transcode.Samples ) );
  audioConvert->Transcode.SamplesChannelMap = NULL;
}

// ---------------------------------------------------------------------
void TranscodeClose( PAUDIO_CONVERT_TYPE audioConvert )
{
  unsigned long r;
  int           c;

  for ( r = 0; r < audioConvert->Transcode.RateConvertIdx; r++ ) {
    for ( c = 0; c < MAX_CHANNELS; c++ ) {
      if ( audioConvert->Transcode.RateConvert[r].Filter[c].priv.Imp ) {
        free( audioConvert->Transcode.RateConvert[r].Filter[c].priv.Imp - 1 );
        audioConvert->Transcode.RateConvert[r].Filter[c].priv.Imp = NULL;
      }
      if ( audioConvert->Transcode.RateConvert[r].Filter[c].priv.X ) {
        free( audioConvert->Transcode.RateConvert[r].Filter[c].priv.X );
        audioConvert->Transcode.RateConvert[r].Filter[c].priv.X = NULL;
      }
    }
  }
}

// ---------------------------------------------------------------------
uint32_t TranscodeChangeRate( PAUDIO_CONVERT_TYPE audioConvert )
{
  int       idx = -1;
  int32_t   rateLen = 0;
  int32_t   c;
  int       r;
  uint32_t  srcSampleRate;
  int16_t   srcChannels;
  uint32_t  dstSampleRate;
  int16_t   dstChannels;

  // compute src sample rate and channels
  switch ( audioConvert->Decode.FrameHdr.FrameSync ) {
    case MPEG_FRAME_SYNC:
      srcSampleRate = audioConvert->Decode.FrameHdr.u.mpeg.SampleRate;
      srcChannels   = audioConvert->Decode.OutputChannels;
      break;

    case AC3_FRAME_SYNC:
      srcSampleRate = audioConvert->Decode.FrameHdr.u.ac3.SampleRate;
      srcChannels   = audioConvert->Decode.OutputChannels;
      break;

    default:
      // unsupported format
      return 0;
  }

  // compute dst sample rate and channels
  switch ( audioConvert->Encode.FrameHdr.FrameSync ) {
    case MPEG_FRAME_SYNC:
      dstSampleRate = audioConvert->Encode.FrameHdr.u.mpeg.SampleRate;
      dstChannels   = audioConvert->Encode.Channels;
      break;

    case AC3_FRAME_SYNC:
      dstSampleRate = audioConvert->Encode.FrameHdr.u.ac3.SampleRate;
      dstChannels   = audioConvert->Decode.OutputChannels;
      break;

    default:
      // unsupported format
      return 0;
  }

  // check if src and dst rates are the same
  if ( srcSampleRate == dstSampleRate ) {

    // are the number of channels the same
    if ( srcChannels == dstChannels ) {

      rateLen = audioConvert->Decode.Synth.Pcm.Length;
      for ( c = 0; c < srcChannels; c++ ) {
        if ( !InQueueBlk( &audioConvert->Transcode.Samples.SampleQueue[c], audioConvert->Decode.DecodedSamples[c], rateLen ) ) {
          return 0;
        }
      }
      SetChannelMap( audioConvert, srcChannels, rateLen );
    }
    else {
      // set rate len
      rateLen = audioConvert->Decode.Synth.Pcm.Length;

      // copy src samples into proper place
      for ( c = 0; c < srcChannels; c++ ) {
        memcpy( audioConvert->Transcode.TranscodedSamples[c], audioConvert->Decode.DecodedSamples[c], rateLen * sizeof( short ) );
      }

      // if mono src and requested stereo dst convert
      if ( (srcChannels == 1) && (dstChannels >= 2) ) {
        // make stereo output buffer from mono input buffer
        memcpy( audioConvert->Transcode.TranscodedSamples[1], audioConvert->Transcode.TranscodedSamples[0], rateLen * sizeof( short ) );
        dstChannels = 2;
      }

      // if stereo src and requested mono dst convert
      if ( (srcChannels >= 2) && (dstChannels == 1) ) {
        StereoToMono( audioConvert->Transcode.TranscodedSamples[0], sizeof( audioConvert->Transcode.TranscodedSamples[0] ),
                      audioConvert->Transcode.TranscodedSamples[0], audioConvert->Transcode.TranscodedSamples[1], rateLen );
      }

      // now place samples in sample queue
      for ( c = 0; c < dstChannels; c++ ) {
        if ( !InQueueBlk( &audioConvert->Transcode.Samples.SampleQueue[c], audioConvert->Transcode.TranscodedSamples[c], rateLen ) ) {
          return 0;
        }
      }
      SetChannelMap( audioConvert, dstChannels, rateLen );
    }
  }
  else {
    // make sure we are dealing with types that we can handle
    if ( !ValidRate( dstSampleRate ) || !ValidRate( srcSampleRate ) ) {
      return 0;
    }

    // see if we already have the proper rate converter created
    for ( r = 0; r < MAX_RATE_CONVERT; r++ ) {

      // check rates
      if ( (audioConvert->Transcode.RateConvert[r].DstRate == dstSampleRate) &&
           (audioConvert->Transcode.RateConvert[r].SrcRate == srcSampleRate) ) {
        idx = r;
        break;
      }
    }

    // create a new rate converter if needed
    if ( idx == -1 ) {

      // do we have room?
      if ( audioConvert->Transcode.RateConvertIdx >= MAX_RATE_CONVERT ) {
        // no, just return
        return 0;
      }

      // set working index and adjust master counter
      idx = audioConvert->Transcode.RateConvertIdx++;

      // compute rates
      audioConvert->Transcode.RateConvert[idx].DstRate = dstSampleRate;
      audioConvert->Transcode.RateConvert[idx].SrcRate = srcSampleRate;

      // create rate filter(s)
      for ( c = 0; c < MAX_CHANNELS; c++ ) {
        if ( !RateSetRateChange( &audioConvert->Transcode.RateConvert[idx].Filter[c],
                                  audioConvert->Transcode.RateConvert[idx].SrcRate,
                                  audioConvert->Transcode.RateConvert[idx].DstRate ) ) {

          // adjust master counter & cleanup entry
          audioConvert->Transcode.RateConvertIdx--;
          memset( &audioConvert->Transcode.RateConvert[idx], 0, sizeof( rate_filter_t ) );

          // no, just return
          return 0;
        }
      }
    }

    // change the rate for each channel
    for ( c = 0; c < srcChannels; c++ ) {
      // make rate change happen
      rateLen = RateChangeRate( &audioConvert->Transcode.RateConvert[idx].Filter[c],
                                audioConvert->Decode.DecodedSamples[c],
                                audioConvert->Decode.Synth.Pcm.Length,
                                audioConvert->Transcode.TranscodedSamples[c] );
      if ( !rateLen ) {
        return 0;
      }
    }

    // if mono src and requested stereo dst convert
    if ( (srcChannels == 1) && (dstChannels >= 2) ) {
      // make stereo output buffer from mono input buffer
      memcpy( audioConvert->Transcode.TranscodedSamples[1], audioConvert->Transcode.TranscodedSamples[0], rateLen * sizeof( short ) );
      dstChannels = 2;
    }

    // if stereo src and requested mono dst convert
    if ( (srcChannels >= 2) && (dstChannels == 1) ) {
      StereoToMono( audioConvert->Transcode.TranscodedSamples[0], sizeof( audioConvert->Transcode.TranscodedSamples[0] ),
                    audioConvert->Transcode.TranscodedSamples[0], audioConvert->Transcode.TranscodedSamples[1], rateLen );
    }


    // now place samples in sample queue
    for ( c = 0; c < dstChannels; c++ ) {
      if ( !InQueueBlk( &audioConvert->Transcode.Samples.SampleQueue[c], audioConvert->Transcode.TranscodedSamples[c], rateLen ) ) {
        return 0;
      }
    }
    SetChannelMap( audioConvert, dstChannels, rateLen );
  }

  // return total samples produced, may be one or more channels
  return( rateLen );
}

// ---------------------------------------------------------------------
void *TranscodeGetFullFrame( PAUDIO_CONVERT_TYPE audioConvert, int32_t requestedSamples, int16_t *channels )
{
  int16_t   c;
  uint32_t  numSamples;

  // does channel map say we have any samples
  if ( !audioConvert->Transcode.SamplesChannelMap ) {

    // double check queue
    if ( QueueUsedItems( &audioConvert->Transcode.Samples.SampleQueue[0] ) ) {
      LOG_TY_FATAL( "Channel Map and sample queue don't match\n" );
    }
    return NULL;
  }

  // compute the number of samples we can provide
  if ( (audioConvert->Transcode.SamplesChannelMap->Samples < requestedSamples) && !audioConvert->Transcode.SamplesChannelMap->Next ) {
    // only partial frame left
    return NULL;
  }

  // we have more samples after this batch but different number of channels
  if ( (audioConvert->Transcode.SamplesChannelMap->Samples < requestedSamples) ) {
    // we will have to pad,
    numSamples = audioConvert->Transcode.SamplesChannelMap->Samples;

    // clear out samples with PCM silence
    memset( &audioConvert->Transcode.Samples.Samples, 0, sizeof( audioConvert->Transcode.Samples.Samples ) );
  }
  else {
    // we have the proper number of samples
    numSamples = requestedSamples;
  }

  // safety check
  if ( QueueUsedItems( &audioConvert->Transcode.Samples.SampleQueue[0] ) < numSamples ) {
    LOG_TY_FATAL( "Channel Map and sample queue are not in sync\n" );
  }

  // peek and fill up the Samples buffer
  for ( c = 0; c < audioConvert->Transcode.SamplesChannelMap->Channels; c++ ) {
    QueuePeak( &audioConvert->Transcode.Samples.SampleQueue[c], audioConvert->Transcode.Samples.Samples[c], numSamples );
  }

  // set channels
  *channels = audioConvert->Transcode.SamplesChannelMap->Channels;

  // return pointer to samples buffer
  return &audioConvert->Transcode.Samples.Samples;
}

// ---------------------------------------------------------------------
void *TranscodeGetParitalFrame( PAUDIO_CONVERT_TYPE audioConvert, int32_t *obtainedSamples, int16_t *channels )
{
  int32_t  c;

  // does channel map say we have any samples
  if ( !audioConvert->Transcode.SamplesChannelMap ) {

    // double check queue
    if ( QueueUsedItems( &audioConvert->Transcode.Samples.SampleQueue[0] ) ) {
      LOG_TY_FATAL( "Channel Map and sample queue don't match\n" );
    }
    return NULL;
  }

  // get number of samples in the channel map
  *obtainedSamples = audioConvert->Transcode.SamplesChannelMap->Samples;

  // make sure we have that many in the queue
  if ( *obtainedSamples > (int32_t)QueueUsedItems( &audioConvert->Transcode.Samples.SampleQueue[0] ) ) {
    LOG_TY_FATAL( "Channel Map and sample queue are not in sync\n" );
    *obtainedSamples = (int32_t)QueueUsedItems( &audioConvert->Transcode.Samples.SampleQueue[0] );
  }

  // did we find any samples
  if ( *obtainedSamples ) {

    // clear out samples with PCM silence
    memset( &audioConvert->Transcode.Samples.Samples, 0, sizeof( audioConvert->Transcode.Samples.Samples ) );

    // peek and fill up the Samples buffer
    for ( c = 0; c < audioConvert->Transcode.SamplesChannelMap->Channels; c++ ) {
      QueuePeak( &audioConvert->Transcode.Samples.SampleQueue[c], audioConvert->Transcode.Samples.Samples[c], *obtainedSamples );
    }

    // set channels
    *channels = audioConvert->Transcode.SamplesChannelMap->Channels;

    // return pointer to samples buffer
    return &audioConvert->Transcode.Samples.Samples;
  }
  return NULL;
}

// ---------------------------------------------------------------------
void TranscodeDoneWithCurrentFrame( PAUDIO_CONVERT_TYPE audioConvert, int32_t usedSamples )
{
  int32_t       c;
  int16_t       channels;
  PCHANNEL_MAP  tmap = NULL;

  // clean up channel map
  if ( audioConvert->Transcode.SamplesChannelMap ) {

    // save off number of channels
    channels = audioConvert->Transcode.SamplesChannelMap->Channels;

    // will this make the head empty
    if ( (audioConvert->Transcode.SamplesChannelMap->Samples - usedSamples) <= 0 ) {

      // save head
      tmap = audioConvert->Transcode.SamplesChannelMap;

      // remove head
      audioConvert->Transcode.SamplesChannelMap = audioConvert->Transcode.SamplesChannelMap->Next;

      // free old head
      free( tmap );
    }
    else {
      // just adjust the sample count
      audioConvert->Transcode.SamplesChannelMap->Samples -= usedSamples;
    }

    // eat up the requested Samples
    for ( c = 0; c < channels; c++ ) {
      QueueEat( &audioConvert->Transcode.Samples.SampleQueue[c], usedSamples );
    }
  }
}

// ---------------------------------------------------------------------
int ValidRate( int rate )
{
  switch ( rate ) {
    case 8000:
    case 16000:
    case 24000:
    case 32000:
    case 44100:
    case 48000:
      return( 1 );
  }
  return( 0 );
}

// ---------------------------------------------------------------------
// dstlen is in bytes
void StereoToMono( int16_t *dstBuf, uint32_t dstLen, int16_t *channel1, int16_t *channel2, uint32_t srcShorts )
{
  // make sure we have room in dst buffer
  if ( srcShorts > (dstLen / 2) ) {
    srcShorts = dstLen / 2;
  }

  // combine channel1 and channel2 to make output
  while ( srcShorts-- ) {
    // combine channels and normalize
    int32_t l = (((int32_t)*channel1++) + ((int32_t)*channel2++)) / 2;

    // check for cliping
    if ( l > 32767 ) {
      l = 32767;
    }
    if ( l < -32768 ) {
      l = -32768;
    }
    *dstBuf++ = (int16_t)l;
  }
}

// ---------------------------------------------------------------------
// Process options
static int st_resample_getopts( prate_filter_t rate, int n, char **argv )
{
  presample_t r = (presample_t) &rate->priv;

  // These defaults are conservative with respect to aliasing.
  r->rolloff = 0.80;
  r->beta    = 16; // anything <=2 means Nutall window
  r->quadr   = 0;
  r->Nmult   = 45;

  // This used to fail, but with sox-12.15 it works. AW
  if ( (n >= 1) ) {
    if ( !strcmp( argv[0], "-qs" ) ) {
      r->quadr = 1;
      n--;
      argv++;
    }
    else if ( !strcmp( argv[0], "-q" ) ) {
      r->rolloff = 0.875;
      r->quadr   = 1;
      r->Nmult   = 75;
      n--;
      argv++;
    }
    else if ( !strcmp( argv[0], "-ql" ) ) {
      r->rolloff = 0.94;
      r->quadr = 1;
      r->Nmult = 149;
      n--;
      argv++;
    }
  }

  if ( (n >= 1) && (sscanf( argv[0], "%lf", &r->rolloff ) != 1) ) {
    LOG_TY_FATAL("Usage: resample [ rolloff [ beta ] ]");
    return 1;
  }
  else if ( (r->rolloff <= 0.01) || (r->rolloff >= 1.0) ) {
    LOG_TY_FATAL1( "resample: rolloff factor (%f) no good, should be 0.01<x<1.0", r->rolloff );
    return 1;
  }

  if ( (n >= 2) && !sscanf( argv[1], "%lf", &r->beta ) ) {
    LOG_TY_FATAL( "Usage: resample [ rolloff [ beta ] ]" );
    return 1;
  }
  else if ( r->beta <= 2.0 ) {
    r->beta = 0;
    LOG_TY_USERDIAG1( "resample opts: Nuttall window, cutoff %f\n", r->rolloff );
  }
  else {
//RLM    LOG_TY_USERDIAG2( "resample opts: Kaiser window, cutoff %f, beta %f\n", r->rolloff, r->beta );
  }
  return 0;
}

// ---------------------------------------------------------------------
// here for linear interp.  might be useful for other things
static int32_t st_gcd( int32_t a, int32_t b )
{
  if ( b == 0 ) {
    return a;
  }
  else {
    return st_gcd( b, a % b );
  }
}

/* LpFilter()
 *
 * reference: "Digital Filters, 2nd edition"
 *            R.W. Hamming, pp. 178-179
 *
 * Izero() computes the 0th order modified bessel function of the first kind.
 *    (Needed to compute Kaiser window).
 *
 * LpFilter() computes the coeffs of a Kaiser-windowed low pass filter with
 *    the following characteristics:
 *
 *       c[]  = array in which to store computed coeffs
 *       frq  = roll-off frequency of filter
 *       N    = Half the window length in number of coeffs
 *       Beta = parameter of Kaiser window
 *       Num  = number of coeffs before 1/frq
 *
 * Beta trades the rejection of the lowpass filter against the transition
 *    width from passband to stopband.  Larger Beta means a slower
 *    transition and greater stopband rejection.  See Rabiner and Gold
 *    (Theory and Application of DSP) under Kaiser windows for more about
 *    Beta.  The following table from Rabiner and Gold gives some feel
 *    for the effect of Beta:
 *
 * All ripples in dB, width of transition band = D*N where N = window length
 *
 *               BETA    D       PB RIP   SB RIP
 *               2.120   1.50  +-0.27      -30
 *               3.384   2.23    0.0864    -40
 *               4.538   2.93    0.0274    -50
 *               5.658   3.62    0.00868   -60
 *               6.764   4.32    0.00275   -70
 *               7.865   5.0     0.000868  -80
 *               8.960   5.7     0.000275  -90
 *               10.056  6.4     0.000087  -100
 */



// ---------------------------------------------------------------------
static double Izero( double x )
{
  double sum, u, halfx, temp;
  int32_t n;

  sum = u = n = 1;
  halfx = x / 2.0;
  do {
    temp = halfx / (double)n;
    n += 1;
    temp *= temp;
    u *= temp;
    sum += u;
  } while ( u >= IzeroEPSILON * sum );
  return sum;
}

// ---------------------------------------------------------------------
static void LpFilter( double *c, int32_t N, double frq, double Beta, int32_t Num )
{
  int32_t i;

  // Calculate filter coeffs:
  c[0] = frq;
  for ( i = 1; i < N; i++ ) {
    double x = M_PI * (double)i / (double)(Num);
    c[i] = sin( x * frq ) / x;
  }

  if ( Beta > 2 ) {
    // Apply Kaiser window to filter coeffs:
    double IBeta = 1.0 / Izero( Beta );
    for ( i = 1; i < N; i++ ) {
      double x = (double)i / (double)(N);
      c[i] *= Izero( Beta * sqrt( 1.0 - x * x ) ) * IBeta;
    }
  }
  else {
    // Apply Nuttall window:
    for ( i = 0; i < N; i++ ) {
      double x = M_PI * i / N;
      c[i] *= 0.36335819 + 0.4891775 * cos( x ) + 0.1365995 * cos( 2 * x ) + 0.0106411 * cos( 3 * x );
    }
  }
}

// ---------------------------------------------------------------------
// over 90% of CPU time spent in this iprodUD() function
// quadratic interpolation
static double qprodUD( double Imp[], double *Xp, int32_t Inc, double T0, int32_t dhb, int32_t ct )
{
  const double f = 1.0 / (1 << La);
  double v;
  int32_t Ho;

  Ho = (int32_t)(T0 * dhb);
  Ho += (ct - 1) * dhb; // so Float sum starts with smallest coef's
  Xp += (ct - 1) * Inc;
  v = 0;
  do {
    double coef;
    int32_t Hoh;
    Hoh = Ho >> La;
    coef = Imp[Hoh];
    {
      double dm,dp,t;
      dm = coef - Imp[Hoh-1];
      dp = Imp[Hoh+1] - coef;
      t = (Ho & Amask) * f;
      coef += ((dp - dm) * t + (dp + dm)) * t * 0.5;
    }
    // filter coef, lower La bits by quadratic interpolation
    v += coef * *Xp;   // sum coeff * input sample
    Xp -= Inc;     // Input signal step. NO CHECK ON ARRAY BOUNDS
    Ho -= dhb;     // IR step
  } while( --ct );

  return v;
}

// ---------------------------------------------------------------------
// linear interpolation
static double iprodUD( double Imp[], double *Xp, int32_t Inc, double T0, int32_t dhb, int32_t ct )
{
  const double f = 1.0 / (1 << La);
  double v;
  int32_t Ho;

  Ho = (int32_t)(T0 * dhb);
  Ho += (ct - 1) * dhb; // so Float sum starts with smallest coef's
  Xp += (ct - 1) * Inc;
  v = 0;
  do {
    double coef;
    int32_t Hoh;
    Hoh = Ho >> La;
    // if (Hoh >= End) break;
    coef = Imp[Hoh] + (Imp[Hoh+1] - Imp[Hoh]) * (Ho & Amask) * f;
    // filter coef, lower La bits by linear interpolation
    v += coef * *Xp;    // sum coeff * input sample
    Xp -= Inc;          // Input signal step. NO CHECK ON ARRAY BOUNDS
    Ho -= dhb;          // IR step
  } while ( --ct );

  return v;
}

// ---------------------------------------------------------------------
// From resample:filters.c
// Sampling rate conversion subroutine
static int32_t SrcUD( presample_t r, int32_t Nx )
{
   double *Ystart, *Y;
   double Factor;
   double dt;                  // Step through input signal
   double time;
   double (*prodUD)();
   int n;

   prodUD = (r->quadr) ? qprodUD : iprodUD; // quadratic or linear interp
   Factor = r->Factor;
   time = r->Time;
   dt = 1.0/Factor;        // Output sampling period
   //fprintf(stderr,"Factor %f, dt %f, ",Factor,dt);
   //fprintf(stderr,"Time %f, ",r->Time);
   // (Xh * dhb)>>La is max index into Imp[]
   //fprintf(stderr,"ct=%d\n",ct);
   //fprintf(stderr,"ct=%.2f %d\n",(double)r->Nwing*Na/r->dhb, r->Xh);
   //fprintf(stderr,"ct=%ld, T=%.6f, dhb=%6f, dt=%.6f\n"
	 //	         r->Xh, time-floor(time),(double)r->dhb/Na,dt);
   Ystart = Y = r->Y;
   n = (int)ceil( (double)Nx / dt );
   while ( n-- ) {
     double *Xp;
     double v;
     double T;
     T = time - floor( time );        /* fractional part of Time */
     Xp = r->X + (int32_t)time;      /* Ptr to current input sample */

     // Past  inner product:
     v = (*prodUD)( r->Imp, Xp, -1, T, r->dhb, r->Xh ); // needs Np*Nmult in 31 bits
     // Future inner product:
     v += (*prodUD)( r->Imp, Xp + 1, 1, (1.0 - T), r->dhb, r->Xh ); // prefer even total

     if ( Factor < 1 ) {
       v *= Factor;
     }
     *Y++ = v;              // Deposit output
     time += dt;            // Move to next sample by time increment
   }
   r->Time = time;
   //fprintf(stderr,"Time %f\n",r->Time);
   return Y - Ystart;        // Return the number of output samples
}

// ---------------------------------------------------------------------
// exact coeff's
static double prodEX( const double Imp[], double *Xp, int32_t Inc, int32_t T0, int32_t dhb, int32_t ct )
{
  double v;
  const double *Cp;

  Cp  = Imp + (ct - 1) * dhb + T0; // so Float sum starts with smallest coef's
  Xp += (ct - 1) * Inc;
  v = 0;
  do {
    v += *Cp * *Xp;   // sum coeff * input sample
    Cp -= dhb;        // IR step
    Xp -= Inc;        // Input signal step.
  } while ( --ct );

  return v;
}

// ---------------------------------------------------------------------
static int32_t SrcEX( presample_t r, int32_t Nx )
{
  double *Ystart, *Y;
  double Factor;
  int32_t a,b;
  int32_t time;
  int n;

  Factor = r->Factor;
  time = r->t;
  a = r->a;
  b = r->b;
  Ystart = Y = r->Y;
  n = (Nx * b + (a - 1)) / a;
  while ( n-- ) {
	  double *Xp;
	  double v;
	  int32_t T;
	  T = time % b;              // fractional part of Time
	  Xp = r->X + (time/b);      // Ptr to current input sample

	  // Past  inner product:
	  v = prodEX( r->Imp, Xp, -1, T, b, r->Xh );
	  // Future inner product:
	  v += prodEX( r->Imp, Xp+1, 1, b-T, b, r->Xh );

	  if (Factor < 1) v *= Factor;
	  *Y++ = v;             // Deposit output
	  time += a;            // Move to next sample by time increment
  }
  r->t = time;
  return Y - Ystart;        // Return the number of output samples
}

// ---------------------------------------------------------------------
static int makeFilter( double *Imp, int32_t Nwing, double Froll, double Beta, int32_t Num, int Normalize )
{
  double *ImpR;
  int32_t Mwing, i;

  // Check for valid parameters
  if ( Nwing > MAXNWING ) {
    return -1;
  }
  if ( (Froll <= 0) || (Froll > 1) ) {
    return -2;
  }

  // it does help accuracy a bit to have the window stop at
  // a zero-crossing of the sinc function
  Mwing = (int32_t)(floor( (double)Nwing / (Num / Froll) ) * (Num / Froll) + 0.5);
  if ( Mwing == 0 ) {
    return -4;
  }

  ImpR = (double *) malloc( sizeof( double ) * Mwing );
  if ( !ImpR ) {
    LOG_TY_FATAL( "Unable to allocate memory in makeFilter\n" );
  }

  // Design a Nuttall or Kaiser windowed Sinc low-pass filter
  LpFilter( ImpR, Mwing, Froll, Beta, Num );

  // 'correct' the DC gain of the lowpass filter
  if ( Normalize ) {
    int32_t Dh;
    double DCgain;
    DCgain = 0;
    Dh = Num;                  // Filter sampling period for factors>=1
    for ( i = Dh; i < Mwing; i += Dh ) {
      DCgain += ImpR[i];
    }
    DCgain = 2 * DCgain + ImpR[0];    // DC gain of real coefficients
    //st_report("DCgain err=%.12f",DCgain-1.0);

    DCgain = 1.0 / DCgain;
    for ( i = 0; i < Mwing; i++ ) {
      Imp[i] = ImpR[i] * DCgain;
    }
  }
  else {
    for ( i = 0; i < Mwing; i++ ) {
      Imp[i] = ImpR[i];
    }
  }
  free( ImpR );
  for ( i = Mwing; i <= Nwing; i++ ) {
    Imp[i] = 0;
  }
  // Imp[Mwing] and Imp[-1] needed for quadratic interpolation
  Imp[-1] = Imp[1];

  return Mwing;
}

// ---------------------------------------------------------------------
static int st_resample_start( prate_filter_t rate )
{
  presample_t r = (presample_t) &rate->priv;
  int32_t    Xoff, gcdrate;
  int        i;

  if ( rate->inputrate == rate->outputrate ) {
    LOG_TY_FATAL( "Input and Output rates must be different to use resample effect" );
  }

  r->Factor = (double)rate->outputrate / (double)rate->inputrate;
  gcdrate   = st_gcd( (long)rate->inputrate, (long)rate->outputrate );
  r->a      = rate->inputrate / gcdrate;
  r->b      = rate->outputrate / gcdrate;

  if ( r->a <= r->b && r->b <= NQMAX ) {
    r->quadr = -1; // exact coeff's
    r->Nq = r->b;  // MAX(r->a,r->b);
  }
  else {
    r->Nq = Nc; // for now
  }

  // Check for illegal constants
#if 0
  if (Lp >= 16) st_fail("Error: Lp>=16");
  if (Nb+Nhg+NLpScl >= 32) st_fail("Error: Nb+Nhg+NLpScl>=32");
  if (Nh+Nb > 32) st_fail("Error: Nh+Nb>32");
#endif

  // Nwing: # of filter coeffs in right wing
  r->Nwing = r->Nq * (r->Nmult/2+1) + 1;

  r->Imp = (double *)malloc( sizeof( double ) * (r->Nwing + 2) ) + 1;
  // need Imp[-1] and Imp[Nwing] for quadratic interpolation
  // returns error # <=0, or adjusted wing-len > 0
  i = makeFilter( r->Imp, r->Nwing, r->rolloff, r->beta, r->Nq, 1 );
  if ( i <= 0 ) {
//RLM
    i = makeFilter( r->Imp, MAXNWING, r->rolloff, r->beta, r->Nq, 1 );
  }
  if ( i <= 0 ) {
    LOG_TY_FATAL( "resample: Unable to make filter\n" );
    return 1;
  }

  //st_report("Nmult: %ld, Nwing: %ld, Nq: %ld\n",r->Nmult,r->Nwing,r->Nq);

  // exact coeff's method
  if (r->quadr < 0) {
    r->Xh = r->Nwing/r->b;
//RLM    LOG_TY_USERDIAG2( "resample: rate ratio %ld:%ld, coeff interpolation not needed\n", r->a, r->b );
  }
  else {
    r->dhb = Np;  // Fixed-point Filter sampling-time-increment
    if ( r->Factor<1.0 ) {
      r->dhb = (int32_t)(r->Factor*Np + 0.5);
    }
    r->Xh = (r->Nwing<<La)/r->dhb;
    // (Xh * dhb)>>La is max index into Imp[]
  }

  // reach of LP filter wings + some creeping room
  Xoff = r->Xh + 10;
  r->Xoff = Xoff;

  r->Xp = Xoff;         // Current "now"-sample pointer for input to filter
  r->Xread = Xoff;      // Position in input array to read into
  r->Time = Xoff;       // Current-time pointer for converter
  if ( r->quadr < 0 ) { // exact coeff's method
    r->t = Xoff*r->Nq;
  }

  // Check input buffer size
  i = BUFFSIZE - 2 * Xoff;
  if ( i < r->Factor + 1.0 / r->Factor ) {
    LOG_TY_FATAL( "Factor is too small or large for BUFFSIZE" );
    return 1;
  }

  r->Xsize = (int32_t)(2 * Xoff + i / (1.0 + r->Factor));
  r->Ysize = BUFFSIZE - r->Xsize;
  // st_report("Xsize %d, Ysize %d, Xoff %d",r->Xsize,r->Ysize,r->Xoff);

  r->X = (double *) malloc( sizeof( double ) * ( BUFFSIZE ) );
  r->Y = r->X + r->Xsize;

  // Need Xoff zeros at beginning of sample
  for ( i = 0; i < Xoff; i++ ) {
    r->X[i] = 0;
  }
  return 0;
}

// ---------------------------------------------------------------------
// Processed signed long samples from ibuf to obuf.
// Return number of samples processed.
static int st_resample_flow( prate_filter_t rate, int32_t *ibuf, int32_t *obuf, int32_t *isamp, int32_t *osamp )
{
	presample_t r = (presample_t)&rate->priv;
	int32_t     i, last, Nout, Nx, Nproc;

	// constrain amount we actually process
	//fprintf(stderr,"Xp %d, Xread %d, isamp %d, ",r->Xp, r->Xread,*isamp);

	Nproc = r->Xsize - r->Xp;

	i = (r->Ysize < *osamp) ? r->Ysize : *osamp;
	if ( Nproc * r->Factor >= i ) {
	  Nproc = (int32_t)(i / r->Factor);
	}

	Nx = Nproc - r->Xread; // space for right-wing future-data
	if ( Nx <= 0 ) {
    //st_fail("resample: Can not handle this sample rate change. Nx not positive: %d", Nx);
	  return 1;
	}
	if ( Nx > *isamp ) {
		Nx = *isamp;
	}
	//fprintf(stderr,"Nx %d\n",Nx);

	if ( ibuf == NULL ) {
		for( i = r->Xread; i < Nx + r->Xread; i++ ) {
			r->X[i] = 0;
		}
	}
	else {
		for( i = r->Xread; i < Nx + r->Xread; i++ ) {
			r->X[i] = (double)(*ibuf++) / ISCALE;
		}
	}
	last = i;
	Nproc = last - r->Xoff - r->Xp;

	if ( Nproc <= 0 ) {
		// fill in starting here next time
		r->Xread = last;
		// leave *isamp alone, we consumed it
		*osamp = 0;
		return 0;
	}

  // exact coeff's method
	if ( r->quadr < 0 ) {
		int32_t creep;
		Nout = SrcEX( r, Nproc );
		//fprintf(stderr,"Nproc %d --> %d\n",Nproc,Nout);
		// Move converter Nproc samples back in time
		r->t -= Nproc * r->b;
		// Advance by number of samples processed
		r->Xp += Nproc;
		// Calc time accumulation in Time
		creep = r->t / r->b - r->Xoff;
		if ( creep ) {
		  r->t -= creep * r->b;  // Remove time accumulation
		  r->Xp += creep;        // and add it to read pointer
		  //fprintf(stderr,"Nproc %ld, creep %ld\n",Nproc,creep);
		}
	}
	else { // approx coeff's method
		int32_t creep;
		Nout = SrcUD( r, Nproc );
		//fprintf(stderr,"Nproc %d --> %d\n",Nproc,Nout);
		// Move converter Nproc samples back in time
		r->Time -= Nproc;
		// Advance by number of samples processed
		r->Xp += Nproc;
		// Calc time accumulation in Time
		creep = (int32_t)(r->Time - r->Xoff);
		if ( creep ) {
		  r->Time -= creep;   // Remove time accumulation
		  r->Xp += creep;     // and add it to read pointer
		  // fprintf(stderr,"Nproc %ld, creep %ld\n",Nproc,creep);
		}
	}

	{
	int32_t i,k;
	// Copy back portion of input signal that must be re-used
	k = r->Xp - r->Xoff;
	//fprintf(stderr,"k %d, last %d\n",k,last);
	for ( i = 0; i < last - k; i++ ) {
	  r->X[i] = r->X[i+k];
	}

	// Pos in input buff to read new data into
	r->Xread = i;
	r->Xp = r->Xoff;

	for( i = 0; i < Nout; i++ ) {
		*obuf++ = (int32_t)(r->Y[i] * ISCALE);
	}

	*isamp = Nx;
	*osamp = Nout;
	}
	return 0;
}

// ---------------------------------------------------------------------
// Process tail of input samples.
static int st_resample_drain( prate_filter_t rate , int32_t *obuf, int32_t *osamp )
{
	presample_t r = (presample_t)&rate->priv;
	int32_t     isamp_res, *Obuf, osamp_res;
	int         rc;

	// fprintf(stderr,"Xoff %d, Xt %d  <--- DRAIN\n",r->Xoff, r->Xt);

	// stuff end with Xoff zeros
	isamp_res = r->Xoff;
	osamp_res = *osamp;
	Obuf = obuf;
	while ( isamp_res > 0 && osamp_res > 0 ) {
		int32_t Isamp, Osamp;
		Isamp = isamp_res;
		Osamp = osamp_res;
		rc = st_resample_flow( rate, NULL, Obuf, &Isamp, &Osamp );
		if ( rc ){
		  return rc;
		}
	  // fprintf(stderr,"DRAIN isamp,osamp  (%d,%d) -> (%d,%d)\n",
		//     isamp_res,osamp_res,Isamp,Osamp);
		Obuf += Osamp;
		osamp_res -= Osamp;
		isamp_res -= Isamp;
	}
	*osamp -= osamp_res;
	// fprintf(stderr,"DRAIN osamp %d\n", *osamp);
	if ( isamp_res ) {
		LOG_TY_ERROR1( "drain overran obuf by %d\n", isamp_res );
	}
	return 0;
}

// ---------------------------------------------------------------------
int RateSetRateChange( prate_filter_t rate, int inRate, int outRate )
{
  rate->inputrate  = inRate;  // sampling rate
  rate->outputrate = outRate; // sampling rate

  st_resample_getopts( rate, 0, NULL );
  st_resample_start( rate );
  return( 1 );
}

// ---------------------------------------------------------------------
int32_t RateChangeRate( prate_filter_t rate, int16_t *inputBuf, int32_t isamp, int16_t *outputBuf )
{
  int32_t   ibuf[MY_BUFSIZE/2];
	int32_t   osamp;
	int32_t   i;

	for ( i = 0; i < isamp; i++ ) {
		ibuf[i] = (((int32_t)inputBuf[i]) << 16);
	}
  osamp = rate->outputrate >> 2;
	st_resample_flow( rate, ibuf, rate->obuf, &isamp, &osamp );
	for ( i = 0; i < osamp; i++ ) {
		outputBuf[i] = (int16_t)(rate->obuf[i] >> 16);
	}

//RLM - Drain
	return osamp;
}

// ---------------------------------------------------------------------
void SetChannelMap( PAUDIO_CONVERT_TYPE audioConvert, int16_t channels, uint32_t nitems )
{
  PCHANNEL_MAP  cmap = NULL;
  PCHANNEL_MAP  lmap = NULL;

  // check for first one
  if ( !audioConvert->Transcode.SamplesChannelMap ) {
    // allocate struct
    audioConvert->Transcode.SamplesChannelMap = calloc( 1, sizeof( CHANNEL_MAP ) );
    audioConvert->Transcode.SamplesChannelMap->Samples  = nitems;
    audioConvert->Transcode.SamplesChannelMap->Channels = channels;
    return;
  }

  // can we just add to the head
  if ( !audioConvert->Transcode.SamplesChannelMap->Next &&
       (audioConvert->Transcode.SamplesChannelMap->Channels = channels) ) {
    audioConvert->Transcode.SamplesChannelMap->Samples += nitems;
    return;
  }

  // walk to the end of list and check if we can add to it
  cmap = audioConvert->Transcode.SamplesChannelMap->Next;
  lmap = audioConvert->Transcode.SamplesChannelMap;
  while ( cmap ) {
    if ( !cmap->Next && (cmap->Channels = channels) ) {
      cmap->Samples += nitems;
      return;
    }
    lmap = cmap;
    cmap = cmap->Next;
  }

  // if we got here then we need to create a new channel map and add to tail
  if ( lmap ) {
    cmap = calloc( 1, sizeof( CHANNEL_MAP ) );
    cmap->Samples  = nitems;
    cmap->Channels = channels;
    lmap->Next = cmap;
  }
  else {
    LOG_TY_FATAL( "Channel Map is corrupt\n" );
  }
}

// ---------------------------------------------------------------------
uint32_t InQueueBlk( PQUEUE_TYPE queue, int16_t *items, uint32_t nitems )
{
  uint32_t  tlen = nitems;
  int16_t  *dptr;

  // make sure the queue has room
  if ( nitems > (MAX_QUEUE_SIZE - queue->Size) ) {
    return( 0 );
  }

  // compute offset of new item
  dptr = &queue->Item[queue->Tail];

  // insert data, check for wrap around
  if ( (queue->Tail + nitems) >= MAX_QUEUE_SIZE ) {

    tlen = (MAX_QUEUE_SIZE - queue->Tail);
    memcpy( dptr, items, tlen * sizeof( int16_t ) );
    dptr   = queue->Item;
    items += tlen;
    tlen   = (nitems - tlen);
  }
  // copy items into queue
  if ( nitems < MAX_QUEUE_SIZE ) {
    memcpy( dptr, items, tlen * sizeof( int16_t ) );
  }

  // adjust tail and size
  queue->Tail += nitems;
  if ( queue->Tail >= MAX_QUEUE_SIZE ) {
    queue->Tail -= MAX_QUEUE_SIZE;
  }
  queue->Size += nitems;
  return( 1 );
}

// ---------------------------------------------------------------------
uint32_t OutQueueBlk( PQUEUE_TYPE queue, int16_t *items, uint32_t nitems )
{
  uint32_t tlen = nitems;
  int16_t *sptr;

  // make sure queue has at least nitems in it
  if ( queue->Size < nitems ) {
    return( 0 );
  }

  // copy data out of queue
  sptr = &queue->Item[queue->Head];

  // insert data, check for wrap around
  if ( (queue->Head + nitems) >= MAX_QUEUE_SIZE ) {
      tlen = (MAX_QUEUE_SIZE - queue->Head);
      memcpy( items, sptr, tlen * sizeof( int16_t ) );
      sptr   = queue->Item;
      items += tlen;
      tlen   = (nitems - tlen);
  }
  // copy items into queue
  if ( nitems < MAX_QUEUE_SIZE ) {
      memcpy( items, sptr, tlen * sizeof( int16_t ) );
  }

  // adjust tail and size
  queue->Head += nitems;
  if ( queue->Head >= MAX_QUEUE_SIZE ) {
      queue->Head -= MAX_QUEUE_SIZE;
  }
  queue->Size -= nitems;

  return( 1 );
}

// ---------------------------------------------------------------------
uint32_t QueuePeak( PQUEUE_TYPE queue, int16_t *items, uint32_t nitems )
{
  uint32_t tlen = nitems;
  int16_t *sptr;

  // make sure queue has at least nitems in it
  if ( queue->Size < nitems ) {
    return( 0 );
  }

  // copy data out of queue
  sptr = &queue->Item[queue->Head];

  // insert data, check for wrap around
  if ( (queue->Head + nitems) >= MAX_QUEUE_SIZE ) {
    tlen = (MAX_QUEUE_SIZE - queue->Head);
    memcpy( items, sptr, tlen * sizeof( int16_t ) );
    sptr   = queue->Item;
    items += tlen;
    tlen   = (nitems - tlen);
  }
  // copy items into queue
  if ( nitems < MAX_QUEUE_SIZE ) {
    memcpy( items, sptr, tlen * sizeof( int16_t ) );
  }
  return 1;
}

// ---------------------------------------------------------------------
void QueueEat( PQUEUE_TYPE queue, uint32_t nitems )
{
  // adjust tail and size
  queue->Head += nitems;
  if ( queue->Head >= MAX_QUEUE_SIZE ) {
      queue->Head -= MAX_QUEUE_SIZE;
  }
  queue->Size -= nitems;
}

// ---------------------------------------------------------------------
void QueueFlush( PQUEUE_TYPE queue )
{
  // flush queue
  queue->Size = 0;
  queue->Head = 0;
  queue->Tail = 0;
}

// ---------------------------------------------------------------------
uint32_t QueueFreeItems( PQUEUE_TYPE queue )
{
  return (MAX_QUEUE_SIZE - queue->Size);
}

// ---------------------------------------------------------------------
uint32_t QueueUsedItems( PQUEUE_TYPE queue )
{
  return queue->Size;
}

