/*
 * Copyright (C) 2002  Rowan McFarland <tytranscode@hotmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "tycommon.h"
#include "tytranscode.h"
#include "audio_convert_structs.h"

#include <tygetopt.h>

#include <string.h>
#include <stdarg.h>

// defines & typedefs
typedef struct {
  char *Name;
  long  Val;
} TEXT_VALUE;

// globals
pipe_handle_t           m_hPipe_In = 0;
pipe_handle_t           m_hPipe_Out = 0;
char                   *m_InputFn = NULL;
unsigned long           m_FileSize = 0;
int                     m_OutFh = -1;
char                   *m_OutputFn = NULL;
int                     m_DumpAllHdrs = 0;
int                     m_OutputRawPcm = 0;
ENCODE_SETTINGS         m_EncodeSettings;
int                     m_DbAdjust = DEFAULT_DB_ADJUST;
int                     m_argc = 0;
char                  **m_argv = NULL;
static mutex_handle_t   m_process_lock = 0;
static pipe_handle_t    m_write_pipe = 0;
int                     m_RunAsLib;
AUDIO_CONVERT_TYPE      m_Convert;
double                  m_DbAdjustCoefTbl[(MAX_DB_RANGE * 2) + 1];
int8_t                  m_RawOutput = 0;
int32_t                 m_RawInputSampleRate = 0;

#ifdef WIN32
#define vsnprintf _vsnprintf
#endif

static const char *m_AudioVerText[] = {
  "MPEG version 2.5",
  "Reserved",
  "MPEG version 2",
  "MPEG version 1"
};
static const char *m_LayerText[] = {
  "Reserved",
  "Layer III",
  "Layer II",
  "Layer I"
};
static const char *m_ProtectionText[] = {
  "Protected by CRC",
  "Not Protected by CRC"
};
static const char *m_BitrateText[5][16] = {
  {"free", "32", "64", "96", "128", "160", "192", "224", "256", "288", "320", "352", "384", "416", "448", "bad"},
  {"free", "32", "48", "56", "64",  "80",  "96",  "112", "128", "160", "192", "224", "256", "320", "384", "bad"},
  {"free", "32", "40", "48", "56",  "64",  "80",  "96",  "112", "128", "160", "192", "224", "256", "320", "bad"},
  {"free", "32", "48", "56", "64",  "80",  "96",  "112", "128", "144", "160", "176", "192", "224", "256", "bad"},
  {"free", "8",  "16", "24", "32",  "40",  "48",  "56",  "64",  "80",  "96",  "112", "128", "144", "160", "bad"}
};
static const char *m_SampleRateText[3][3] = {
  {"44100", "48000", "32000"},
  {"22050", "24000", "16000"},
  {"11025", "12000", "8000"}
};
static const char *m_ChannelModeText[] = {
  "Stereo",
  "Joint stereo",
  "Dual channel (Stereo)",
  "Single channel (Mono)"
};
static const char *m_CopyrightText[] = {
  "Not copyrighted",
  "Copyrighted"
};
static const char *m_OriginalText[] = {
  "Copy of media",
  "Original media"
};
static const char *m_EmphasisText[] = {
  "None",
  "50/15 ms",
  "reserved",
  "CCIT J.17"
};
static const char *m_ErrorTxt[] = {
  "OK",
  "Error",
  "No Frame Found",
  "No Frame Info Available",
  "Not MPEG1",
  "Not LAYER II",
  "Transcode Failed",
  "Decode failed",
  "Invalid sample rate",
  "Invalid Bit Rate",
  "AC3 Init Failed",
  "AC3 Frame Failed",
  "AC3 Block Failed",
  "Decode Ran out of input bytes",
  "Invalid Output Format",
  "MPEG encoder initilization failed",
  "AC3 encoder initilization failed"
};
static const char *m_Ac3ChannelTxt[] = {
  "Channel",
  "Mono",
  "Stereo",
  "3F",
  "2F1R",
  "3F1R",
  "2F2R",
  "3F2R",
  "CHANNEL1",
  "CHANNEL2",
  "DOLBY"
};

// ----------------------------------------------------------------------------
void DebugTrace( const char *str, ... )
{
  char    buf[2048];
  va_list arglist;

  va_start( arglist, str );
  vsnprintf( buf, sizeof( buf ), str, arglist);
  va_end(arglist);
  buf[sizeof( buf ) - 1] = 0;
  printf( "%s", buf );
}

// ---------------------------------------------------------------------
void DumpHdr( FRAME_HEADER_INFO hdr, long framecnt )
{
  int bitidx = 0;
  int sampleidx = 0;

  switch ( hdr.FrameSync ) {
    case MPEG_FRAME_SYNC:
      switch ( hdr.u.mpeg.AudioVer ) {
        default:

       case MPEG_1_VER:
          sampleidx = 0;
          switch ( hdr.u.mpeg.LayerDescrition ) {
            default:
            case MPEG_LAYER1:
              bitidx = 0;
              break;

            case MPEG_LAYER2:
              bitidx = 1;
              break;

            case MPEG_LAYER3:
              bitidx = 2;
              break;
          }
          break;

        case MPEG_2_VER:
          sampleidx = 1;
          switch ( hdr.u.mpeg.LayerDescrition ) {
            default:
            case MPEG_LAYER1:
              bitidx = 3;
              break;

            case MPEG_LAYER2:
            case MPEG_LAYER3:
              bitidx = 4;
              break;
          }
          break;

        case MPEG_2_5_VER:
          sampleidx = 2;
          break;
      }

      LOG_TY_USERDIAG2( "MPEG Hdr %d, len = %d\n", framecnt, hdr.u.mpeg.FrameLen );

      LOG_TY_USERDIAG1( "  FrameSync    = 0x%x\n", hdr.FrameSync );
      LOG_TY_USERDIAG1( "  Ver          = %s\n", m_AudioVerText[hdr.u.mpeg.AudioVer] );
      LOG_TY_USERDIAG1( "  Layer        = %s\n", m_LayerText[hdr.u.mpeg.LayerDescrition] );
      LOG_TY_USERDIAG1( "  Protection   = %s\n", m_ProtectionText[hdr.u.mpeg.Protection] );
      LOG_TY_USERDIAG1( "  Bitrate      = %s\n", m_BitrateText[bitidx][hdr.u.mpeg.BitrateIndex] );
      if ( !m_DumpAllHdrs ) {
        LOG_TY_USERDIAG2( "  SampleRate   = %s -> %d\n", \
        m_SampleRateText[sampleidx][hdr.u.mpeg.SampleRateIndex], m_EncodeSettings.u.mpeg.SampleRate );
      }
      else {
        LOG_TY_USERDIAG1( "  SampleRate   = %s\n", m_SampleRateText[sampleidx][hdr.u.mpeg.SampleRateIndex] );
      }

      LOG_TY_USERDIAG1( "  Padding      = %d\n", hdr.u.mpeg.Padding );
      LOG_TY_USERDIAG1( "  Private      = %d\n", hdr.u.mpeg.Private );
      LOG_TY_USERDIAG1( "  Channel Mode = %s\n", m_ChannelModeText[hdr.u.mpeg.ChannelMode] );
      LOG_TY_USERDIAG1( "  Ext Mode     = %d\n", hdr.u.mpeg.ModeExtension );
      LOG_TY_USERDIAG1( "  Copyright    = %s\n", m_CopyrightText[hdr.u.mpeg.Copyright] );
      LOG_TY_USERDIAG1( "  Original     = %s\n", m_OriginalText[hdr.u.mpeg.Original] );
      LOG_TY_USERDIAG1( "  Emphasis     = %s\n", m_EmphasisText[hdr.u.mpeg.Emphasis] );
      break;

    case AC3_FRAME_SYNC:
      LOG_TY_USERDIAG2( "AC3 Hdr %d, len = %d\n", framecnt, hdr.u.ac3.FrameLen );

      LOG_TY_USERDIAG1( "  FrameSync  = 0x%x\n", hdr.FrameSync );
      LOG_TY_USERDIAG1( "  Bitrate    = %d\n", hdr.u.ac3.BitRate );
      LOG_TY_USERDIAG2( "  SampleRate = %d -> %d\n", hdr.u.ac3.SampleRate, m_EncodeSettings.u.ac3.SampleRate );
      LOG_TY_USERDIAG2( "  Channels   = %s%s\n", m_Ac3ChannelTxt[hdr.u.ac3.Flags & A52_CHANNEL_MASK], (hdr.u.ac3.Flags & A52_LFE) ? ", Sub Channel is present" : "" );
      break;

    default:
      LOG_TY_USERDIAG2( "Error: Hdr %d Unknown header type (0x%X)\n", framecnt, hdr.FrameSync );
      break;
  }
}

// ---------------------------------------------------------------------
int HdrChangeCheck( PFRAME_HEADER_INFO thdr, PFRAME_HEADER_INFO hdr )
{
  if ( thdr->FrameSync != hdr->FrameSync ) {
    return 1;
  }
  switch ( thdr->FrameSync ) {
    case MPEG_FRAME_SYNC:
      return memcmp( &thdr->u.mpeg, &hdr->u.mpeg, sizeof( hdr->u.mpeg ) );

    case AC3_FRAME_SYNC:
      return memcmp( &thdr->u.ac3, &hdr->u.ac3, sizeof( hdr->u.ac3 ) );
  }
  return 0;
}

// ---------------------------------------------------------------------
void ParseFile()
{
  FRAME_HEADER_INFO hdr;
  FRAME_HEADER_INFO thdr;
  long              framecnt = 0;   // make part of control
  unsigned long     frameSize;
  int               flen;
  void             *fptr;
  int               err;
  char              time_str[255];
  int               cnt;
  int               db;


  if ( m_InputFn ) {
    LOG_TY_USERDIAG1( "\nParsing input file '%s':\n", m_InputFn );
  }
  memset( &hdr, 0, sizeof( hdr ) );

  LOG_TY_USERDIAG1( "Output Format: %s\n", (m_EncodeSettings.FrameSync == MPEG_FRAME_SYNC) ? "MPEG" : "AC3" );

  // clear out memory
  memset( &m_Convert, 0, sizeof( m_Convert ) );

  // create db adjust coef table
  for ( cnt = 0, db = -(MAX_DB_RANGE * DB_STEP); cnt < (MAX_DB_RANGE * 2) + 1; db += DB_STEP, cnt++ ) {
    m_DbAdjustCoefTbl[cnt] = pow( 10.0, (((double)db) / 20.0) );
  }

  // init converter
  err = AudioInitConverter( &m_Convert, &m_EncodeSettings, m_DbAdjust );
  if ( err != AC_OK ) {
    LOG_TY_ERROR1( "ERROR: %s\n", m_ErrorTxt[abs( err )] );
    return;
  }

  // Set up the input and output data streams
  PipeInit( &(m_Convert.InputStream), m_hPipe_In, 0, STREAM_BUFFERSIZE );

  // make sure we have not reached the end of the stream
  while ( !thread_pipe_iseof( m_hPipe_In ) ) {

    // try to find a valid frame hdr
    err = AudioFindFrameHeader( &m_Convert, &thdr );
    if ( err != AC_OK ) {
		  if( err == AC_NO_FRAME_FOUND && thread_pipe_iseof( m_hPipe_In ) ) {
	        LOG_TY_USERDIAG( "End of Data Reached\n" );
		  }
		  else {
	        LOG_TY_ERROR1( "ERROR: %s\n", m_ErrorTxt[abs( err )] );
		  }
      break;
    }

    if ( !m_RawInputSampleRate && (!framecnt || HdrChangeCheck( &thdr, &hdr )) ) {

      // display audio format change
      if ( framecnt ) {
        LOG_TY_PROGRESS1( "Frame: %d\r", ++framecnt );
        LOG_TY_USERDIAG( "\nAudio Format Change detected:\n" );
      }

      // save the new hdr
      hdr = thdr;

      // display hdr
      DumpHdr( hdr, framecnt );

      // make sure this is mpeg1 layer II only
      if ( hdr.FrameSync == MPEG_FRAME_SYNC ) {

        if ( (hdr.u.mpeg.AudioVer != MPEG_1_VER) || (hdr.u.mpeg.LayerDescrition != MPEG_LAYER2) ) {
          LOG_TY_ERROR2( "%s %s is not a support format\n", m_AudioVerText[hdr.u.mpeg.AudioVer], m_LayerText[hdr.u.mpeg.LayerDescrition] );
          return;
        }
      }
    }
    if ( !(++framecnt % 100) ) {
      if ( m_FileSize ) {
        if ( !m_RawInputSampleRate ) {
          switch ( hdr.FrameSync ) {
            default:
            case MPEG_FRAME_SYNC:
              frameSize = hdr.u.mpeg.FrameLen;
              break;

            case AC3_FRAME_SYNC:
              frameSize = hdr.u.ac3.FrameLen;
              break;
          }
        }
        else {
          frameSize = 4;  // short * 2 channels
        }
        LOG_TY_PROGRESS2( "Frame: %d of %d\r", ++framecnt, (m_FileSize / frameSize) );
      }
      else {
        LOG_TY_PROGRESS1( "Frame: %d\r", ++framecnt );
      }
    }

    // are we just dumping hdrs
    if ( !m_DumpAllHdrs ) {

      // now decode frame
      err = AudioProcessFrame( &m_Convert );
      if ( err != AC_OK ) {
        LOG_TY_ERROR1( "\nERROR: %s\n", m_ErrorTxt[abs( err )] );
        break;
      }

      // get output frame(s)
      while ( (fptr = AudioGetOutputFrame( &m_Convert, &flen )) != NULL ) {

        // flush it to output pipe
        thread_pipe_write( m_hPipe_Out, flen, fptr );
      }
    }

    // move pointer to next frame
    switch ( hdr.FrameSync ) {
      case MPEG_FRAME_SYNC:
        if ( m_DumpAllHdrs ) {
          // we need to manually skip past the frame
          PipeSkipBytes( &m_Convert.InputStream, m_Convert.Decode.FrameHdr.u.mpeg.FrameLen );
        }
        break;

      case AC3_FRAME_SYNC:
        if ( m_DumpAllHdrs ) {
          // we need to manually skip past the frame
          PipeSkipBytes( &m_Convert.InputStream, m_Convert.Decode.FrameHdr.u.ac3.FrameLen );
        }
        break;

      default:
        if ( !m_RawInputSampleRate ) {
          return;
        }
        // raw PCM does not have a hdr
        break;
    }
  }

  if ( !m_DumpAllHdrs ) {
    // get any leftover output frames
    while ( (fptr = AudioGetLastFrame( &m_Convert, &flen )) != NULL ) {

      // flush it to output pipe
      thread_pipe_write( m_hPipe_Out, flen, fptr );
      framecnt++;
    }
  }

  timer_string( m_Convert.Decode.Stats.TotalPlayTime, time_str, "%lu:%02u:%02u.%u" );

  LOG_TY_USERDIAG2( "%d frames decoded (%s)\n", framecnt, time_str );

  if ( !m_DumpAllHdrs ) {
    if ( hdr.FrameSync == MPEG_FRAME_SYNC ) {
      LOG_TY_USERDIAG1( "%+.1f dB peak amplitude\n", 20 * log10( mad_f_todouble( m_Convert.Decode.Stats.PeakSample ) ) );
    }
    else {
      LOG_TY_USERDIAG1( "%+.1f dB peak amplitude\n", 20 * log10( (double)m_Convert.Decode.Stats.PeakSample ) );
    }

    LOG_TY_USERDIAG1( "%lu clipped samples\n", m_Convert.Decode.Stats.ClippedSamples );
  }

  AudioShutDownConverter( &m_Convert );
}

// ---------------------------------------------------------------------
void HelpScreen( char *pname, int runaslib )
{
  if ( !runaslib ) {
    printf( "%s: -i inputfile -o outputfile [options]\n\n", pname );
    printf( "  options are:\n" );
    printf( "        -i inputFile  - input filename\n" );
    printf( "        -o outputFile - output filename, .raw ext for raw PCM output\n" );
    printf( "        -d            - dump all headers (as they change) and quit\n" );
    printf( "        -m            - output format is MPEG (default)\n" );
    printf( "        -a            - output format is AC3\n" );
    printf( "        -s rate       - new sample rate (i.e. 32000 or 44100 or 48000 (default)\n" );
    printf( "        -b rate       - new bit rate (i.e. 192000 (default for mpeg),\n" );
    printf( "                        384000 (default for AC3)\n" );
    printf( "        -l dbLevel    - (-)attenuate or (+)amplify in Db,\n" );
    printf( "                        default: ac3->mpeg (+10 Db), mpge->ac3 (+5)\n" );
    printf( "                        Warning: if the db gain is too high you will\n" );
    printf( "                        get clipping, it will sound like you have static\n" );
    printf( "                        in the audio\n" );
    printf( "        -r rate       - raw PCM input, with sample rate (stereo only)\n" );
    printf( "        -h            - get help\n\n" );
  }
}

// ---------------------------------------------------------------------
int ProcessArgs( int argc, char *argv[], int runaslib )
{
  int         flags;
  int32_t     sampleRate = 48000;   // default to DVD rate
  int32_t     bitRate    = 0;
  FRAME_SYNCS frameSync  = MPEG_FRAME_SYNC;
  int         fh;
  struct stat stat;

  // save running mode
  m_RunAsLib = runaslib;

  // if no arguments
  if ( !m_hPipe_In && argc <= 1 ) {
    HelpScreen( argv[0], runaslib );
    return 1;
  }

  tygetopt_lock();

  // Command line parsing
  while ((flags = getopt( argc, argv, "pdma?hi:o:s:b:l:r:" )) != -1) {
    switch ( flags ) {
      case 'i':
        if ( optarg[0]=='-' ) HelpScreen( argv[0], runaslib );
        m_InputFn = optarg;
        break;

      case 'o':
        if ( optarg[0]=='-' ) HelpScreen( argv[0], runaslib );
        m_OutputFn = optarg;
        break;

      case 's':
        if ( optarg[0]=='-' ) HelpScreen( argv[0], runaslib );
        sampleRate = atol( optarg );
        break;

      case 'b':
        if ( optarg[0]=='-' ) HelpScreen( argv[0], runaslib );
        bitRate = atol( optarg );
        break;

      case 'l':
        if ( optarg[0]=='-' ) HelpScreen( argv[0], runaslib );
        m_DbAdjust = atol( optarg );
        if ( !m_DbAdjust ) {
          m_DbAdjust = -1;
        }
        break;

      case 'r':
        if ( optarg[0]=='-' ) HelpScreen( argv[0], runaslib );
        m_RawInputSampleRate = atol( optarg );
        break;

      case 'd':
        m_DumpAllHdrs = 1;
        break;

      case 'p':
        m_OutputRawPcm = 1;
        break;

      case 'm':
        frameSync  = MPEG_FRAME_SYNC;
        break;

      case 'a':
        frameSync  = AC3_FRAME_SYNC;
        break;

      case '?':
      case 'h':
      default:
        HelpScreen( argv[0], runaslib );
		tygetopt_unlock();
        return 1;
    }
  }

  tygetopt_unlock();

  // check input file
  if ( !m_hPipe_In  && !m_InputFn ) {
    HelpScreen( argv[0], runaslib );
    return 1;
  }

  // if not -d check output file
  if ( !m_hPipe_Out && (!m_DumpAllHdrs && !m_OutputFn)) {
    HelpScreen( argv[0], runaslib );
    return 1;
  }

  // make sure rate is value
  switch ( sampleRate ) {
    case 32000:
    case 44100:
    case 48000:
      // we are good to go
      break;

    default:
      HelpScreen( argv[0], runaslib );
      return 1;
  }

  // make sure rate is value
  if ( m_RawInputSampleRate ) {
    switch ( m_RawInputSampleRate ) {
      case 32000:
      case 44100:
      case 48000:
        // we are good to go
        break;

      default:
        HelpScreen( argv[0], runaslib );
        return 1;
    }
  }

  // check for raw output
  if ( !runaslib && m_OutputFn ) {

    if ( strstr( m_OutputFn, ".raw" ) || strstr( m_OutputFn, ".RAW" ) ) {
      m_RawOutput = 1;
    }
  }

  // set encoding struct
  switch ( frameSync ) {
    case MPEG_FRAME_SYNC:
      // fix bitRate if needed
      if ( !bitRate ) {
        // set default
        bitRate = 192000;
      }
      m_EncodeSettings.FrameSync          = MPEG_FRAME_SYNC;
      m_EncodeSettings.u.mpeg.SampleRate  = sampleRate;
      m_EncodeSettings.u.mpeg.BitRate     = bitRate;
      m_EncodeSettings.u.mpeg.ChannelMode = CHANNEL_MODE_STEREO;
      m_EncodeSettings.u.mpeg.PsychoModel = PSYCO_MODEL_1;
      break;

    case AC3_FRAME_SYNC:
      // fix bitRate if needed
      if ( !bitRate ) {
        // set default
        bitRate = 384000;
      }
      m_EncodeSettings.FrameSync         = AC3_FRAME_SYNC;
      m_EncodeSettings.u.ac3.SampleRate  = sampleRate;
      m_EncodeSettings.u.ac3.BitRate     = bitRate;
      break;

    default:
      HelpScreen( argv[0], runaslib );
      return 1;
  }

  // are we running in standalone mode
  if ( !runaslib && m_InputFn ) {

    fh = open( m_InputFn, OPEN_READONLY_FLAGS );
    if ( fh != -1 ) {

      // get file size
      if ( fstat( fh, &stat ) != -1 ) {
        m_FileSize = stat.st_size;
      }
      close( fh );
    }
  }
  return( 0 );
}

// ---------------------------------------------------------------------
/* Failsafe to ensure the caller is not permanently blocked if we exit prematurely */
void transcode_exit_handler( void )
{
  int j;

  // clean up arguments
  if ( m_argc && m_argv ) {
		for( j = 0; j < m_argc; ++j ) {
		  free( m_argv[j] );
		}
		free( m_argv );

    m_argv = NULL;
    m_argc = 0;
  }

	if( m_write_pipe ) {
		LOG_FATAL("transcode finished prematurely and did not exit gracefully\n");
		thread_pipe_eof( m_write_pipe, 0 );
		m_write_pipe = 0;
	}
	if( m_process_lock ) {
		thread_mutex_unlock( m_process_lock );
		m_process_lock = 0;
	}
}

// ---------------------------------------------------------------------
/* Entry point for the mplex thread */
int transcode_thread( void *user_data )
{
	int j;
  char buf[15];
  FILE *logfile = stderr;

  transcode_start_params_t *params;

  atexit( transcode_exit_handler );

  if( !user_data )
    return -1;   /* We need some parameters, something has gone horribly wrong  */
  params = ( transcode_start_params_t *) user_data;

  /* Initialise the logger */
  if( params->logfile ) {
	logfile = params->logfile;
  }
  logger_init( params->loglevel, logfile, params->loghookfn, 1 );

  /* Use the pipes passed in the parameters */
  m_hPipe_In = params->input_pipe;
  m_hPipe_Out = params->output_pipe;
  m_write_pipe = m_hPipe_Out;

  // were argument supplied?
  if ( !params->argc && !params->argv ) {

    // make our own
		params->argv = (char **) malloc( sizeof( char *) * 8 );
		params->argv[0] = strdup( "-tcode" );
		++params->argc;
    sprintf( buf, "-%c", params->format ? 'a' : 'm' );
		params->argv[1] = strdup( buf );
		++params->argc;
		params->argv[2] = strdup( "-s" );
		++params->argc;
    sprintf( buf, "%d", params->rate );
		params->argv[3] = strdup( buf );
		++params->argc;
		params->argv[4] = strdup( "-b" );
		++params->argc;
    sprintf( buf, "%d", params->bitrate );
		params->argv[5] = strdup( buf );
		++params->argc;
		params->argv[6] = strdup( "-l" );
		++params->argc;
    sprintf( buf, "%d", params->dbadjust );
		params->argv[7] = strdup( buf );
		++params->argc;

    // save them
    m_argv = params->argv;
    m_argc = params->argc;
  }

  /* Process tytranscode commands */
  ProcessArgs( params->argc, params->argv, 1 );

  // clean up arguments
  if ( m_argc && m_argv ) {
		for( j = 0; j < params->argc; ++j ) {
		  free( params->argv[j] );
		}
		free( params->argv );

    m_argv = params->argv = NULL;
    m_argc = params->argc = 0;
  }

  if( params->thread_lock ) {
    /* Lock the main process while we are running */
    thread_mutex_lock( params->thread_lock );
  }
  m_process_lock = params->thread_lock;

  /* Tell the main program we are running and that it should block itself on the thread_lock mutex */
  params->running = 1;

  /* All set up, let's go */
  ParseFile();

  /* Mark EOF on our output pipe */
  thread_pipe_eof( m_hPipe_Out, 1 );
  m_write_pipe = 0;

  logger_free();

  if( params->thread_lock ) {
    /* All done so release the main process */
    thread_mutex_unlock( params->thread_lock );
  }
  m_process_lock = 0;

  return 0;
}

